#include <stdio.h>
#include <stdlib.h>

#include "tinecdi.h"
#include "database.h"
#include "cdi.h"	// interface to user program
#include "cdiCanLib.h"

#ifndef CDI_CAN_LINUX
# include "windows.h"

#else

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <memory.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <setjmp.h>

#include <math.h>
#include <stropts.h>	/* ioctl	*/
#include <semaphore.h>

#endif

int n_baudrate = 14;
int esdBaudrate[] = {1000, 667, 500, 333, 250, 166, 125, 100, 67, 50, 33, 20, 13, 10};

#ifdef CDI_CAN_LINUX
__attribute__((constructor)) void cdiSedacMain(void)
{
  int cc = 0;

  if ((cc=cdiRegisterBusFilter(CAN_LIB_NAME,filterCanbus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusConfiguration(CAN_LIB_NAME,configCanCard)) != 0) goto err;
  if ((cc=cdiRegisterBusInitialization(CAN_LIB_NAME,initCanLineForCDI)) != 0 ) goto err;
  if ((cc=cdiRegisterBusHandler(CAN_LIB_NAME,execCanbus)) != 0 ) goto err;
  if ((cc=cdiRegisterBusCleanup(CAN_LIB_NAME,terminateCanLine)) != 0 ) goto err;

err:
  if (cc)
  {
    cdilog("%s : failure in bus registration CAN",erlst[cc]);
  }
  return;
}
#endif

int NGdebug = 1;
int canMsgPrint = 1;
#define MAX_BLEN 256

CAN_LINE canLine[MAX_NUMBER_CAN_LINES] = {
#ifdef CDI_CAN_LINUX
	{0}, {0}, {0}, {0}
#else
	{NULL}, {NULL}, {NULL}, {NULL}
#endif
};

CDI_CAN_EXPORT int canScanFcn(char *target,char *result,int resultBufferSize)
{
  return 0;
}
CDI_CAN_EXPORT int filterCanbus(int *parameter, int accessFlag)
{
  if (parameter == NULL) return argument_list_error;
  if (accessFlag != cdiWriteFlag &&
      (parameter[canOpenId] == PDO_TRANSMIT || parameter[canOpenId] == PDO2_TRANSMIT || parameter[canOpenId] == NMT_COMMANDS) ) return argument_list_error;
  return 0;
}

/****************************************************************************/
/* send request! input active flag = transRequest or transRequest2nd */
/* after, set active flag to finishRequest, recvRequest, or recvRequest2nd */
/****************************************************************************/
CDI_CAN_EXPORT void execCanbus(CdiRequestInfoBlk *pReq)
{ 
  int canRet, wFlag;

  if ( pReq->active == transRequest ) /* start first transfere */
  {
    pReq->active = recvRequest; /* afterwards goto receive */
    if ( (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) &&
    pReq->devParameters[canAutoReceive] ) return; /* auto receive, without request! */
    wFlag = (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag) ? 1 : 0;
  }
  else
  {
    pReq->active = recvRequest2nd; /* goto 2ndReceive */
    if ( pReq->accessFlag == cdiWriteReadFlag && pReq->devParameters[canAutoReceive] ) return;
    wFlag = pReq->accessFlag == cdiReadWriteFlag ? 1 : 0;
  }

  canRet = canMsgSend((int)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine,
              (short)(pReq->devParameters[canBroadcast] ? 0 : pReq->pDev->cdiAddr),
              (short)pReq->devParameters[canOpenId], 
              (short)pReq->outLength, 
              (char *)pReq->pOutData,
              (short)pReq->devParameters[canOpenObj], 
              (short)pReq->devParameters[canOpenSubIndex], 
              (short)wFlag);

  if ( NGdebug )
    cdilog("canMsgSend return code: %d! line: %d\n",canRet, ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine);
  
  if ( canRet < 0 )
  {
    pReq->errorCode = -canRet; /* write failed */
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( pReq->errorCode == hardware_error )
    {
      pReq->deviceError = net_write_error;    // can write error
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest;
    return;
  }
  /* without message back */
  if ( (pReq->devParameters[canOpenId] == PDO_TRANSMIT || pReq->devParameters[canOpenId] == PDO2_TRANSMIT || pReq->devParameters[canOpenId] == NMT_COMMANDS) &&
  pReq->accessFlag == cdiWriteFlag ) pReq->active = finishRequest;

  cdilog("active = %d; preq addr: 0x%x\n", pReq->active, pReq);

  return;
}

/****************************************/
/*	can receive thread					*/
/****************************************/
#ifdef CDI_CAN_LINUX
void *GetCanDataThread(void *param)
#else
DWORD WINAPI GetCanDataThread(void *param)
#endif
{ int i;
  CAN_LINE *pCan = (CAN_LINE *)param;
  do 
  {
    pCan->recLength = N_MSGS;
    pCan->rcvStatus = canRead(pCan->fd, pCan->cmsg, &(pCan->recLength), NULL);
    if ( pCan->rcvStatus == NTCAN_SUCCESS ) for ( pCan->recCnt = 0; pCan->recCnt < pCan->recLength; pCan->recCnt++ )
    {
      pCan->rcvPmsg = &pCan->cmsg[pCan->recCnt];
      pCan->retId = (short)pCan->rcvPmsg->id;
      pCan->rtr = (pCan->rcvPmsg->len >> 4) & 0x1;
      pCan->length = pCan->rcvPmsg->len & 0xf;

      memcpy(pCan->retData, pCan->rcvPmsg->data, pCan->length);

      pCan->remoteNode = pCan->retId & 0x7f;
      pCan->remoteCommand = pCan->retId & 0xff80;
  
    if ( pCan->remoteCommand != NMT_ERROS )
    {
      if ( canMsgPrint )
      {
        printf("Master ID %x, rtr: %d; length: %d : ", pCan->retId, pCan->rtr, pCan->length);
        for(i=0; i<pCan->length; i++) printf("%2x ", ((int)pCan->retData[i] & 0xff));
        printf("\n");
      }
      if ( pCan->receiveCanMessage ) pCan->receiveCanMessage(pCan->cdiLine, pCan->retData, pCan->length, pCan->retId, pCan->rtr, pCan->remoteNode, pCan->remoteCommand);
    }

    }
    else
      cdilog("ESD CAN revcice Can error! %d\n",pCan->rcvStatus);
  } while(1);
}

/****************************************/
/*	CAN send API						*/
/*	return -1: line number wrong		*/
/*	return -2: not init yet				*/
/*	return =0: send error				*/
/****************************************/
int canMsgSend(int line, short nodeId, short canOpenId, short length, char *data, short ObjNum, short subIndex, short wFlag)
{ CAN_LINE *pCan;
  int rtr;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
  if ( pCan->fd < 0 ) return -file_not_open;

  if ( nodeId < 0 ) nodeId = 0;

//  printf("***canMsgSend ==> line: %d; nodeID: %d; CanOPENID: %d\n",line, nodeId, canOpenId);

  switch (canOpenId)
  {
  	case NMT_COMMANDS:
	    canOpenId = NMT_COMMANDS;
	    rtr = 0;
	    length = 2;
	    pCan->xmtMsg.data[0] = data[0];	// only one byte data for EXEC_REMOTE
	    pCan->xmtMsg.data[1] = (unsigned char)nodeId;
	    break;
	  case SDO_TRANSMIT:
	    pCan->xmtMsg.data[0] = wFlag ? SDO_WRITE : SDO_READ;
	    *(short *)&pCan->xmtMsg.data[1] = ObjNum;	// object 6300; sub index 1
	    pCan->xmtMsg.data[3] = (unsigned char)subIndex;
	    if ( wFlag ) memcpy(&pCan->xmtMsg.data[4],data,length);
//      if ( wFlag && length == 4 ) printf("can save data: %x\n",*(int *)&FastMsg.data[6]);
	    length = 8;
	    rtr = 0;
      break;
  	case PDO_TRANSMIT:
  	case PDO2_TRANSMIT:
	    memcpy(&pCan->xmtMsg.data[0],data,length);
	    rtr=0;
	    break;
	  case PDO_DI_READ:
	  case PDO_AI_READ:
	    length = 0;
	    rtr = 1;
	    break;
	  default:
	    cdilog("canSendAPI: ****Wrong CanOpenID: 0x%x; line: %d\n",canOpenId, line+1);
	    return -3;
  }
  pCan->xmtMsg.id = canOpenId + nodeId;
  pCan->xmtMsg.len = (rtr ? 0x10 : 0) + length;
  pCan->xmtLength = 1;

  cdilog("id: 0x%x; len: 0x%x\n",pCan->xmtMsg.id, pCan->xmtMsg.len);

  pCan->xmtStatus = canWrite(pCan->fd, &pCan->xmtMsg, &pCan->xmtLength, NULL);
  if ( pCan->xmtStatus != NTCAN_SUCCESS )
  {
    pCan->numSendErrors++;
    cdilog("ESD CAN send Error! return code: 0x%x\n",pCan->xmtStatus); 
    return -hardware_error;
  }
  return 0;
}

/****************************************/
/*	load Can card information			*/
/****************************************/
CDI_CAN_EXPORT int configCanCard(void)
{ int i;
  CAN_LINE *pCan;

  for ( i=0; i<MAX_NUMBER_CAN_LINES; i++ )
  {
    pCan = &(canLine[i]);
#ifdef JANZCAN_H_
    pCan->canRate = (i%2) == 0 ? BTR_500KB : BTR_20KB;
    pCan->canCardType = JANZ_PC_CARD;
#endif
    pCan->dpm_x = i/2;		// card number
    pCan->dpm_y = i%2;		// channel number
    pCan->lineNumber = i;
  }
  return 0;
}

/****************************************/
/*	translateBaudRate					*/
/****************************************/
int transBaudRate(int baudrate)
{ int rate;

#ifdef ESDCAN_H_
  for ( rate=0; rate<n_baudrate; rate++ ) if ( baudrate == esdBaudrate[rate] ) break;
  return rate < n_baudrate ? rate : -1;

#else
  switch ( baudrate )		// set baud rate
  {
	case 800:
		rate = BTR_800KB;
		break;
	case 500:
		rate = BTR_500KB;
		break;
	case 250:
		rate = BTR_250KB;
		break;
	case 125:
		rate = BTR_125KB;
		break;
	case 50:
		rate = BTR_50KB;
		break;
	case 20:
		rate = BTR_20KB;
		break;
	case 10:
		rate = BTR_10KB;
		break;
	case 1000:
		rate = BTR_1MB;
		break;
	default:
		printf(" baud rate not supported: %d",baudrate);
		return -1;
  }
  return rate;
#endif
}

/****************************************/
/*	init can line						*/
/****************************************/
/**************************************/
void canMessage(int cdiLine, char *data, int leng, int retId, int rtr, int remoteNode, int remoteCommand)
{ 
  /* CdiLineInfoBlk *pLine = busHandle[cdiLine]; */
  CdiLineInfoBlk *busHandle;
  CdiRequestInfoBlk *pReq;
  short readFlag, returnObj;
  char *str;
  unsigned int datalen;

  /* find out device */
  if ((busHandle=cdiGetBusHandle(cdiLine)) == NULL)
  {
    cdilog("canMessage: Wrong cdi_line: %d for canMessage() call\n",cdiLine);
    return;
  }
  
  if (NGdebug) 
    cdilog("CanMessage call Back: cdi line: %d, can line number: %d, data: %s, length: %d, retId: %d, rtr: %d\n",
            cdiLine, busHandle->busLine, data, leng, retId, rtr);
    cdilog("Number of active request: %d\n",busHandle->numberActiveRequest);

  for ( busHandle->readCounter=0; busHandle->readCounter < busHandle->numberActiveRequest; busHandle->readCounter++ )
  {
    cdilog("enter loop ! Number of active request: %d; active flag = %d; preq addr: 0x%x\n",busHandle->numberActiveRequest, busHandle->pReqList[ busHandle->readCounter ].active, &busHandle->pReqList[ busHandle->readCounter ]);

    if ( (busHandle->pReqList[ busHandle->readCounter ].active != recvRequest &&
      busHandle->pReqList[ busHandle->readCounter ].active != recvRequest2nd) ||
      busHandle->pReqList[ busHandle->readCounter ].pDev->cdiAddr != remoteNode ) continue;


    pReq = &busHandle->pReqList[ busHandle->readCounter ];
    if ( pReq->active == recvRequest ) /* first part */
      readFlag = (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag) ? 1 : 0;
    else /* second part! */
      readFlag = (pReq->accessFlag == cdiWriteReadFlag) ? 1 : 0;
    returnObj = *(short *)&data[1];

    datalen = leng;
    str = data;
    if ( pReq->devParameters[canOpenId] == SDO_TRANSMIT ) /* check messagesType */
    {
      if ( remoteCommand != SDO_RECEIVE ) continue;
      if ( pReq->devParameters[canOpenObj] != returnObj || (char)pReq->devParameters[canOpenSubIndex] != data[3] ) continue;
      if ( (data[0] & 0xf0) == 0x80 ) // sdo abort Domain Transfer
      {
        pReq->deviceError = *(int *)&data[4];
        if ( pReq->deviceError == 0x05040001 ) pReq->errorCode = illegal_mode;
        else if ( pReq->deviceError == 0x06010002 ) pReq->errorCode = illegal_operator;
        else if ( pReq->deviceError == 0x06020000 ) pReq->errorCode = invalid_index;
        else if ( pReq->deviceError == 0x06070010 ) pReq->errorCode = wrong_type;
        else if ( pReq->deviceError == 0x06090011 ) pReq->errorCode = illegal_equipment_number;
        else pReq->errorCode = not_defined;
        if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
        if (NGdebug) cdilog("SDO abort command: %x",pReq->deviceError);
        cdiExecRequest(&(busHandle->pReqList[busHandle->readCounter]), readTaskIndex);
        continue;
      }
      else if ( readFlag ) /* SDO read */
      {
        if ( (data[0] & 0xf0) != 0x40 ) continue;
        if ( (data[0] & 0xf) == 0xf ) datalen = 1;
        if ( (data[0] & 0xf) == 0xb ) datalen = 2;
        if ( (data[0] & 0xf) == 0x7 ) datalen = 3;
        if ( (data[0] & 0xf) == 0x3 ) datalen = 4;
        else datalen = 4;
      }
      else   /* SDO write */
      {
        if ( (data[0] & 0xf0) != 0x60 ) continue;
        datalen = leng - 4;
      }
      /* SDO data back.... */
      datalen = leng - 4;
      str = &data[4];
    }
    else if ( pReq->devParameters[canOpenId] != remoteCommand ) continue; /* same OpenId? */
    else if ( readFlag == 0 ) /* for write command device should not answer! */
    {
      if (NGdebug) 
      {
        cdilog("canMessage Wrong: unexpected return message for write conmmand!");
        cdilog("CanOpenId: %d; canLine: %d; module CanId: %d; module Name: %s",
                remoteCommand,busHandle->busLine,remoteNode,pReq->pDev->devName);
      }
      continue;
    }

    if ( readFlag ) /* check return condition */
    {
      if (NGdebug) cdilog("CAN data check condition for readFlag canOpenID: %d",remoteCommand);
      checkReturnCondition(pReq, pReq->pInData, str, datalen, 0);
      if ( pReq->dataReturnCondition == 0 ) continue; 
    }

    if (NGdebug) cdilog("can data return condition OK canOpenID: %d",remoteCommand);
    
    memcpy( pReq->pInData, str, datalen > pReq->inLength ? pReq->inLength : datalen);

    cdiExecRequest(&(busHandle->pReqList[busHandle->readCounter]), readTaskIndex);
  }
}

#define CDI_PARAMETER_DELIMETER ":"
int cdiParseParamString(char *paramstr,int *parameters)
{
  char *s,*c=NULL;
  int i;
  if (paramstr == NULL || parameters == NULL) return argument_list_error;
  for (s=strtok_r(paramstr,CDI_PARAMETER_DELIMETER,&c), i=0; 
       i<NUMBER_EXEC_PARAM && s != NULL; 
       s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c), i++)
  {
    parameters[i] = atoi(s);
  }
  return 0;
}

CDI_CAN_EXPORT int initCanLineForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  int parameter[NUMBER_EXEC_PARAM];
  memset(parameter,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdiParseParamString(parameterList,parameter);
  return initCanLine(busLine,parameter,canMessage,cdiLine,numberDevices);
}

int initCanLine(int line, int *parameters, void (*callBack)(int, char *, int, int, int, int, int), int cdiLine, int numberDevices)
{ int ana, id;
  CAN_LINE *pCan;
  DWORD dwCdiThreadId;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES )
  {
    cdilog("Open wrong Can line: %d; valid lines: 1 to %d\n",line, MAX_NUMBER_CAN_LINES);
    return -illegal_line;
  }
  line--;

  pCan = &(canLine[line]);
#ifdef CDI_CAN_LINUX
  if ( pCan->fd != 0 )
#else
  if ( pCan->fd != NULL )
#endif
  {
    cdilog("Can line: %d already opened! \n",line+1);
    return 0;
  }

  pCan->canRate = transBaudRate( parameters[paramCanRate] );
  if ( pCan->canRate == -1 )
  {
    cdilog("Can line: %d; with Baudrate: %d not supported! \n",line,parameters[paramCanRate]);
    return -argument_list_error;
  }
  pCan->NumberDevices = numberDevices;
  pCan->cdiLine = cdiLine;
  pCan->receiveCanMessage = callBack;

  pCan->txqueuesize = NTCAN_MAX_TX_QUEUESIZE;
  pCan->rxqueuesize = NTCAN_MAX_RX_QUEUESIZE;
  pCan->txtimeout = 0;
  pCan->rxtimeout = 0;
  pCan->mode = 0;
  pCan->net = line;

  ana = canOpen(pCan->net, pCan->mode, pCan->txqueuesize, pCan->rxqueuesize, pCan->txtimeout, pCan->rxtimeout, &(pCan->fd));
  if ( ana != NTCAN_SUCCESS )
  {
    cdilog("canOpen() failed with error: %d; ESD can line: %d\n",ana, line + 1);
    return -file_not_open;    
  }

  ana = canSetBaudrate(pCan->fd, pCan->canRate);
  if ( ana != 0 )
  {
    cdilog("canSetBaudrate() failed with error: %d; ESD can line: %d\n",ana, line + 1);
    terminateCanLine(line + 1);
    return -file_not_open;    
  }

  for (id=0; id<MAX_CAN_IDS; id++)
  {
    ana = canIdAdd(pCan->fd, id+SDO_RECEIVE);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(pCan->fd, id+PDO_DI_READ);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(pCan->fd, id+PDO_AI_READ);
    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(pCan->fd, id+NMT_ERROS);
//    if ( ana == NTCAN_SUCCESS ) ana = canIdAdd(pCan->fd, id+EMERGENCE_MESSAGE);
    if ( ana != NTCAN_SUCCESS )
    {
      cdilog("canIdAdd() failed with error: %d; ESD can line: %d; id = %d\n",ana, line + 1, id+SDO_RECEIVE);
      terminateCanLine(line + 1);
      return -file_not_open;    
    }
  }

#ifdef CDI_CAN_LINUX
  if (pthread_create(&pCan->tIdMaster,NULL,GetCanDataThread,(void *)pCan) != 0)
#else
  if ( (pCan->tIdMaster = CreateThread(NULL,0,GetCanDataThread,(void *)pCan,0,&dwCdiThreadId)) == NULL )
#endif
  {
	  cdilog ("CreateThread foe CAN-ESD line: %d error",line + 1);
    terminateCanLine(line + 1);
	  return -un_allocated;
  }

  pCan->numSendErrors = 0;
  pCan->statusOK = 0;
  cdilog("canOpen() OK; ESD can line: %d\n", line + 1);

  return (int)pCan->fd;
}

/****************************************/
/*	terminate can line					*/
/****************************************/
CDI_CAN_EXPORT int terminateCanLine(int line)
{ CAN_LINE *pCan;
  int ret = 0;

  if ( line < 1 || line > MAX_NUMBER_CAN_LINES ) return -illegal_line;
  line--;

  pCan = &(canLine[line]);
#ifdef CDI_CAN_LINUX
  if ( pCan->fd == 0 ) return -file_not_open;
#else
  if ( pCan->fd == NULL ) return -file_not_open;
#endif

  if ( pCan->tIdMaster) cdiCancelThread((int)pCan->tIdMaster);
  ret = canClose( pCan->fd );
  if ( ret != NTCAN_SUCCESS )
  {
    cdilog("canClose() failed with error: %d; ESD can line: %d\n",ret, line + 1);
    ret = -file_not_open;    
  }
#ifdef CDI_CAN_LINUX
  pCan->fd = 0;
#else
  pCan->fd = NULL;
#endif

  return ret;
}
