#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#ifdef UNIX
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <sys/time.h>
#include <dlfnc.h>
#endif
#include "prolog.h"
#include "cdi.h"	// interface to user program
#include "cdiSedacLib.h"

int nSedacLines = N_SEDAC_LINES;
int SedacSimulate = -1;

__attribute__((constructor)) void cdiSedacMain(void)
{
    int cc = 0;

    if ((cc=cdiRegisterBusFilter("SEDPC",filterSedacBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusInitialization("SEDPC",initSedacForCDI)) != 0 ) goto err;
    if ((cc=cdiRegisterBusHandler("SEDPC",execSedacBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusCleanup("SEDPC",exitSedac)) != 0 ) goto err;
err:
    if (cc)
    {
      cdilog("%s : failure in bus registration SEDPC",erlst[cc]);
    }
    return;
}
SEDAC_STRUCT sedacInfo[N_SEDAC_LINES] = {
	{-1}, {-1}, {-1}, {-1},  {-1}, {-1}, {-1}, {-1}
};

int filterSedacBus(int *parameter, int accessFlag)
{
  if (parameter == NULL) return argument_list_error;
  return 0;
}
int initSedacForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{
  return initSedac(busLine);
}
int doSedacRequest(CdiRequestInfoBlk *pReq, int step)
{ int RWflag, ret;

  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);

  if ( step == 1 ) /* first request */
	  RWflag = (pReq->accessFlag == cdiWriteFlag || pReq->accessFlag == cdiWriteReadFlag) ? WRITE : READ;
  else /* second */
	  RWflag = pReq->accessFlag == cdiReadWriteFlag ? WRITE : READ;
  ret = sedac((u16)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine, 
          (u16)pReq->pDev->cdiAddr,
//          (u16)(pReq->pDev->cdiSubAddr + pReq->devParameters[step == 1 ? sedacSubOffset : sedacSubOffsetSecond]), 
          (u16)(pReq->pDev->cdiSubAddr + pReq->devParameters[RWflag == READ ? sedacSubOffset : sedacSubOffsetSecond]), 
          (int)RWflag,
          (u16 *)(RWflag == READ ? pReq->pInData : pReq->pOutData),
          (int)(RWflag == READ ? pReq->inLength : pReq->outLength) / sizeof(short) );
  if ( ret != 0 )
  {
    pReq->errorCode = ret < 0 ? -ret : sedac_error;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( ret > 0 ) pReq->pDev->numberDeviceError++;
    pReq->active = finishRequest; /* finish request */
  }

  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  return pReq->errorCode;
}
/********************************************************/
/* only call once with pReq->active == transRequest */
/********************************************************/
void execSedacBus(CdiRequestInfoBlk *pReq)
{ 
  /* do first part; */
  doSedacRequest(pReq, 1);
  if ( pReq->active == finishRequest ) return;

  /* check return condition */
  if (pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiReadWriteFlag)
  {
    checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, 2, 0); /* check first short data only */
    pReq->bufLastData[0] = *(short *)(pReq->pInData);

    if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
    { /* data back for sync call */
      if ( pReq->callback == NULL )
      {
        pReq->errorCode = invalid_data;	
        if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      pReq->active = finishRequest;
    }
  }
  if ( pReq->active == finishRequest ) return;

  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    return;
  }

  /* do second part */
  doSedacRequest(pReq, 2);
  if ( pReq->active == finishRequest ) return;

  /* check return condition */
  if (pReq->accessFlag == cdiWriteReadFlag)
  { /* check first short data only */
    checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, 2, 0);
    pReq->bufLastData[0] = *(short *)(pReq->pInData);
    if ( pReq->callback == NULL && pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
    {
      pReq->errorCode = invalid_data;
      if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    }
  }

  pReq->active = finishRequest; /* finish request */
}

/* ============  init routine  =================================== */
int initSedac(int ln)
{ char fname[80];
  SEDAC_STRUCT *pSedac;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  if (SedacSimulate) return ln;		// ln number as file handle!

  pSedac = &sedacInfo[ln];
  if ( pSedac->sedac_fd >= 0 )
  {
    cdilog("sedac already opened! sedac line: %d!\n", ln+1);
    return pSedac->sedac_fd;
  }

  sprintf(fname,"/dev/sedac%d",ln);
  if(( pSedac->sedac_fd = open(fname, O_RDWR) ) >= 0 )
  {
    pSedac->da = NULL;
    pSedac->dataSize = 0;	
    pSedac->sedacSimulationCount = 0;

    cdilog("Open pktr sedac line OK! line: %d\n",ln+1);	
    return pSedac->sedac_fd;
  }

  cdilog( "initSedac for PKTR card open() error! line: %d; strerror: %s", ln+1, strerror(errno) );
  fprintf(stderr, "%s: %s; error Number: %d\n", fname, strerror(errno), errno);
  return -file_not_open;
}

// ============  exit routine  ===================================
int exitSedac(int ln)
{ SEDAC_STRUCT *pSedac;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  if (SedacSimulate) return 0;

  cdilog("close sedac line! line number: %d", ln+1);

  pSedac = &sedacInfo[ln];
  if ( pSedac->sedac_fd < 0 ) return -file_not_open;
  
  close(pSedac->sedac_fd);
  pSedac->sedac_fd = -1;

  if ( pSedac->da )
  {
    free(pSedac->da);
    pSedac->da = NULL;
    pSedac->dataSize = 0;	
    pSedac->sedacSimulationCount = 0;
  }

  return 0;
}

// ============  getBuffer  ===================================
PSEDAC_DATA getBuffer(int ln, int number)
{ SEDAC_STRUCT *pSedac;

  pSedac = &sedacInfo[ln];
  if ( pSedac->da )
  {
    free(pSedac->da);
    pSedac->da = NULL;
  }

  pSedac->da = (PSEDAC_DATA)malloc( SEDAC_DATA_SIZE * (number + 1) );
  if ( pSedac->da != NULL )	pSedac->dataSize = number;
  else
  {
    cdilog("sedac pktr line: %d; => get buffer failed!! size: %d\n",ln+1,number);
    return NULL;
  }
  cdilog("get buffer: %x; size: %d",(int)(pSedac->da),number);
  return pSedac->da;
}
int msleep(int msec)
{
  struct timespec ts;
  if (msec <= 0) return 0;
  ts.tv_sec = msec/1000;
  ts.tv_nsec = (msec%1000) * 1000000;
  return nanosleep(&ts,NULL);
}
int execSedac(u16 ln, u16 *crate, u16 *subadd, int *RWF, u16 *data, int nwords, int flag)
{ int i, ret;
  SEDAC_STRUCT *pSedac;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  pSedac = &sedacInfo[ln];

  if ( SedacSimulate == 0 && (pSedac->da == NULL || pSedac->dataSize < nwords)
	&& getBuffer(ln, nwords) == NULL ) return -out_of_local_memory;
 
  if ( SedacSimulate )
  {
    i = nwords * 250;
//    if ( i > 4 ) msleep ( i/4 ); else msleep(1);
    for (i=0; i<nwords; i++, data++)
    {
      ret = flag ? RWF[i] : RWF[0];
      if ( ret == READ )
      *data = pSedac->sedacSimulationBuffer + pSedac->sedacSimulationCount++;
      else
      pSedac->sedacSimulationBuffer = *data;
    }
    return 0;
  }

  if ( pSedac->sedac_fd < 0 ) return -file_not_open;

  for (i=0; i<nwords; i++) 
  {
    ret = flag ? RWF[i] : RWF[0];
  	if ( ret == WRITE ) pSedac->da[i+1].Data = data[i];
    pSedac->da[i+1].ExcessMode = ret == READ ? SE_READ : SE_WRITE;
    pSedac->da[i+1].Line = (BYTE)ln;
    pSedac->da[i+1].SubAdr = flag ? subadd[i] : subadd[0];
    pSedac->da[i+1].CaAdr = flag ? crate[i] : crate[0];
  }
  ret = ioctl(pSedac->sedac_fd, nwords, pSedac->da);
  if ( ret == nwords + 1 )
  {
    for (i=0; i<nwords; i++)
      if ( (flag ? RWF[i] : RWF[0]) == READ ) data[i] = pSedac->da[i+1].Data;
    return ((PSEDAC_HEADER)pSedac->da)->NumberOfError;
  }

  if ( ret < 0 )
  {
    cdilog("pktr sedac ioclt() error: line: %d, getLastError: %s\n",ln+1, strerror(errno));
    fprintf(stderr, "ioctl error (request #Telegrams= %d): %s; return code: %d \n",nwords, strerror(errno), ret);
    return -software_error;
  }
  return -operation_timeout;
}

// ================  read/write nwords telegram to same addesses  ===================================
int sedac(u16 ln, u16 crate, u16 subadd, int RWF, u16 *data, int nwords)
{
  return execSedac(ln, &crate, &subadd, &RWF, data, nwords, 0);
}

// ================  read/write nwords telegram to different addresses ===================================
int nsedac(u16 ln, u16 *crate, u16 *subadd, int *RWF, u16 *data, int nwords)
{
  return execSedac(ln, crate, subadd, RWF, data, nwords, 1);
}

