#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <time.h>      /* time() */
#include <memory.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <setjmp.h>

#include <math.h>

#include "prolog.h"
#include "cdi.h"
#include "tinecdi.h"
#include "rpctypes.h"
#include "servdef.h"
#include "database.h"

int gSimulationFlag = FALSE;
CdiTemplate *CdiTemplateTbl[MAX_NUM_TEMPLATES];
int CdiTemplateTblSize = 0;

PARAM_STRUCT *paramList = NULL;
int paramListSize = 0;

CdiLineInfoBlk *busHandle[NUMBER_LINES];
int CdiNumIndiciesPerLine[NUMBER_LINES];
int numConnectedHandles = 0;
int cdiInitFlag = 0; /* calling cdiInitialize() */

cdiRedirXRefStruct **cdiRedirXref = NULL;
int nCdiRedirectedDevices = 0;

CdiBusInfoBlk busList[NUMBER_LINES];
int numBusList = 0;

#define COMPLEX_RANGE_TOKEN " - "
#define SIMPLE_RANGE_TOKEN "-"
/*#define SEPERATOR_TOKEN ",;+" */
#define SEPERATOR_TOKEN ",;"

double applyClbr(CdiRule *rule,BYTE *data,int dataFormat);
void assignClbr(double dval,BYTE *data,int dataFormat);

CDI_EXPORT CdiLineInfoBlk *cdiGetBusHandle(int cdiLine) 
{
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return NULL;
  return busHandle[cdiLine];
}
CDI_EXPORT int cdi_GetNumberConnectedHandles(void) { return numConnectedHandles; }
CDI_EXPORT int cdi_GetNumberRegisteredBus(void) { return numBusList; }
CDI_EXPORT int cdi_GetNumberRequests(int cdiLine)
{ int i, cnt;
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
  for (i=0, cnt=0; i<busHandle[cdiLine]->numberActiveRequest; i++) if ( busHandle[cdiLine]->pReqList[i].active ) cnt++;
  return cnt;
}
int cdiGetNextAvailableBusType(void)
{ /* find the next free (unnamed) slot */
  int i;
  for (i=0; i<NUMBER_LINES && strlen(busList[i].busName) > 0; i++);
  if (i == NUMBER_LINES) 
  {
    cdilog("cdi bus identifiers exhausted!");
    return -resources_exhausted;
  }
  return i;
}
int cdiGetBusType(char *busName)
{ 
  int i;
  for (i=0; i<numBusList && strncmp(busName, busList[i].busName,16); i++);
  if (i == numBusList) return -1;
  return busList[i].busType;
}
CDI_EXPORT int cdi_RegisterBus(
  char *busName,
  int busType,
  int (*filterBusFcn)(int *parameter, int accessFlag),
  int (*cfgBusFcn)(void),
  int (*initBusFcn)(int busline,int cdiline,int ndevices,char *paramList),
  void (*execBusFcn)(CdiRequestInfoBlk *pReq),
  int (*freeBusFcn)(int busline)
  )
{
  int i;
  for (i=0; i<numBusList && busType != busList[i].busType && strncmp(busName, busList[i].busName,16); i++);
  if (i == numBusList) /* it's new !*/
  {
    if (i == NUMBER_LINES) 
    {
      cdilog("cdi bus %.16s : no more identifiers !",busName);
      return -resources_exhausted;
    }
    numBusList++;
    cdilog("cdi bus %s (id %d) registered",busName,busType);
  }
  else 
  {
    cdilog("cdi bus %s (id %d) : already registered",busName,busType);
    return -invalid_index; /*  same type index! */
  }
  strncpy(busList[i].busName,busName,16);
  busList[i].busType = busType;
  busList[i].filterBusFcn = filterBusFcn;
  busList[i].cfgBusFcn = cfgBusFcn;
  busList[i].initBusFcn = initBusFcn;
  busList[i].execBusFcn = execBusFcn;
  busList[i].freeBusFcn = freeBusFcn;
  return i;
}

/************************************/
CDI_EXPORT int getTimeMilliseconds(void)
{ 
  struct timeval tv; 
  gettimeofday (&tv, NULL);
  return ((tv.tv_sec - gCdiStartupTime) * 1000 + tv.tv_usec / 1000);
}

/********************************************/
CDI_EXPORT int getTypeLength(short dataType)
{
  if ( dataType == cdiChar ) return 1;
  else if ( dataType == cdiByte ) return 1;
  else if ( dataType == cdiShort ) return 2;
  else if ( dataType == cdiInteger ) return 4;
  else if ( dataType == cdiFloat ) return 4;
  else if ( dataType == cdiDouble ) return 8;
  else if ( dataType == cdiChar32 ) return 32;
  return 0;
}
/********************************************/
/* compare pattern or last data             */
/********************************************/
CDI_EXPORT int cdiDataCmp(CdiRequestInfoBlk *pReq, char *lastDataOrPattern, char *str1, int length1, int stringType)
{ int cmpLength, i, j;
  int delta, lastData, window, *iTmp;
  double fdelta, flastData, fwindow;
  char *plast, *pnow, last[8], now[8];		/* for rs232 string type data comparison, convert to number first!!! */
  short *sTmp;
  BYTE *bTmp;
  float *fTmp;
  double *dTmp;

  cmpLength = length1 < (int)pReq->inLength ? length1 : (int)pReq->inLength;
  if ( stringType == 0 )
  {
    plast = lastDataOrPattern;
    pnow = str1;
  }
  else		/* convert string "1234" to number */
  {
    plast = last;
    pnow = now;
    if ( pReq->inType == cdiByte )
    {
      sscanf(lastDataOrPattern, "%hc", (char *)last ); 
      sscanf(str1, "%hc", (char *)now ); 
    }
    else if ( pReq->inType == cdiShort )
    {
      sscanf(lastDataOrPattern, "%hd", (short *)last ); 
      sscanf(str1, "%hd", (short *)now ); 
    }
    else if ( pReq->inType == cdiInteger )
    {
      sscanf(lastDataOrPattern, "%d", (int *)last ); 
      sscanf(str1, "%d", (int *)now ); 
    }
    else if ( pReq->inType == cdiFloat )
    {
      sscanf(lastDataOrPattern, "%f", (float *)last ); 
      sscanf(str1, "%f", (float *)now ); 
    }
    else
    {
      sscanf(lastDataOrPattern, "%f", (double *)last ); 
      sscanf(str1, "%f", (double *)now ); 
    }
  }
  if ( pReq->conditionFlag == maskChangeWindow || pReq->conditionFlag == maskChangePercent )
  {
    if ( pReq->inType == cdiChar || pReq->inType == cdiChar32 ) return 0;    /* no tolerance for string data */
    else if ( pReq->inType == cdiByte || pReq->inType == cdiShort || pReq->inType == cdiInteger )
    {
      if ( pReq->inType == cdiByte )
      {
        window = *(BYTE *)pReq->patternFilter;
        lastData = *(BYTE *)plast;
        delta = *(BYTE *)plast - *(BYTE *)pnow;
	    }
	    else if ( pReq->inType == cdiShort )
	    {
        window = *(short *)pReq->patternFilter;
        lastData = *(short *)plast;
        delta = *(short *)plast - *(short *)pnow;
      }
      else
      {
        window = *(int *)pReq->patternFilter;
        lastData = *(int *)plast;
        delta = *(int *)plast - *(int *)pnow;
      }
      if ( delta < 0 ) delta = -delta;
      if ( pReq->conditionFlag == maskChangeWindow )
	      return delta >= window ? 1 : 0;
      else
	      return delta * 100 >= window * lastData ? 1 : 0;			  
    }
    else	/* no mask for float and double */
    {
      if ( pReq->inType == cdiFloat )
      {
        fwindow = *(float *)pReq->patternFilter;
        flastData = *(float *)plast;
        fdelta = *(float *)plast - *(float *)pnow;
      }
      else
      {
        fwindow = *(double *)pReq->patternFilter;
        flastData = *(double *)plast;
        fdelta = *(double *)plast - *(double *)pnow;
      }
      if ( fdelta < 0 ) fdelta = -fdelta;
      if ( pReq->conditionFlag == maskChangeWindow )
	      return fdelta > fwindow ? 1 : 0;
      else
	      return fdelta * 100 > fwindow * flastData ? 1 : 0;			  
    }
  }
  else  /* compare pattern */
  {
    if ( pReq->inType == cdiChar || pReq->inType == cdiChar32 )  /* compare string */
    {
      cmpLength = length1 < (int)strlen(pReq->patternFilter) ? length1 : strlen(pReq->patternFilter);
      for ( i=0; i<cmpLength; i++ ) if ( str1[i] != ((char *)pReq->patternFilter)[i] ) break;
      return strncmp(str1, (char *)pReq->patternFilter, cmpLength);
    }
    else    /* compare number */
    {
      if ( pReq->inType == cdiByte )
      {
        bTmp = (BYTE *)str1;
        for (i=0; i<length1; i+=1 )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && i < pReq->pDev->dataPatternStartIndex && i > pReq->pDev->dataPatternStopIndex ) continue;
          if ( bTmp[i] != *(BYTE *)pReq->patternFilter ) break;
        }
      }
      else if ( pReq->inType == cdiShort )
      {
        sTmp = (short *)str1;
        for (i=j=0; j<length1; j+=2, i++ )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && (i < pReq->pDev->dataPatternStartIndex || i > pReq->pDev->dataPatternStopIndex) ) continue;
          if ( sTmp[i] != *(short *)pReq->patternFilter ) break;
        }
      }
      else if ( pReq->inType == cdiInteger )
      {
        iTmp = (int *)str1;
        for (i=j=0; j<length1; j+=4, i++ )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && (i < pReq->pDev->dataPatternStartIndex || i > pReq->pDev->dataPatternStopIndex) ) continue;
          if ( iTmp[i] != *(int *)pReq->patternFilter ) break;
        }
      }
      else if ( pReq->inType == cdiFloat )
      {
        fTmp = (float *)str1;
        for (i=j=0; j<length1; j+=4, i++ )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && (i < pReq->pDev->dataPatternStartIndex || i > pReq->pDev->dataPatternStopIndex) ) continue;
          if ( (int) fTmp[i] != *(int *)pReq->patternFilter ) break;
        }
      }
      else if ( pReq->inType == cdiDouble )
      {
        dTmp = (double *)str1;
        for (i=j=0; j<length1; j+=8, i++ )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && (i < pReq->pDev->dataPatternStartIndex || i > pReq->pDev->dataPatternStopIndex) ) continue;
          if ( (int) dTmp[i] != *(int *)pReq->patternFilter ) break;
        }
      }
      if ( (i <= pReq->pDev->dataPatternStartIndex || i <= pReq->pDev->dataPatternStopIndex) && pReq->conditionFlag == maskCompNotEqual ) return 1; /* can not compare then return notEqual */
      return j < length1 ? 1 : 0;
    }
#ifdef DONOTDELEAT
    for ( i=0; i<cmpLength; i++ ) 
	  {
      if ( pReq->hasDataMask )
      {
        if ( (str1[i] & pReq->mask[i]) != pReq->patternFilter[i] ) break; /* is mask guaranteed to be equal in length ? */
      }
      else if ( str1[i] != ((char *)pReq->patternFilter)[i] ) break;
 	  }
#endif
  }
  return 0;
}

/*******************************************************/
/* called from busplug after data comes from Hardware, set return conditionflag */
/*******************************************************/
CDI_EXPORT void checkReturnCondition(CdiRequestInfoBlk *pReq, char *lastData, char *str, int datalen, int stringType)
{ short *sTmp;
  BYTE *bTmp;
  int i, j, comFlag, *iTmp;
  float fp;

  if (pReq == NULL) return;

  /* check return data format */
  if ( stringType != 0 )
  {
    if ( pReq->inType >= cdiFloat && sscanf(str,"%f",&fp) != 1 ) return;
    else if ( pReq->inType >= cdiByte && sscanf(str,"%d",&i) != 1 ) return;
  }
  /* should use default data type not input(stringType) ??????? */
  if ( stringType == 0 && pReq->hasDataMask ) /* mask for byte, short or integer type data */
  {
    if ( pReq->inType == cdiByte )
    {
      bTmp = (BYTE *)str;
      for (i=0; i<datalen; i+=1 )
      {
        if ( pReq->pDev->dataMaskStartIndex >= 0 && i < pReq->pDev->dataMaskStartIndex && i > pReq->pDev->dataMaskStopIndex ) continue;
        (*bTmp) &= (BYTE)pReq->maskValue;
        bTmp++;
      }
    }
    else if ( pReq->inType == cdiShort )
    {
      sTmp = (short *)str;
      for (i=j=0; j<datalen; j+=2, i++ )
      {
        if ( pReq->pDev->dataMaskStartIndex >= 0 && i < pReq->pDev->dataMaskStartIndex && i > pReq->pDev->dataMaskStopIndex ) continue;
        (*sTmp) &= (short)pReq->maskValue;
        sTmp++;
      }
    }
    else if ( pReq->inType == cdiInteger )
    {
      iTmp = (int *)str;
      for (i=j=0; j<datalen; j+=4, i++ )
      {
        if ( pReq->pDev->dataMaskStartIndex >= 0 && i < pReq->pDev->dataMaskStartIndex && i > pReq->pDev->dataMaskStopIndex ) continue;
        (*iTmp) &= (int)pReq->maskValue;
        iTmp++;
      }
    }
  }

  if ( pReq->conditionFlag == anyData )
  {
    if ( stringType == 0 && (pReq->accessFlag == cdiReadWriteFlag || pReq->accessFlag == cdiWriteReadWriteFlag)) /* read then write bits */
    { /* mask the readData to the write Buffer */
      i = pReq->accessFlag == cdiReadWriteFlag ? 0 : 1;
      if ( pReq->inType == cdiByte )
      {
        if ( ((BYTE *)pReq->pOutData)[i+1] )
          ((BYTE *)pReq->pOutData)[i] |= *(BYTE *)str;
        else
          ((BYTE *)pReq->pOutData)[i] = (~((BYTE *)pReq->pOutData)[i]) & *(BYTE *)str;
      }
      else if ( pReq->inType == cdiShort )
      {
        if ( ((short *)pReq->pOutData)[i+1] )
          ((short *)pReq->pOutData)[i] |= *(short *)str;
        else
          ((short *)pReq->pOutData)[i] = (~((short *)pReq->pOutData)[i]) & *(short *)str;
      }
      else if ( pReq->inType == cdiInteger )
      {
        if ( ((int *)pReq->pOutData)[i+1] )
          ((int *)pReq->pOutData)[i] |= *(int *)str;
        else
          ((int *)pReq->pOutData)[i] = (~((int *)pReq->pOutData)[i]) & *(int *)str;
      }
    }
    pReq->dataReturnCondition = DATA_PASS_FILTER;
  }
  else if ( pReq->conditionFlag == maskChangeWindow || pReq->conditionFlag == maskChangePercent )
  {
    if ( pReq->requestCounter == 1 ) 
      pReq->dataReturnCondition = DATA_PASS_FILTER; /* first call of Async or sync */
    else 
    {
      /*if ( pReq->patternFilter == NULL ) */
      if ( pReq->hasPatternFilter == FALSE ) 
      {
        pReq->dataReturnCondition = DATA_PASS_FILTER; /* no filter => any data */
      }
      else /* data changed? compare to last one */
      {
        comFlag = cdiDataCmp(pReq,lastData,str,datalen, stringType);
        if ( comFlag != 0 ) pReq->dataReturnCondition = DATA_PASS_FILTER;
        else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK; /* data back for polling request! */
      }
    }
  }
  else
  {
    comFlag = cdiDataCmp(pReq,pReq->patternFilter,str,datalen, stringType);
	  if (cdiDebugLevel) dbglog("checkReturnCondition: %d-%d: data: %d --- %d(%d); datalen: %d; stringtype: %d; cmp: %d\n",pReq->conditionFlag, pReq->dataReturnCondition, *(short *)lastData, *(short *)str, *(short *)pReq->patternFilter, datalen, stringType, comFlag);
    if ( comFlag == 0 )
    {
      if ( pReq->conditionFlag == maskCompEqual ) pReq->dataReturnCondition = DATA_PASS_FILTER;
      else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK;
    }
    else if ( pReq->conditionFlag == maskCompNotEqual ) pReq->dataReturnCondition = DATA_PASS_FILTER;
    else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK;
	  if (cdiDebugLevel) dbglog("checkReturnCondition: %d-%d: data: %d --- %d(%d); datalen: %d; stringtype: %d; cmp: %d\n",pReq->conditionFlag, pReq->dataReturnCondition, *(short *)lastData, *(short *)str, *(short *)pReq->patternFilter, datalen, stringType, comFlag);
  }
}

/************************************/
/* check if need to transfer */
/************************************/
CDI_EXPORT int checkRequestTime(CdiRequestInfoBlk *pReq)
{
  if ( pReq->active == transRequest )
  {
    if ( pReq->callback == NULL || pReq->timeout == 0 ) return 1; /* sync or no timeout */
    if ( pReq->responseTime < pReq->requestTime ) return 1; /* first call for asyn */
    if ( pReq->timeout > 0 )
    {
      if ( getTimeMilliseconds() > pReq->requestTime + pReq->timeout ) return 1; /* poll interval */
    }
    else
    {
      if ( getTimeMilliseconds() > pReq->requestTime - pReq->timeout ) return 1; /* poll interval */
    }
    return 0;
  }
  else if ( pReq->active == transRequest2nd ) return 1;
  return 0;
}

/************************************/
CDI_EXPORT int signGroupInOutFormatCode(CdiRequestInfoBlk *pReq, int inFormat, int outFormat)
{ int i, j, cnt;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL )
  {
    pReq->outFormatCode = inFormat;
    pReq->inFormatCode = outFormat;
    return 1;
  }
  for ( j = 0, cnt = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      pLine->pReqList[i].outFormatCode = inFormat;
      pLine->pReqList[i].inFormatCode = outFormat;
      cnt++;
    }
  }
  return cnt;
}

/************************************/
CDI_EXPORT int checkGroupReturnFlags(CdiRequestInfoBlk *pReq)
{ int i, j, cnt;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL ) return 0;
  pReq->NgroupRets = 1;
  cnt = 0;
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      if ( pLine->pReqList[i].NgroupRets ) cnt++;
    }
  }
  return ( ((CdiRequestInfoBlk *)pReq->groupReq)->asynCallFlag == 0 && ((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs == cnt ) ? 1 : 0;
}

/************************************/
/* return value: 0 for single request; number of erros for group request  */
CDI_EXPORT int setGroupReturnFlags(CdiRequestInfoBlk *pReq, int index)
{ int i, j;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL ) return 0;
  pReq->NgrpErrors = 0;
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      if ( index & 0x1 ) pLine->pReqList[i].userCallDown++;
      if ( index & 0x2 ) pLine->pReqList[i].groupSyncStart = 1; /* mark next start  */
      if ( index & 0x4 ) pLine->pReqList[i].NgroupRets = 0;     /* reset return flags  */
      if ( pLine->pReqList[i].errorCode ) pReq->NgrpErrors++;
    }
  }
  return pReq->NgrpErrors;
}

/************************************/
/* exec request, step by step */
/* all 4 steps arranged here! */
/************************************/
CDI_EXPORT void cdiExecRequest(CdiRequestInfoBlk *pReq, int taskIndex)
{
  if (cdiDebugLevel) dbglog("Call cdiExecRequest: taskIndex === %d; activeFlag: %d; cancelFlag: %d; requestNumber: 0x%x\n",taskIndex, pReq->active, pReq->cancelFlag, pReq->requestNumber);

  if ( pReq->active <= idleState || pReq->active >= finishRequest ) return;

  /* send out request */
  if ( pReq->active == transRequest ) /* init transmission parameters only on first transfer */
  {
    pReq->groupSyncStart = 0;
    pReq->requestTime = getTimeMilliseconds();
    pReq->dataReturnCondition = 0;
    pReq->errorCode = 0;
    if (pReq->perror != NULL) *pReq->perror = 0;
    pReq->requestCounter++;
    if (cdiDebugLevel > 3) dbglog("set request Time: %d; preq: %x\n",pReq->requestTime, pReq);
  }
  if ( pReq->active == recvRequest2nd ) pReq->active = finishRequest; /* second receive */
  else if ( pReq->active == recvRequest ) /* first receive => change to finish or 2ndSend */
  {
    /* data came back! */
    if (pReq->errorCode || pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag) pReq->active = finishRequest;
    else if ( pReq->dataReturnCondition != DATA_BACK ) pReq->active = transRequest2nd;
    else pReq->active = finishRequest; /* data do not pass condition, then finish! */
  }

  if ( pReq->active != finishRequest ) /* only first or second request */
  {
    if ( gSimulationFlag ) 
    {
      pReq->active = finishRequest;
    }
    else
    {
      if (busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].execBusFcn != NULL)
        busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].execBusFcn(pReq);
      if ( pReq->active == finishRequest && pReq->errorCode == 0 && pReq->dataReturnCondition == 0 )
        pReq->dataReturnCondition = DATA_PASS_FILTER; /* flag for callBack */
    }
  }
  if ( pReq->active != finishRequest ) return;

  /* finish request!! */
  pReq->responseTime = getTimeMilliseconds(); /* avoid timeout check!!!! */
  if (cdiDebugLevel > 3) dbglog("set response time: %d, preq: %x\n",pReq->responseTime, pReq);
  if ( pReq->callback == NULL )
  {
    if ( pReq->groupReq == NULL ) cdiPostSemaphore(&pReq->sync_SemId);
    else if ( checkGroupReturnFlags(pReq) ) cdiPostSemaphore(&((CdiRequestInfoBlk *)pReq->groupReq)->sync_SemId);
  }
  else
  {
    if ( (pReq->timeout >= 0 && pReq->dataReturnCondition == DATA_PASS_FILTER) || pReq->errorCode )
    { 
      if ( pReq->groupReq == NULL || checkGroupReturnFlags(pReq) )
      {
        if (cdiDebugLevel) dbglog("post callback semaphore: task: 0x%x\n",pReq);
        if ( pReq->groupReq == NULL ) pReq->asynCallFlag = 1;
        else ((CdiRequestInfoBlk *)pReq->groupReq)->asynCallFlag = 1;
        setGroupReturnFlags(pReq, 0x2);
        if ( pReq->groupReq == NULL ) cdiPostSemaphore(&(((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->call_SemId));
        else cdiPostSemaphore(&(((CdiLineInfoBlk *)((CdiRequestInfoBlk *)pReq->groupReq)->pDev->pcdiLine)->call_SemId));
      }
    }
    if ( pReq->pollFlag && pReq->cancelFlag == 0 )
    {
      pReq->active = transRequest; /* start next time interval */
      /* Let background task start, avoid continuous recursive call */
      /*      if ( pReq->timeout == 0 ) cdiExecRequest(pReq, taskIndex); */
    }
  }
  return;
}

/**************************************/
CDI_EXPORT void freeRequestBlock(CdiLineInfoBlk *pLine)
{ 
  int i, findActive;

  findActive = 0;
  for ( i = pLine->numberActiveRequest - 1; i >= 0 ; i-- )
  {
    if ( pLine->pReqList[i].active == removeRequest ) 
    {
      pLine->pReqList[i].active = 0;
      if (pLine->pReqList[i].freeOutDataMemoryOnCancel) 
      { 
        free(pLine->pReqList[i].pOutData); pLine->pReqList[i].pOutData = NULL; 
        pLine->pReqList[i].freeOutDataMemoryOnCancel = FALSE;
      }
      if (pLine->pReqList[i].freeInDataMemoryOnCancel) 
      { 
        free(pLine->pReqList[i].pInData); pLine->pReqList[i].pInData = NULL; 
        pLine->pReqList[i].freeInDataMemoryOnCancel = FALSE;
      }
    }
    if ( pLine->pReqList[i].active ) findActive = 1;
    if ( pLine->pReqList[i].active == 0 && findActive == 0 )	/* adjust max number */
    {
      pLine->numberActiveRequest = i;
      if (cdiDebugLevel) dbglog("set number active request: %d\n", pLine->numberActiveRequest);
    }
    if ( pLine->pReqList[i].active != finishRequest || pLine->pReqList[i].userCallDown == 0 ) continue;

  	if ( pLine->pReqList[i].cancelFlag == 2 )
	    pLine->pReqList[i].active = removeRequest;
    else
	    pLine->pReqList[i].active = pLine->pReqList[i].lockRequestFlag ? idleState : removeRequest;
	  
    if ( pLine->pReqList[i].active == removeRequest ) pLine->pReqList[i].groupReq = NULL;
  }
}

/**************************************/
CDI_EXPORT THREAD_TYPE ThreadCallback(THREAD_PARAM param)
{ 
  CdiLineInfoBlk *pLine = (CdiLineInfoBlk *)param, *pL;
  int j, k, cc, applyRules = 0, hasErrLst = 0, hasNameLst = 0, doutlen;
  DTYPE dout;
  int *errLst = NULL;
  NAME16 *nameLst = NULL, *nameLstPtr;

  cdilog("ThreadCallback starting: cdi line: %d, hardware line number: %d", pLine->cdiLine, pLine->busLine);
  for(;;)
  {
    if (cdiWaitSemaphore(&pLine->call_SemId,0))
    {
      cdilog("ThreadCallback : semaphore ERROR! line: %d, busLine: %d", pLine->cdiLine, pLine->busLine);
      continue;
    }
    for ( pLine->callCounter = 0; pLine->callCounter < pLine->numberActiveRequest; pLine->callCounter++ )
    {
      if ( pLine->pReqList[pLine->callCounter].asynCallFlag == 0 ) continue;
      /* do any calibration /reformatting here */
      if (pLine->pReqList[pLine->callCounter].needsCalibration)
      {
        applyRules = pLine->pReqList[pLine->callCounter].needsCalibration & RETURN_CALIBRATED;
        hasErrLst = pLine->pReqList[pLine->callCounter].needsCalibration & RETURN_STATUS;
        hasNameLst = pLine->pReqList[pLine->callCounter].needsCalibration & RETURN_NAMES;
        if (hasNameLst)
        { /* point to the right spot in memory to put the names ... */
          doutlen = pLine->pReqList[pLine->callCounter].NgroupsReqs;
          if (nameLst == NULL)
            nameLst = (NAME16 *)((int *)pLine->pReqList[pLine->callCounter].pInData + doutlen*2);
        }
        if (pLine->pReqList[pLine->callCounter].groupReq == NULL) 
        {
          cc = clbrRequest(&pLine->pReqList[pLine->callCounter],applyRules);
          if (nameLst != NULL)
          {
            strncpy(nameLst->name,pLine->pReqList[pLine->callCounter].pDev->devName,16);
          }
        }
        else
        {
          nameLstPtr = nameLst;
          for (j=0; j<numConnectedHandles; j++)	/* search all devices belonging to this group */
          {
            if (busHandle[j] == NULL ) continue;
            pL = busHandle[j];
            for (k=0; k<pL->numberActiveRequest; k++)
            {
              if (pL->pReqList[k].groupReq != pLine->pReqList[pLine->callCounter].groupReq) continue;
              if ((cc=clbrRequest(&pL->pReqList[k],applyRules)) < 0) break;
              if (nameLstPtr != NULL)
              {
                strncpy(nameLstPtr->name,pL->pDevList[k].devName,16);
                nameLstPtr++;
              }
            }
            if (cc < 0) break;
          }
        }
        if (pLine->pReqList[pLine->callCounter].errorCode != 0) pLine->pReqList[pLine->callCounter].errorCode = -cc;
      }
      if (hasErrLst)
      {
        dout.dArrayLength = pLine->pReqList[pLine->callCounter].NgroupsReqs;
        dout.dFormat = pLine->pReqList[pLine->callCounter].inFormatCode;
        dout.data.vptr = pLine->pReqList[pLine->callCounter].pInData;
        errLst = pLine->pReqList[pLine->callCounter].perror;
        cdiRemapDataSets(&dout,dout.data.bptr,errLst,nameLst);
      }
      pLine->pReqList[pLine->callCounter].callback( pLine->pReqList[pLine->callCounter].callBackIndex, pLine->pReqList[pLine->callCounter].groupReq ? -pLine->pReqList[pLine->callCounter].NgrpErrors :  -pLine->pReqList[pLine->callCounter].errorCode);
      setGroupReturnFlags(&(pLine->pReqList[pLine->callCounter]), 0x4);
      pLine->pReqList[pLine->callCounter].asynCallFlag = 0;
      if ( pLine->pReqList[pLine->callCounter].pollFlag == 0 )
      {
        pLine->pReqList[pLine->callCounter].userCallDown++;
        setGroupReturnFlags(&(pLine->pReqList[pLine->callCounter]), 0x1);
      }
    }
  }
}


/**************************************/
CDI_EXPORT THREAD_TYPE ThreadMaster(THREAD_PARAM param)
{ 
  CdiLineInfoBlk *pLine = (CdiLineInfoBlk *)param;

  cdilog("ThreadMaster starting: cdi line: %d, hardware bus line number: %d", pLine->cdiLine, pLine->busLine);
  for(;;)
  {
    pLine->threadBusy = FALSE;
    if ( pLine->activeCounter )
    {
      pLine->backReq.tv_sec = 0;
      pLine->backReq.tv_nsec = BACK_THREAD_TIME * 1000;
      nanosleep( &pLine->backReq, NULL );
    }
    else if (cdiWaitSemaphore(&pLine->back_SemId,0))
    {
      cdilog("ThreadMaster : semaphore ERROR! cdi line: %d, busLine: %d", pLine->cdiLine, pLine->busLine);
      continue;
    }

    pLine->threadBusy = TRUE;
    pLine->activeCounter = 0;
    gettimeofday (&pLine->tv, NULL);
    pLine->backTime = (pLine->tv.tv_sec - gCdiStartupTime) * 1000 + pLine->tv.tv_usec / 1000;
    
    for ( pLine->backCounter = 0; pLine->threadActive && pLine->backCounter < pLine->numberActiveRequest; pLine->backCounter++ )
    {
      if ( pLine->pReqList[pLine->backCounter].active <= idleState ) continue;
      pLine->activeCounter++;
      /* check group call, callback not done yet ==> continue */
		  if ( pLine->pReqList[pLine->backCounter].groupReq )
      {
        if ( ((CdiRequestInfoBlk *)pLine->pReqList[pLine->backCounter].groupReq)->asynCallFlag ) continue;
      } 
      else 
      {
        if ( pLine->pReqList[pLine->backCounter].asynCallFlag ) continue;
      }

      /* check timeout */
      if ( pLine->pReqList[pLine->backCounter].active < finishRequest && pLine->pReqList[pLine->backCounter].responseTime < pLine->pReqList[pLine->backCounter].requestTime &&
           ((pLine->pReqList[pLine->backCounter].timeout > 0 && (pLine->backTime > pLine->pReqList[pLine->backCounter].requestTime + pLine->pReqList[pLine->backCounter].timeout)) ||
           (pLine->pReqList[pLine->backCounter].timeout < 0 && (pLine->backTime > pLine->pReqList[pLine->backCounter].requestTime - pLine->pReqList[pLine->backCounter].timeout)) ) )
      {
        cdilog("timeout: %d; response %d, request %d, background %d",
          pLine->pReqList[pLine->backCounter].timeout, 
          pLine->pReqList[pLine->backCounter].responseTime, 
          pLine->pReqList[pLine->backCounter].requestTime, 
          pLine->backTime);

        pLine->pReqList[pLine->backCounter].errorCode = operation_timeout;
        if ( pLine->pReqList[pLine->backCounter].perror ) *(pLine->pReqList[pLine->backCounter].perror) = operation_timeout;
        pLine->pReqList[pLine->backCounter].active = finishRequest; /* finish */
        pLine->pReqList[pLine->backCounter].responseTime = pLine->backTime;
        /* treat sync and async differently */
        if ( pLine->pReqList[pLine->backCounter].callback == NULL )
    		{
		      if ( pLine->pReqList[pLine->backCounter].groupReq == NULL ) cdiPostSemaphore(&pLine->pReqList[pLine->backCounter].sync_SemId);
		      else if ( checkGroupReturnFlags(&pLine->pReqList[pLine->backCounter]) ) cdiPostSemaphore(&((CdiRequestInfoBlk *)pLine->pReqList[pLine->backCounter].groupReq)->sync_SemId);
      	}
        else
        {
		      if ( pLine->pReqList[pLine->backCounter].groupReq == NULL || checkGroupReturnFlags(&pLine->pReqList[pLine->backCounter]) )
    		  {
		      	if ( pLine->pReqList[pLine->backCounter].groupReq == NULL ) pLine->pReqList[pLine->backCounter].asynCallFlag = 1;
            else ((CdiRequestInfoBlk *)pLine->pReqList[pLine->backCounter].groupReq)->asynCallFlag = 1;
            setGroupReturnFlags(&(pLine->pReqList[pLine->backCounter]), 0x2);
			      if ( pLine->pReqList[pLine->backCounter].groupReq == NULL ) cdiPostSemaphore(&pLine->call_SemId);
            else cdiPostSemaphore(&(((CdiLineInfoBlk *)((CdiRequestInfoBlk *)pLine->pReqList[pLine->backCounter].groupReq)->pDev->pcdiLine)->call_SemId));
		      }
          if ( pLine->pReqList[pLine->backCounter].pollFlag  && pLine->pReqList[pLine->backCounter].cancelFlag == 0 ) pLine->pReqList[pLine->backCounter].active = transRequest; /* start next time interval */
        }
        continue;
      }
      if ( pLine->pReqList[pLine->backCounter].active != transRequest ) continue;
      if ( pLine->pReqList[pLine->backCounter].callback == NULL &&   /* start the rest of sync group call */
           pLine->pReqList[pLine->backCounter].active == transRequest &&
           (pLine->pReqList[pLine->backCounter].groupReq == NULL || pLine->pReqList[pLine->backCounter].groupReq == &pLine->pReqList[pLine->backCounter]) 
         ) continue;

      if ( checkRequestTime( &pLine->pReqList[pLine->backCounter] ) )
		    if ( pLine->pReqList[pLine->backCounter].groupReq == NULL || pLine->pReqList[pLine->backCounter].groupSyncStart )
          cdiExecRequest(&(pLine->pReqList[pLine->backCounter]), backTaskIndex);
    }
    if ( pLine->activeCounter ) freeRequestBlock(pLine);
  }
}

/**************************************/
CDI_EXPORT int disconnectFromBus ( int busType, int busLine, int fileHandle )
{
  int i;
  for (i=0; i<numBusList; i++) if (busList[i].busType == busType) 
  {
    if (busList[i].freeBusFcn != NULL) 
    {
      if ( gSimulationFlag ) return 0;
      return busList[i].freeBusFcn(busLine);
    }
  }
  return -not_implemented;
}
/**************************************/
CDI_EXPORT int connectToBus ( int busType, int busLine, char *parameter, int cdiLine, int numberDevices)
{
  int i;
  for (i=0; i<numBusList; i++) if (busList[i].busType == busType) 
  {
    if (busList[i].initBusFcn != NULL) 
    {
      if ( gSimulationFlag ) return 0;
      return busList[i].initBusFcn(busLine,cdiLine,numberDevices,parameter);
    }
    if (parameter != NULL) 
      cdilog("bus parameters for %.16s with no initialization handler",busList[i].busName);
  }
  return -1;
}

/**************************************/
CDI_EXPORT int checkLine(int busType, int busLine)
{ 
  int i;

  for ( i = 0; i < numConnectedHandles; i++ )
    if ( busHandle[i] != NULL && busList [ busHandle[i]->busIndex ].busType == busType && busHandle[i]->busLine == busLine ) break;
  return i < numConnectedHandles ? i : -1;
}

/**************************************/
CDI_EXPORT int findFreeLine()
{ 
  int i;
  if ( numConnectedHandles == 0 ) return 0;
  for ( i = 0; i < numConnectedHandles; i++ ) if ( busHandle[i] == NULL ) return i;
  if ( numConnectedHandles >= NUMBER_LINES ) return -1; /* slots full! */
  return numConnectedHandles;
}

/************************************/
/* load bus specific information */
/************************************/
time_t gCdiStartupTime = 0;
CDI_EXPORT int cdiInitialize(void)
{ 
  gCdiStartupTime = time(NULL);
  cdiSystemInitialization();
  cdiInitFlag = 1;
  return 0;
}

/**************************************/
CDI_EXPORT int getFreeRequestBlock(CdiLineInfoBlk *pLine)
{ 
  int i;

  if ( pLine->numberActiveRequest == 0 ) return 0;
  /* find if there is a Loch */
  for ( i = 0; i < pLine->numberActiveRequest; i++ ) 
  {
    if ( pLine->pReqList[i].active == 0 ) return i;
  }
  if ( pLine->numberActiveRequest >= pLine->numberRequest ) return -link_exhausted; /* slots full! */
  return pLine->numberActiveRequest;
}

/**************************************/
CDI_EXPORT void makeCdiHashTable(int cdiLine, int ndevices)
{
  int cc=0,n=ndevices;

  if (numConnectedHandles == 0 || busHandle == NULL) return;
  if (cdiLine < 0 || cdiLine > numConnectedHandles || busHandle[cdiLine] == NULL) return;
  if (busHandle[cdiLine]->devHashList != NULL) return; /* already have hash table */
  if (n < 1) n = busHandle[cdiLine]->numberDevices;/* get total number of devices */
  if (n < CDI_PAYOFF_HASHSIZE) return;
  if (n < CDI_SMALL_HASHSIZE) busHandle[cdiLine]->devHashSize = CDI_SMALL_HASHSIZE;
  else if (n < CDI_MEDIUM_HASHSIZE) busHandle[cdiLine]->devHashSize = CDI_MEDIUM_HASHSIZE;
//  else if (n < CDI_LARGE_HASHSIZE) busHandle[cdiLine]->devHashSize = CDI_LARGE_HASHSIZE;
  else busHandle[cdiLine]->devHashSize = CDI_LARGE_HASHSIZE;
  if ((busHandle[cdiLine]->devHashList=(cdiHashXRefStruct **)calloc(busHandle[cdiLine]->devHashSize,sizeof(cdiHashXRefStruct *))) == NULL) 
    ccerr(out_of_server_memory);
err:
  if (cc) cdilog("makeCdiHashTable: %s",erlst[cc]);
  return;
}
CDI_EXPORT int findDevice(char *devName, int *pCdiLine, int *pDevIndex)
{ 
  int i, j, idx, devnr = -1, cc = 0;
  cdiHashXRefStruct *xref;
  char scratch[128],*c=NULL;

  if (devName == NULL || pCdiLine == NULL || pDevIndex == NULL) ccerr(argument_list_error);
  if (devName[0] == '#') /* given as a number */
  {
    strncpy(scratch,&devName[1],32); scratch[32] = 0;
    if ((c=strchr(scratch,'.')) != NULL)
    { /* given as #<ln>.<index> */
      *c++ = 0;
      i = atoi(scratch);
      j = atoi(c);
      if (i < 0 || i >= numConnectedHandles) ccerr(illegal_line);
      if (j < 0 || j >= busHandle[i]->numberDevices) ccerr(invalid_index);
      *pCdiLine = i;
      *pDevIndex = j;
      return busHandle[i]->pDevList[j].devNumber;
    }
    /* given as #<nr> */
    devnr = atoi(scratch);
  }
  for (i=0; i<numConnectedHandles; i++) /* search through all bus lines */
  {
    if (busHandle[i] == NULL) continue;
    if (busHandle[i]->devHashSize == 0) /* no hash table ...*/
    {
      for (j=0; j<busHandle[i]->numberDevices; j++)
      {
        if (devnr >= 0)
        {
          if (busHandle[i]->pDevList[j].devNumber != devnr) continue;
        }
        else
        {
          if (busHandle[i]->pDevList[j].devName[0] == 0) continue;
          if (strncmp(devName, busHandle[i]->pDevList[j].devName,16)) continue;
        }
        *pCdiLine = i;
        *pDevIndex = j;
        return busHandle[i]->pDevList[j].devNumber;
      }
    }
    else
    {
      idx = (int)(ElfHash((unsigned char *)devName) % busHandle[i]->devHashSize);
      for (xref=busHandle[i]->devHashList[idx]; xref; xref=xref->nxt)
      {
        if (devnr >= 0)
        {
          if (devnr != xref->devNumber) continue;
          j = xref->cdiIndex;
        }
        else
        {
          j = xref->cdiIndex;
          if (busHandle[i]->pDevList[j].devName[0] == 0) continue;
          if (strncmp(devName, busHandle[i]->pDevList[j].devName,16)) continue;
        }
        *pCdiLine = i;
        *pDevIndex = j;
        return busHandle[i]->pDevList[j].devNumber;
      }
    }
  }
  cc = device_not_connected;
err:
  return -cc;
}
int isAccessAllowed(CdiDeviceInfoBlk *pDev,int accessFlag)
{
  int accessMask = 0x00;
  if (pDev == NULL) return FALSE;
  if (pDev->allowedAccess == 0 || pDev->allowedAccess == 0xffff) return TRUE;
  accessMask |= (0x01 << accessFlag);
  return (pDev->allowedAccess & accessMask);
}

/************************************/
CDI_EXPORT int checkReqWriteData(CdiRequestInfoBlk *pReq, int *paccessFlag, int returnCondition)
{ int flen, xlen, size, retFlag = 0;
  /* use default input data from the database */
  size = getTypeLength((short)pReq->pDev->dataFormat);
  if ( pReq->pDev->dataInput != NULL )
  { /* there is registered input data for this device! do I need to use it? */
    if (*paccessFlag == cdiReadFlag)
    {
      if ( pReq->pOutData == NULL ) /* no input data supplied by the caller, change RD to WRRD */
      {
        if (pReq->pDev->allowedAccess & 0x08)
        {
          if ( pReq->pDev->dataInputSize < size ) return -dimension_error;
          *paccessFlag = cdiWriteReadFlag;
          pReq->pOutData = pReq->pDev->dataInput;
          pReq->outLength = size;
          pReq->outType = pReq->pDev->dataFormat;
          retFlag = 1;
        }
        else if (pReq->pDev->allowedAccess == 0x02) return -illegal_command;
      }
    }
    else if (*paccessFlag == cdiWriteFlag)  /* change WR to WRWR */
    {
      if (pReq->pDev->allowedAccess & 0x10)
      {
        if ( pReq->pDev->dataInputSize < size ) return -dimension_error;
        *paccessFlag = cdiWriteWriteFlag;
        memcpy(pReq->cpOutData, pReq->pDev->dataInput, size);
        flen = pReq->outLength;
        if ( flen > DEF_INPUT_BYTES - size ) return -dimension_error;
        if ( flen > DEF_INPUT_BYTES - size ) flen = DEF_INPUT_BYTES - size;
        memcpy(&pReq->cpOutData[size], pReq->pOutData, flen);
        pReq->pOutData = pReq->cpOutData;
        pReq->outLength = size + flen;
        retFlag = 1;
      }
    }
    else if (*paccessFlag == cdiReadWriteFlag)  /* change RDWR to WRRDWR */
    { /* default firstWrite and SecondWrite  */
      xlen = (pReq->pDev->dataPattern && returnCondition > anyData) ? 1 : 2;
//      if ( pReq->pDev->dataInputSize >= size * xlen && pReq->pDev->allowedAccess & 0x20)
      if (pReq->pDev->dataInputSize >= size && pReq->pOutData && (pReq->pDev->dataInputSize + (int)pReq->outLength) >= size * xlen && pReq->pDev->allowedAccess & 0x20)
      {
        *paccessFlag = cdiWriteReadWriteFlag;
        memcpy(pReq->cpOutData, pReq->pDev->dataInput, pReq->pDev->dataInputSize);
        flen = pReq->outLength;
        if ( flen > DEF_INPUT_BYTES - pReq->pDev->dataInputSize ) return -dimension_error;
        if ( flen > DEF_INPUT_BYTES - pReq->pDev->dataInputSize ) flen = DEF_INPUT_BYTES - pReq->pDev->dataInputSize;
        memcpy(&pReq->cpOutData[pReq->pDev->dataInputSize], pReq->pOutData, flen);
        pReq->pOutData = pReq->cpOutData;
        pReq->outLength = pReq->pDev->dataInputSize + flen;
        retFlag = 1;
      }
    }
  }
  if ( *paccessFlag != cdiReadFlag )
  { /* not a simple read access */
    if ( pReq->pDev->dataFormat >= 0 && pReq->pDev->dataFormat != pReq->outType ) return -illegal_format;

    if ( *paccessFlag == cdiWriteWriteFlag ) flen = 2;
    else if ( *paccessFlag == cdiReadWriteFlag)
    {
      if ( pReq->pDev->dataPattern == NULL || returnCondition == anyData ) flen = 2;
    }
    else if ( *paccessFlag == cdiWriteReadWriteFlag)
    {
      flen = (pReq->pDev->dataPattern == NULL || returnCondition == anyData) ? 3 : 2;
    }
    else flen = 1;

    if ( (int)pReq->outLength < size * flen ) return -dimension_error;

    if ( retFlag == 0 ) /* input from user */
    {
      if ( pReq->pDev->dataInput == NULL && 
           pReq->pDev->dataInputSize && 
           (int)pReq->outLength > pReq->pDev->dataInputSize + (flen-1) * size ) 
        return -dimension_error;
    }
    else if ( pReq->pDev->dataInputSize > 0)  /* use default input data */
    { /* do not check WriteWrite */
      if ( *paccessFlag != cdiWriteWriteFlag && (int)pReq->outLength > size * flen ) return -dimension_error;
    }
  }
  return retFlag;
}
int rvrsClbrRequest(CdiRequestInfoBlk *pReq,int applyRules)
{
  int i, flen, tlen, dlen;
  CdiDeviceInfoBlk *pDev;
  int fromFormat, toFormat;
  BYTE *fromData, *toData;
  CdiRule *rule = NULL;
  double dval;
  /* the request has input calibrated data which need to be converted to hdw bus data */
  if ((fromFormat=pReq->outType) < 0) return -illegal_format;
  flen = getTypeLength((short)fromFormat);
  dlen = pReq->outLength/flen;
  if (flen <= 0) return -illegal_format;
  fromData = pReq->pOutData;
  pDev = pReq->pDev;
  toFormat = pDev->dataFormat;
  if (toFormat < 0) return -illegal_format;
  if ((tlen=getTypeLength((short)toFormat)) <= 0) return -illegal_format;
  if (!applyRules && toFormat == fromFormat) return 0; /* nothing to do */
  /* allocate separate buffer to hold the input hdw bus data */
  toData = (BYTE *)calloc(dlen,getTypeLength((short)pDev->dataFormat));
  if (toData == NULL) return -out_of_server_memory;
  rule = applyRules ? pDev->rvrsRules : NULL;
  for (i=dlen-1; i>=0; i--)
  { /* start from the end and work backwards for in situ assignments */
    dval = applyClbr(rule,fromData+i*flen,fromFormat);
    assignClbr(dval,toData+i*tlen,toFormat);
  }
  pReq->pOutData = toData;
  pReq->outType = toFormat;
  pReq->outLength = tlen;
  pReq->freeOutDataMemoryOnCancel = TRUE;
  return 0;
}

void cdi_reverseCalibrate(CdiRequestInfoBlk *pReq,int cdiNeedsCalibration)
{
  if (cdiNeedsCalibration & INPUT_CALIBRATED)
  {
    rvrsClbrRequest(pReq,TRUE);
  }
  else if (cdiNeedsCalibration & INPUT_REFORMATED)
  {
    rvrsClbrRequest(pReq,TRUE);
  }
  return;
}

/************************************/
/* find and init requestBuffer */
/************************************/
CDI_EXPORT int init_RequestBuffer(char *devName, void *inData, short readType, int inLength, void *outData, short writeType, int outLength, void *dataPatternWindow, 
                       void *dataMask, int *parameter, int accessFlag, int returnCondition, int timeout, int *perror)
{ 
  int i, cc = 0, flen, tlen, cdiLine, cdiIndex, reqIndex, cdiNeedsCalibration = 0;
  CdiRequestInfoBlk *pReq = NULL;

  /* check parameters */
  if (returnCondition & RETURN_CALIBRATED) cdiNeedsCalibration |= RETURN_CALIBRATED;
  if (returnCondition & RETURN_REFORMATED) cdiNeedsCalibration |= RETURN_REFORMATED;
  if (returnCondition & RETURN_STATUS) cdiNeedsCalibration |= RETURN_STATUS;
  if (returnCondition & RETURN_NAMES) cdiNeedsCalibration |= RETURN_NAMES;
  if (returnCondition & INPUT_CALIBRATED) cdiNeedsCalibration |= INPUT_CALIBRATED;
  returnCondition &= 0x7;
  
  if ( accessFlag < cdiReadFlag || accessFlag > cdiWriteReadWriteFlag ) return -argument_list_error;
  if ( returnCondition < anyData || returnCondition > maskChangePercent ) return -argument_list_error;
  if ( accessFlag != cdiReadFlag )
  {
    if ( outData == NULL || outLength == 0 ) return -argument_list_error;
  }
  if ( accessFlag != cdiWriteFlag && accessFlag != cdiWriteWriteFlag )
  {
    if ( inData == NULL || inLength == 0 ) return -argument_list_error;
  }

  if ( findDevice(devName, &cdiLine, &cdiIndex) < 0 )
  {
    if (cdiDebugLevel) dbglog("invalid device name: %s",devName);
    return -device_not_connected;
  }

  reqIndex = getFreeRequestBlock(busHandle[cdiLine]);
  if ( reqIndex < 0 )
  {
    if (cdiDebugLevel) dbglog("no free memory slot! Number of active maximum slots: %d", busHandle[cdiLine]->numberActiveRequest);
    return reqIndex;
  }

  if (cdiDebugLevel) dbglog("1. get free request Index: %d; cdiLine: %d, activeRequest: %d\n", reqIndex, cdiLine, busHandle[cdiLine]->numberActiveRequest);

  pReq = &(busHandle[cdiLine]->pReqList[reqIndex]); /* load request block */
  pReq->pDev = &(busHandle[cdiLine]->pDevList[cdiIndex]); /* assign device */
  pReq->pDev->reqFormat = readType;

  if (pReq->pDev->dataFormat != readType) cdiNeedsCalibration |= RETURN_REFORMATED;

  /* is there data to write to hardware ? */
  pReq->pOutData = outData;
  pReq->outLength = outLength;
  pReq->outType = writeType;
  /* does input data need to be reformatted ? */
  if (outLength > 0 && pReq->pDev->dataFormat != writeType) 
    cdiNeedsCalibration |= INPUT_REFORMATED;
  cdi_reverseCalibrate(pReq,cdiNeedsCalibration);

  /* check write data format and size  */
  cc = checkReqWriteData(pReq, &accessFlag, returnCondition);
  if ( cc == 0 ) 
  {
    if (!isAccessAllowed(pReq->pDev,accessFlag)) ccerr(illegal_command);
  }
  else if ( cc < 0 ) ccerr(-cc);

  /* check read format  */
  if (pReq->pDev->dataFormat >= 0 && readType >= 0)
  {
    flen = getTypeLength(readType);
    tlen = getTypeLength((short)pReq->pDev->dataFormat);
    if (flen <= 0 || flen < tlen) ccerr(illegal_format);
    pReq->inLength = inLength*tlen/flen;
    pReq->inType = pReq->pDev->dataFormat;
  }
  else
  {
    pReq->inLength = inLength;
    pReq->inType = readType;
  }
  pReq->pInData = inData;

  /* check read size  */
  if ( pReq->pDev->dataOuputSize && (int)pReq->inLength > pReq->pDev->dataOuputSize ) ccerr(dimension_error);

  pReq->hasPatternFilter = FALSE;
  if (pReq->pDev->dataPattern != NULL && returnCondition == anyData && (accessFlag != cdiWriteFlag && accessFlag != cdiWriteWriteFlag) )
  {
    returnCondition = pReq->pDev->dataPatternTrigger == CM_SEEKPATTERN ? maskCompEqual : maskCompNotEqual;
  }
  if ( returnCondition > anyData )	/* check dataPatternWindow */
  {
    if (dataPatternWindow != NULL)
    {
      if (((int *)dataPatternWindow)[0] != 0 || ((int *)dataPatternWindow)[1] != 0) 
      {
        memcpy(pReq->patternFilter,dataPatternWindow,32);
        pReq->hasPatternFilter = TRUE;
      }
    }
    else if (pReq->pDev->dataPattern != NULL)
    {
      memcpy(pReq->patternFilter,pReq->pDev->dataPattern,32);
      pReq->hasPatternFilter = TRUE;
    }
    if ( pReq->hasPatternFilter == FALSE ) ccerr(argument_list_error);
  }

  pReq->hasDataMask = FALSE;
  if ( pReq->inType <= cdiInteger && pReq->inType >= cdiByte && (accessFlag != cdiWriteFlag && accessFlag != cdiWriteWriteFlag) ) /* check dataMask for reading data */
  {
    if (dataMask != NULL)
    {
      if (((int *)dataMask)[0] != -1 || ((int *)dataMask)[1] != -1) 
	    {
        memcpy(pReq->mask,dataMask,8);
        pReq->hasDataMask = TRUE;
	    }
    }
    else if (pReq->pDev->dataMask != NULL)
    {
	    memcpy(pReq->mask,pReq->pDev->dataMask,8);
	    pReq->hasDataMask = TRUE;
    }
    if ( pReq->hasDataMask == TRUE )
    {
      pReq->maskValue = *(int *)pReq->mask;
      if ( pReq->maskValue == 0 || pReq->maskValue == -1 || (pReq->inType == cdiShort && pReq->maskValue == 0xffff) ||
            (pReq->inType == cdiByte && pReq->maskValue == 0xff) || (pReq->inType == cdiInteger && pReq->maskValue == 0xffffffff)) pReq->hasDataMask = FALSE;
    }
  }

  /* wrong parameters! */
  i = busHandle[cdiLine]->busIndex;
  if (busList[i].filterBusFcn != NULL && gSimulationFlag == FALSE )
  {
    if ((cc=busList[i].filterBusFcn(parameter != NULL ? parameter : pReq->pDev->devParameters,accessFlag)) != 0) ccerr(-cc);
  }

  if (parameter != NULL) /* use what was passed */
  {
    memcpy(pReq->devParameters, parameter, NUMBER_EXEC_PARAM * sizeof(int));
  }
  else if (pReq->pDev->devParameters != NULL) /* use the default set */
  {
    memcpy(pReq->devParameters, pReq->pDev->devParameters, NUMBER_EXEC_PARAM * sizeof(int));
  }

  pReq->accessFlag = accessFlag;
  pReq->conditionFlag = returnCondition;
  pReq->timeout = timeout; /* milliseconds */
  pReq->perror = perror;   /* user provides buffer for storing error */
  pReq->needsCalibration = cdiNeedsCalibration;

  pReq->registerTime = getTimeMilliseconds();

  pReq->requestNumber = reqIndex + (cdiLine << 16); /* upper word is cdiLine number */
  pReq->reqIndex = reqIndex;
  pReq->groupReq = NULL;
  pReq->NgroupsReqs = 0;
  pReq->groupSyncStart = 0;
  pReq->bufIndex = -1;

  if ( reqIndex == busHandle[cdiLine]->numberActiveRequest )
  {
    busHandle[cdiLine]->numberActiveRequest++;
    if (cdiDebugLevel) dbglog("set number active request: %d", busHandle[cdiLine]->numberActiveRequest);
  }
err:
  if (cc) 
  {
    if (pReq != NULL && pReq->freeOutDataMemoryOnCancel) 
    {
      free(pReq->pOutData);
      pReq->freeOutDataMemoryOnCancel = FALSE;
    }
    return -cc;
  }
  return pReq->requestNumber;
}

/************************************/
CDI_EXPORT int checkCallbackIndex(void (*callBack)(int, int),int callbackID)
{ int i, j;
  
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all device's callbackID  */
  {
	  if ( busHandle[j] == NULL ) continue;
	  for ( i = 0; i < busHandle[j]->numberActiveRequest; i++ )
	  {
	    if (busHandle[j]->pReqList[i].active == 0 || /* not active */
          busHandle[j]->pReqList[i].callBackIndex != callbackID || /* index is different */
          busHandle[j]->pReqList[i].callback != callBack)   /* function is different */
        continue;
	    return -argument_list_error; /* ids are equal and cb fcns are equal */
	  }
  }
  return 0;
}

/************************************/
/* sync or Async call */
/************************************/
CDI_EXPORT char *cdiGetRedirectedDeviceServer(char *devName)
{
  int idx;
  cdiRedirXRefStruct *xref;
  if (devName == NULL) return NULL;
  if (gCdiRedirectedDevicesHashSize == 0) return NULL;
  idx = (int)(ElfHash((unsigned char *)devName) % gCdiRedirectedDevicesHashSize);
  for (xref=cdiRedirXref[idx]; xref; xref=xref->nxt)
  {
    if (!strncmp(devName,cdiRedirXref[idx]->devName,16)) 
      return cdiRedirXref[idx]->devRedirectedServerName;
  }
  return NULL;
}
CDI_EXPORT int registerRequest(char *devName, void *readData, short readType, unsigned int readLength, void *writeData, short writeType, unsigned int writeLength, void *dataPatternWindow, 
                void *dataMask, int *parameter, int accessFlag, int returnCondition, int timeout, void (*callBack)(int, int), 
                int pollFlag, int *perror, int callbackID, int lockRequest)
{ 
  CdiRequestInfoBlk *pReq;
  int reqLineNumber, cdiLine, reqIndex;
  char *c;

  if ((c=cdiGetRedirectedDeviceServer(devName)) != NULL)
  {
    strncpy(devName,c,64);
    return -server_redirection;
  }

  if ( callBack && callbackID >= 0 && callbackID <= 0xffff )	/* check if callbackID exists */
  {
    reqLineNumber = checkCallbackIndex(callBack,callbackID);
    if ( reqLineNumber < 0 ) return reqLineNumber; /* error return! */
  }

  reqLineNumber = init_RequestBuffer(devName, readData, readType, readLength, writeData, writeType, writeLength, dataPatternWindow, dataMask, parameter, accessFlag, returnCondition, timeout,  perror);
  if ( reqLineNumber < 0 ) return reqLineNumber; /* error return! */

  cdiLine = reqLineNumber >> 16;
  reqIndex = reqLineNumber & 0xffff;
  pReq = &(busHandle[cdiLine]->pReqList[reqIndex]); /* load request block */

  pReq->callBackIndex = (callBack && callbackID >= 0 && callbackID <= 0xffff) ? callbackID : (reqLineNumber + 0x10000);
  pReq->callback = callBack;
  pReq->pollFlag = callBack == NULL ? 0 : pollFlag; /* polling with async together */
  pReq->lockRequestFlag = lockRequest;
  pReq->active = idleState;

  if (cdiDebugLevel) dbglog("preq: 0x%x callback : 0x%x; Callbackindex: %d, return Handle: 0x%x\n",pReq, callBack, pReq->callBackIndex, (reqLineNumber + 0x10000));
  return (reqLineNumber + 0x10000); /* reqLineNumber; */
}

int cdiParseDeviceString(char *devString,NAME16 *devGroup,int *groupSize)
{
  int i, len, num=0, cc=0;
  int cdiLine, index, n, istart, istop;
  char *g=NULL,*c=NULL,*g2=NULL,*cp=NULL,*rngToken=SIMPLE_RANGE_TOKEN;
  char tmpString[1024];

  if (devString == NULL || devGroup == NULL || groupSize == NULL || *groupSize < 1) ccerr(argument_list_error);
  strncpy(tmpString,devString,1024);
  /* split the seperated groups */
  for (i=0, g=strtok_r(tmpString,SEPERATOR_TOKEN,&cp); g != NULL; i++, g=strtok_r(NULL,SEPERATOR_TOKEN,&cp))
  { /* split each group into the individual devices ... */
    while (*g == ' ') g++; /* leading blanks */
    rngToken = (g[0] == '#') ? SIMPLE_RANGE_TOKEN : COMPLEX_RANGE_TOKEN;
    if ((c=strstr(g,rngToken)) != NULL)
    {
      g2 = c + strlen(rngToken);
      *c = 0;
      while (*g2 == ' ') g2++; /* leading blanks in front of range terminator */
      for (len=strlen(g2); len > 0 && g2[len-1] == ' '; g2[len-1] = 0); /* trailing blanks */
    }
    for (len=strlen(g); len > 0 && g[len-1] == ' '; g[len-1] = 0); /* trailing blanks */
    if ((istart=findDevice(g,&cdiLine,&index)) < 0) 
    {
      cdilog("device %s was not found !",g);
      ccerr(device_not_connected);
    }
    if ( c == NULL ) istop = istart;
    else if ((istop=findDevice(g2,&cdiLine,&index)) < 0) 
    {
      cdilog("device %s was not found !",g2);
      ccerr(device_not_connected);
    }

    for (n=istart; ; ) /* allow high to low */
    {
      if (num < *groupSize)
#     ifdef WINNT
        _snprintf(devGroup[num].name,16,"#%d",n);
#     else
        snprintf(devGroup[num].name,16,"#%d",n);
#     endif

      num++;
      if ( n == istop ) break; 
      istart < istop ? (n++) : (n--);
    }

  }
  *groupSize = num;
err:
  return cc;
}

/************************************/
CDI_EXPORT int cdi_Request(char *devName, void *readData, short readType, unsigned int readLength, void *writeData, short writeType, unsigned int writeLength, void *dataPatternWindow, 
                void *dataMask, int *parameter, int accessFlag, int returnCondition, int timeout, void (*callBack)(int, int), 
                int pollFlag, int *perror, int callbackID, int lockRequest)
{ int i, j, numDevices, ret, *pInt;
  NAME16 localDevName, *pDevName;
  int rLen, wLen;

  if (readLength <= 0) readType = -1;
  numDevices = 1;
  ret = cdiParseDeviceString(devName, &localDevName, &numDevices);
  if (cdiDebugLevel) dbglog("cdiParseDeviceString: numDevices: %d; name: %s; return: %d\n",numDevices, localDevName.name, ret);
  if ( ret ) return -ret;
  if ( numDevices == 1 )
  {
    return registerRequest(devName, readData, readType, readLength, writeData, writeType, writeLength, dataPatternWindow, 
                dataMask, parameter, accessFlag, returnCondition, timeout, callBack, pollFlag, perror, callbackID, lockRequest);
  }
  /* for group call */
  if ( (pDevName = (NAME16 *)calloc(numDevices, sizeof(NAME16))) == NULL ) return -out_of_local_memory;
  if ( (pInt = (int *)calloc(numDevices, sizeof(int))) == NULL )
  {
    free(pInt);
    return -out_of_local_memory;
  }
  ret = cdiParseDeviceString(devName, pDevName, &numDevices);
  rLen = (readLength == 0 || readData == NULL ) ?  0 : (readLength / numDevices);
  wLen = (writeLength == 0 || writeData == NULL ) ? 0 : (writeLength / numDevices);
  for (i=0; i<numDevices; i++)
  {
    pInt[i] = registerRequest(pDevName[i].name, 
      rLen == 0 ? NULL : &((char *)readData)[rLen * i], readType, rLen, 
      wLen == 0 ? NULL : &((char *)writeData)[wLen * i], writeType, wLen, 
      dataPatternWindow, dataMask, parameter, accessFlag, returnCondition, timeout, 
      callBack, pollFlag, perror==NULL ? NULL : &perror[i], i == 0 ? callbackID : -1, lockRequest);
    if ( pInt[i] < 0 ) break;
    if ( i > 0 )
    {
      ret = cdi_loadGroupRequests(pInt[0], pInt[i]);
      if ( ret ) break;
    }
  }
  if ( i < numDevices || ret ) /* error */
  {
    for (j=0; j<=i, pInt[i]>0; j++) cdi_CancelRequest(pInt[j]); /* error  */
    if ( ret == 0 ) ret = pInt[i] < 0 ? pInt[i] : -invalid_data; 
  }
  else ret = pInt[0];
  free(pDevName);
  free(pInt);
  return ret;
}
#ifdef CHANGE_CALLBACK_ID
/************************************/
CDI_EXPORT int findReqIndex(int requestIndex)
{ 
  int i, j;

  for (i=0; i<numConnectedHandles; i++)
  {
    if ( busHandle[i] == NULL ) return -1;
    for (j=0; j<busHandle[i]->numberActiveRequest; j++)
      if ( requestIndex == busHandle[i]->pReqList[j].callBackIndex ) break;
    if ( j < busHandle[i]->numberActiveRequest ) return busHandle[i]->pReqList[j].requestNumber + 0x10000;
  }
  return -1;
}
#endif
/************************************/
/* check request index */
/************************************/
CDI_EXPORT int cdi_isValidRequestIndex(int requestIndex)
{ int cdiLine, reqIndex;

  if ( requestIndex < 0x1000 ) return FALSE;

  cdiLine = (requestIndex >> 16) - 1;
  reqIndex= (requestIndex & 0xffff);
  if (cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL || reqIndex >= busHandle[cdiLine]->numberActiveRequest ) return FALSE;
  return TRUE; /* if it is a valid requestIndex ! */
}

/************************************/
CDI_EXPORT int get_RequestBufferPointer(int requestIndex, CdiRequestInfoBlk **pReq)
{ int cdiLine, reqIndex;

  if ( cdi_isValidRequestIndex(requestIndex) == FALSE )
  {
    if (cdiDebugLevel) dbglog("invalid input: %x; (line %d, index %d)",
      requestIndex, (requestIndex >> 16) - 1, requestIndex & 0xffff);
    return -non_existent;
  }
  cdiLine = (requestIndex >> 16) - 1;
  reqIndex = requestIndex & 0xffff;
  *pReq = &(busHandle[cdiLine]->pReqList[reqIndex]); /* load request block */

  return 0;
}

/************************************/
CDI_EXPORT int getGroupRequestInfo(int requestIndex, void *buffers, int index)
{ int i, j, offset;
  CdiLineInfoBlk *pLine;
  CdiRequestInfoBlk *pReq;

  if ( buffers == NULL ) return -illegal_data_size;
  if ( (i = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return i;
  if ( pReq->groupReq == NULL )
  {
    if ( index == CDI_INFO_ERRORS ) *(int *)buffers = pReq->errorCode;
    else if ( index == CDI_INFO_NAMES ) strncpy(buffers, pReq->pDev->devName, 16);
    else if ( index == CDI_INFO_STATUS ) *(int *)buffers = pReq->active;
    else if ( index == CDI_INFO_NDEVS ) *(int *)buffers = 1;
    else if ( index == CDI_INFO_ERRCNT ) *(int *)buffers = pReq->pDev->numberDeviceError;
  }
  else for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belonging to this group */
  {
    if ( index == CDI_INFO_NDEVS )
    {
      *(int *)buffers = ((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs;
      break;
    }
    else if ( index == CDI_INFO_ERRCNT )
    {
      *(int *)buffers = ((CdiRequestInfoBlk *)pReq->groupReq)->NgrpErrors;
      break;
    }
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      if (cdiDebugLevel) dbglog("getGroupRequestInfo-> device group index: %d\n",pLine->pReqList[i].NgroupsReqs);
      offset = &pLine->pReqList[i] == pReq ? 0 :  pLine->pReqList[i].NgroupsReqs;
      if ( index == CDI_INFO_ERRORS ) ((int *)buffers)[ offset ] = pLine->pReqList[i].errorCode;
      else if ( index == CDI_INFO_NAMES ) strncpy( ((NAME16 *)buffers)[offset].name, pLine->pReqList[i].pDev->devName, 16);
      else if ( index == CDI_INFO_STATUS ) ((int *)buffers)[ offset ] = pLine->pReqList[i].active;
    }
  }
  return 0;
}

/************************************/
CDI_EXPORT int cdi_GetRequestErrors(int requestIndex, int *errorArray)
{
  return getGroupRequestInfo(requestIndex, (void *)errorArray, CDI_INFO_ERRORS);
}
/************************************/
CDI_EXPORT int cdi_GetRequestDeviceNames(int requestIndex, NAME16 *namebuffer)
{
  return getGroupRequestInfo(requestIndex, (void *)namebuffer, CDI_INFO_NAMES);
}
CDI_EXPORT int cdi_GetRequestStat(int requestIndex, int *activeStatArray)
{
  return getGroupRequestInfo(requestIndex, (void *)activeStatArray, CDI_INFO_STATUS);
}
CDI_EXPORT int cdi_GetRequestNumberDevices(int requestIndex, int *numberOfDevices)
{
  return getGroupRequestInfo(requestIndex, (void *)numberOfDevices, CDI_INFO_NDEVS);
}

/************************************/
/* first call cdi_Request for all(N), then call this rountine N times to group them */
CDI_EXPORT int loadGroupRequests(CdiRequestInfoBlk *pMainReq, CdiRequestInfoBlk *pFollowerReq)
{
  if ( pMainReq == NULL || pFollowerReq == NULL ) return -invalid_index;	/* no such a request */
  if ( pFollowerReq->callback != pMainReq->callback ) return -ambiguous;	/* not same for sunc or async */
  if ( pFollowerReq->callback && pFollowerReq->pollFlag != pMainReq->pollFlag ) return -ambiguous;	/* not same for polling */
  if ( pMainReq->NgroupsReqs == 0 ) /* first time call  */
  {
    pMainReq->groupReq = pMainReq;
    pMainReq->NgroupsReqs++;
  }
  /* check if already signed  */
  if ( pFollowerReq->groupReq != NULL ) return (pFollowerReq->groupReq != pMainReq->groupReq) ? -illegal_equipment_name : 0;
  pFollowerReq->groupReq = pMainReq->groupReq;  /* add this new element to group  */
  pFollowerReq->NgroupsReqs = pMainReq->NgroupsReqs;  /* index number for this group request: 1, 2, 3....  pMainReq->NgroupsReqs */
  pMainReq->NgroupsReqs++;
  return 0;
}

/************************************/
CDI_EXPORT int cdi_loadGroupRequests(int requestIndexMain, int requestIndexFollower)
{ int retCode;
  CdiRequestInfoBlk *pMainReq, *pFollowerReq;

  if ( (retCode = get_RequestBufferPointer(requestIndexMain, &pMainReq)) < 0 ) return retCode;
  if ( (retCode = get_RequestBufferPointer(requestIndexFollower, &pFollowerReq)) < 0 ) return retCode;
  return loadGroupRequests(pMainReq, pFollowerReq);
}

/************************************/
CDI_EXPORT int postBackTaskSemaphore(CdiRequestInfoBlk *pReq)
{ int i, j, ret=0, tmpret;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL )
  {
    ret = cdiPostSemaphore(&(((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->back_SemId));
    if ( ret )
      if (cdiDebugLevel) dbglog("background SemId : cdiPostSemaphore error");
    return ret;
  }

  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      /* start background task for this cdi line  */
      tmpret = cdiPostSemaphore(&(((CdiLineInfoBlk *)pLine->pReqList[i].pDev->pcdiLine)->back_SemId));
      if (tmpret)
      {
        ret = tmpret;
        if (cdiDebugLevel) dbglog("background SemId : cdiPostSemaphore error");
      }
      break;
    }
  }
  return ret;
}


/************************************/
CDI_EXPORT int cdiStopCancel(CdiRequestInfoBlk *pReq, int cancelFlag)
{ int tmp;

  if ( pReq->active == 0 || pReq->active > finishRequest) return 0;	/* nothing to do */
  if ( pReq->callback == NULL && pReq->active > idleState && pReq->active < finishRequest )
  {
    pReq->active = finishRequest;		/* make it finish!!!!! */
    pReq->errorCode = not_properly;
    if ( pReq->perror ) *(pReq->perror) = not_properly;
    if ( pReq->groupReq == NULL ) cdiPostSemaphore(&pReq->sync_SemId);
    else if ( checkGroupReturnFlags(pReq) ) cdiPostSemaphore(&((CdiRequestInfoBlk *)pReq->groupReq)->sync_SemId);
  }
  else pReq->userCallDown = 1;
  tmp = pReq->active == idleState ? 1 : 0;
  /* avoid user does: first cancel, then stop	*/
  if ( pReq->cancelFlag != 2 ) pReq->cancelFlag = cancelFlag ? 2 : 1;	/* stop: 1; cancel: 2 */
  pReq->active = finishRequest;

  /* activate background task */
  if ( tmp && postBackTaskSemaphore(pReq) ) 
  {
    if (cdiDebugLevel) dbglog("background SemId : cdiPostSemaphore error");
  }
  return 0;
}

/************************************/
/* do rest of the group items	*/
CDI_EXPORT void cdiStopCancelGroup(CdiRequestInfoBlk *pReq, int cancelFlag)
{ int i, j;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL ) return;
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      if ( (CdiRequestInfoBlk *)&(pLine->pReqList[i]) == pReq ) continue;
      cdiStopCancel(&(pLine->pReqList[i]), cancelFlag);
    }
  }
}
void assignClbr(double dval,BYTE *data,int dataFormat)
{
  switch (dataFormat)
  {
    default: 
    case cdiChar: *(data) = (BYTE)dval; break;
    case cdiByte: *(data) = (BYTE)dval; break;
    case cdiShort: *((UINT16 *)data) = (UINT16)dval; break;
    case cdiInteger: *((UINT32 *)data) = (UINT32)dval; break;
    case cdiFloat: *((float *)data) = (float)dval; break;
    case cdiDouble: *((double *)data) = dval; break;
  }
}
void assignMsg(char *msg,BYTE *data,int dataFormat)
{
  switch (dataFormat)
  {
    default:
      return;
    case cdiChar32:
      /* must ensure that there's room to put this ! */
      strncpy((char *)data,msg,32);
      return;
  }
}
char *applyMsg(CdiRule *rule,BYTE *data,int dataFormat)
{
  CdiRule *r = NULL, *msg;
  int hasMessage = FALSE, isSigned = FALSE;
  int mval = 0;

  if (rule != NULL) isSigned = rule->isSigned;
  switch (dataFormat)
  {
    default: 
    case cdiChar: mval = *(data); break;
    case cdiByte: mval = *(data); break;
    case cdiShort: mval = isSigned ? *(short *)(data) : *(UINT16 *)(data); break;
    case cdiInteger: mval = isSigned ?  *(SINT32 *)(data) : *(UINT32 *)(data); break;
  }
  for (r=rule; r != NULL && !hasMessage; r=r->nxt)
  {
    switch (r->operation)
    {
      case cdiRuleSub: 
        mval -= (int)r->operand; 
        break;
      case cdiRuleAdd: 
        mval += (int)r->operand; 
        break;
      case cdiRuleMult: 
        mval *= (int)r->operand; 
        break;
      case cdiRuleDiv: 
        if (r->operand != 0) mval /= (int)r->operand; 
        break;
      case cdiRulePwr: 
        mval = (int)pow((double)mval,r->operand);
        break;
      case cdiRuleMod:
        mval = (mval)%((int)r->operand);
        break;
      case cdiRuleShiftRight:
        mval = (mval)>>((int)r->operand);
        break;
      case cdiRuleShiftLeft:
        mval = (mval)<<((int)r->operand);
        break;
      case cdiRuleXor:
        mval = (mval)^((int)r->operand);
        break;
      case cdiRuleExp: 
        mval = (int)pow(r->operand,(double)mval);
        break;
      case cdiRuleFcn:
        if (r->fcn != NULL) mval = (int)r->fcn((double)mval);
        break;
      case cdiRuleMessage:  
        hasMessage = TRUE; /* will also break the rules chain */
        msg = r;
        break;
      case cdiRuleSigned:
        isSigned = TRUE;
        break;
      default:
        break;
    }
  }
  if (!hasMessage) return NULL;
  return mval == msg->msgOnValue ? msg->msgOn : msg->msgOff;
}
double applyClbr(CdiRule *rule,BYTE *data,int dataFormat)
{
  CdiRule *r = NULL;
  double dval;
  int isSigned = FALSE;
  /* first rule sets the signed/unsigned conversion */
  if (rule != NULL) isSigned = rule->isSigned;
  switch (dataFormat)
  {
    default: 
    case cdiChar: dval = *(data); break;
    case cdiByte: dval = *(data); break;
    case cdiShort: dval = isSigned ? *(short *)(data) : *(UINT16 *)(data); break;
    case cdiInteger: dval = isSigned ? *(SINT32 *)(data) : *(UINT32 *)(data); break;
    case cdiFloat: dval = *(float *)(data); break;
    case cdiDouble: dval = *(double *)(data); break;
  }
  for (r=rule; r != NULL; r=r->nxt)
  {
    switch (r->operation)
    {
      case cdiRuleSub: 
        dval -= r->operand; 
        break;
      case cdiRuleAdd: 
        dval += r->operand; 
        break;
      case cdiRuleMult: 
        dval *= r->operand; 
        break;
      case cdiRuleDiv: 
        if (r->operand != 0) dval /= r->operand; 
        break;
      case cdiRulePwr: 
        dval = pow(dval,r->operand);
        break;
      case cdiRuleLog:
        /* no operand here ! */
        if (dval != 0) dval = log10(dval);
        break;
      case cdiRuleMod:
        dval = ((int)dval)%((int)r->operand);
        break;
      case cdiRuleShiftRight:
        dval = ((int)dval)>>((int)r->operand);
        break;
      case cdiRuleShiftLeft:
        dval = ((int)dval)<<((int)r->operand);
        break;
      case cdiRuleXor:
        dval = ((int)dval)^((int)r->operand);
        break;
      case cdiRuleExp: 
        dval = pow(r->operand,dval);
        break;
      case cdiRuleFcn:
        if (r->fcn != NULL) dval = r->fcn(dval);
        break;
      default: break;
    }
  }
  return dval;
}
static CdiDeviceInfoBlk *getDevPtr(CdiRequestInfoBlk *pReq,int groupIndex)
{ 
  int i, j, k = 0;
  CdiLineInfoBlk *pLine;

  if (pReq->groupReq == NULL) return pReq->pDev;
  for (j=0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for (i=0; i < pLine->numberActiveRequest; i++ )
    {
      if (pLine->pReqList[i].groupReq != pReq->groupReq) continue;
      if (k == groupIndex) return pLine->pReqList[i].pDev;
      k++;
    }
  }
  return NULL;
}
int clbrRequest(CdiRequestInfoBlk *pReq,int applyRules)
{
  int i, flen, tlen, dlen;
  CdiDeviceInfoBlk *pDev;
  int fromFormat, toFormat;
  BYTE *fromData, *toData;
  CdiRule *rule = NULL;
  double dval;
  char *msg;

  if ((fromFormat=pReq->inType) < 0) return -illegal_format;
  flen = getTypeLength((short)fromFormat);
  dlen = pReq->inLength/flen;
  if (flen <= 0) return -illegal_format;
  toData = fromData = pReq->pInData;
  pDev = pReq->pDev;
  toFormat = getCdiBaseFormat(pReq->inFormatCode);
  if (toFormat < 0) return -illegal_format;
  if ((tlen=getTypeLength((short)toFormat)) <= 0) return -illegal_format;
  if (!applyRules && toFormat == fromFormat) return 0; /* nothing to do */
  rule = applyRules ? pDev->clbrRules : NULL;
  for (i=dlen-1; i>=0; i--)
  { /* start from the end and work backwards for in situ assignments */
    if (toFormat == cdiChar32)
    {
      if ((msg=applyMsg(rule,fromData+i*flen,fromFormat)) != NULL)
      {
        assignMsg(msg,toData+i*tlen,toFormat);
      }
    }
    else
    {
      dval = applyClbr(rule,fromData+i*flen,fromFormat);
      assignClbr(dval,toData+i*tlen,toFormat);
    }
  }
  return 0;
}
CDI_EXPORT int cdi_CalibrateRequest(int requestIndex,int applyRules)
{
  int cc, j, k;
  CdiRequestInfoBlk *pReq;
  CdiLineInfoBlk *pLine;

  if ((cc=get_RequestBufferPointer(requestIndex, &pReq)) < 0) return cc;

  if (pReq->groupReq == NULL) return clbrRequest(pReq,applyRules);
  for (j=0; j<numConnectedHandles; j++)	/* search all devices belonging to this group */
  {
    if (busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for (k=0; k<pLine->numberActiveRequest; k++)
    {
      if (pLine->pReqList[k].groupReq != pReq->groupReq) continue;
      if ((cc=clbrRequest(&pLine->pReqList[k],applyRules)) < 0) return cc;
    }
  }      
  return 0;
}

/************************************/
/* cancel Async call */
/************************************/
CDI_EXPORT int cdi_CancelRequest(int requestIndex)
{
  CdiRequestInfoBlk *pReq;
  int retCode;

  if ( (retCode = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return retCode;
  if ( pReq->groupReq && pReq->groupReq != pReq ) return -invalid_index;
  if (pReq->callback != NULL) cdilog("cdi cancel link : %d",requestIndex);
  cdiStopCancelGroup(pReq, 1);
  return cdiStopCancel(pReq, 1);
}

/************************************/
CDI_EXPORT void initStartRequest(CdiRequestInfoBlk *pReq, int flag)
{
  if ( flag & 0x1 )
  {
    pReq->dataReturnCondition = 0;
    pReq->errorCode = 0;
    pReq->deviceError = 0;
    pReq->cancelFlag = 0;
    pReq->requestCounter = 0;
    if ( pReq->perror ) *(pReq->perror) = 0;

    pReq->asynCallFlag = 0;
    pReq->userCallDown = 0;
    pReq->NgroupRets = 0;
    pReq->NgrpErrors = 0;
    pReq->groupSyncStart = 1;
  }
  if ( flag & 0x2 )
  {
    pReq->requestTime = getTimeMilliseconds();
    pReq->responseTime = pReq->requestTime - 1;
    pReq->active = transRequest;	/* mark start flag! */
  }
}	

/************************************/
CDI_EXPORT void doStartRequest(CdiRequestInfoBlk *pReq)
{ int i, j;
  CdiLineInfoBlk *pLine;
  CdiRequestInfoBlk *pR;

  if ( pReq->groupReq == NULL )
  {
    initStartRequest(pReq, 0x3);
    return;
  }
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      pR = (CdiRequestInfoBlk *)&(pLine->pReqList[i]);
      initStartRequest(pR, 0x1);
    }
  }
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      pR = (CdiRequestInfoBlk *)&(pLine->pReqList[i]);
      if ( pR == pReq ) continue;
      initStartRequest(pR, 0x2);
      /* Let background task start the rest of the group call */
      /*      if ( pReq->callback == NULL ) cdiExecRequest(pR, execIndex); */
    }
  }
  initStartRequest(pReq, 0x2);
}

/************************************/
CDI_EXPORT int cdi_StartRequest(int requestIndex)
{ CdiRequestInfoBlk *pReq;
  int retCode;

  if ( (retCode = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return retCode;
  if ( pReq->groupReq && pReq->groupReq != pReq ) return -invalid_index;
  if (cdiDebugLevel) dbglog("cdi_StartRequest, preq: %x,  call back == %x...\n", pReq, pReq->callback);
  if ( pReq->active == 0 ) return -database_not_loaded;	/* should call cdi_Request first or lock it... */
  if ( pReq->active != idleState ) return -operation_busy;	/* not finish yet */

  doStartRequest(pReq);

  if ( pReq->callback == NULL ) /* sync call */
  {
    while ( cdiWaitSemaphore(&pReq->sync_SemId,TRUE) == 0 ); /* clear semaphore, send request */
    cdiExecRequest(pReq, execIndex);
  }
  /* start background task */
  postBackTaskSemaphore(pReq);

  if ( pReq->callback == NULL ) /* sync call return error number */
  {
    if ( cdiWaitSemaphore(&pReq->sync_SemId,0) ) 
    {
      if (cdiDebugLevel) dbglog("wait on sync semaphore ERROR! cdiLine: %d, requestNumber: %d", ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->cdiLine, pReq->reqIndex);
    }
    pReq->userCallDown++;
    retCode = setGroupReturnFlags(pReq, 0x5);
    return retCode ? -retCode : -pReq->errorCode; /* if group call, return -(Number of errors) */
  }
  /* asyn return with requestIndex */
  return requestIndex; /* reqLineNumber; */

}
/************************************/
CDI_EXPORT int cdi_StopRequest(int requestIndex)
{ CdiRequestInfoBlk *pReq;
  int retCode;

  if ( (retCode = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return retCode;
  if ( pReq->groupReq && pReq->groupReq != pReq ) return -invalid_index;
  cdiStopCancelGroup(pReq, 0);
  return cdiStopCancel(pReq, 0);
}

/************************************/
CDI_EXPORT int cdi_getDeviceNumberOfErrors(char *devName)
{ int i, j;
  if ( findDevice(devName, &i, &j) < 0 ) return -device_not_connected;
  return busHandle[i]->pDevList[j].numberDeviceError;
}

/************************************/
CDI_EXPORT char *cdi_getBusNameForCdiLine(int cdiLine)
{
  if ( busHandle[cdiLine] == NULL ) return NULL;
  return busList[ busHandle[cdiLine]->busIndex ].busName;
}

/************************************/
CDI_EXPORT char *cdi_getBusNameForDevice(char *devName)
{ 
  int i, j;

  if ( findDevice(devName, &i, &j) < 0 ) return NULL;
  return busList[ busHandle[i]->busIndex ].busName;
}

/************************************/
CDI_EXPORT int cdi_getBusNames(NAME16 *namebuffer,int namesize)
{ 
  int i;
  for (i=0; i<numBusList && i<namesize; i++) strncpy(namebuffer[i].name,busList[i].busName,16);
  return numBusList;
}

/************************************/
CDI_EXPORT DADDRESS *cdi_getBusAddressForDevice(char *devName)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;
  static DADDRESS da;

  memset(&da,0,sizeof(da));
  if ( findDevice(devName, &i, &j) < 0 ) return &da;
  pDev = &busHandle[i]->pDevList[j];
  da.i1val = ((CdiLineInfoBlk *)pDev->pcdiLine)->cdiLine;
  da.i2val = pDev->cdiAddr;
  da.i3val = pDev->cdiSubAddr;
  da.i4val = pDev->cdiIndex;
  return &da;
}

/************************************/
CDI_EXPORT int cdi_getDescriptionForDevice(char *devName,char *devDesc)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devDesc == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  strncpy(devDesc,pDev->description,64);
  return 0;
}
CDI_EXPORT int cdi_getUnitsForDevice(char *devName,char *devUnits)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devUnits == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  strncpy(devUnits,pDev->units,16);
  return 0;
}
CDI_EXPORT int cdi_getPrecisionForDevice(char *devName,char *devPrec)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devPrec == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  strncpy(devPrec,pDev->precision,16);
  return 0;
}
CDI_EXPORT int cdi_getRangeForDevice(char *devName,float *devRange)
{
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devRange == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  devRange[0] = pDev->minimum;
  devRange[1] = pDev->maximum;
  return 0;
}

/************************************/
CDI_EXPORT int getCDIDeviceNames(NAME16 *namebuffer,int namesize,int cdiLine, int oneLineFlag)
{ 
  int i, j, n=0, max;
  if ( oneLineFlag )
  {
    if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
    i = cdiLine;
    max = cdiLine + 1;
  }
  else
  {
    i = 0;
    max = numConnectedHandles;
  }
  for (; i<max; i++ ) /* search through all bus lines */
  {
    if ( busHandle[i] == NULL ) continue;
    for ( j=0; j<busHandle[i]->numberDevices; j++ )
    {
      if ( busHandle[i]->pDevList[j].devName[0] == 0 ) continue;
      if ( namebuffer != NULL ) strncpy(namebuffer[n].name, busHandle[i]->pDevList[j].devName, 16);
      /*if ( namebuffer != NULL && ++n >= namesize ) return n; */
      n++;
    }
  }
  return n;
}

/************************************/
CDI_EXPORT int cdi_getDeviceNames(NAME16 *namebuffer,int namesize)
{
  return getCDIDeviceNames(namebuffer,namesize,0,0);
}

/************************************/
CDI_EXPORT int cdi_getDeviceNamesByLine(NAME16 *namebuffer, int namesize, int cdiLine)
{
  return getCDIDeviceNames(namebuffer,namesize,cdiLine,1);
}
/************************************/
CDI_EXPORT int cdi_GetNumberDevicesByLine(int cdiLine)
{
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
  return busHandle[cdiLine]->numberDevices;
}
/************************************/
/* Register Parameters with Name */
/* return index number */
/************************************/
CDI_EXPORT int cdi_RegisterParameters(char *paramName, int *parameters)
{ 
  int i;

  if (paramList == NULL) return -database_not_loaded;
  if (paramName == NULL || parameters == NULL || strlen(paramName) >= 32 ) return -argument_list_error;
  for (i=0; i<paramListSize; i++)
  {
    if ( strlen(paramList[i].paramName) == 0 ) break; /* empty slot */
    if ( strncmp(paramList[i].paramName, paramName, 32) == 0 ) break; /* previous registered */
  }
  if (i == paramListSize) return -link_exhausted;
  strncpy(paramList[i].paramName, paramName, 32);
  memcpy(paramList[i].devParameters, parameters, NUMBER_EXEC_PARAM * sizeof(int));
  return i;
}

/****************************************/
/* get Registered Parameters with Name */
/****************************************/
CDI_EXPORT int *cdi_GetParameters(char *paramName)
{ 
  int i;

  if (paramList == NULL) return NULL;
  if (paramName == NULL || strlen(paramName) >= 32 ) return NULL;
  for (i=0; i<paramListSize; i++)
  {
    if ( strlen(paramList[i].paramName) == 0 ) break; /* empty slot */
    if ( strncmp(paramList[i].paramName, paramName, 32) == 0 ) return paramList[i].devParameters; /* previous registered */
  }
  return NULL;
}

/************************************/
/* Register Devices */
/************************************/
int maxCdiRedirectedDevices = MAX_REDIRECTED_DEVICES;
int gCdiRedirectedDevicesHashSize = 0;
CDI_EXPORT int cdi_RegisterRedirectedDevice(char *devName, int devNumber, char *redirectedServerName)
{
  int idx;
  cdiRedirXRefStruct *xref;
  static int done = 0;

  if (!done)
  {
    if (maxCdiRedirectedDevices < CDI_PAYOFF_HASHSIZE) gCdiRedirectedDevicesHashSize = CDI_TINY_HASHSIZE;
    else if (maxCdiRedirectedDevices < CDI_SMALL_HASHSIZE) gCdiRedirectedDevicesHashSize = CDI_SMALL_HASHSIZE;
    else if (maxCdiRedirectedDevices < CDI_MEDIUM_HASHSIZE) gCdiRedirectedDevicesHashSize = CDI_MEDIUM_HASHSIZE;
//    else if (maxCdiRedirectedDevices < CDI_LARGE_HASHSIZE) gCdiRedirectedDevicesHashSize = CDI_LARGE_HASHSIZE;
    else gCdiRedirectedDevicesHashSize = CDI_LARGE_HASHSIZE;
    if ((cdiRedirXref=(cdiRedirXRefStruct **)calloc(gCdiRedirectedDevicesHashSize,sizeof(cdiRedirXRefStruct *))) == NULL)
    {
      cdilog("cannot allocate redirected cross-reference buffer");
      exit(1);
    }
    done = TRUE;
  }
  idx = (int)(ElfHash((unsigned char *)devName) % gCdiRedirectedDevicesHashSize);
  for (xref=cdiRedirXref[idx]; xref && strncmp(devName,cdiRedirXref[idx]->devName,16); xref=xref->nxt);
  if (xref == NULL) /* it's a new one*/
  {
    if ((xref=(cdiRedirXRefStruct *)SystemCalloc(1,sizeof(cdiRedirXRefStruct))) == NULL) 
    {
      cdilog("CDI: cannot allocate hash table entry !");
      return -out_of_server_memory;
    }
    strncpy(xref->devName,devName,16);
    strncpy(xref->devRedirectedServerName,redirectedServerName,64);
    xref->devNumber = devNumber;
    xref->nxt = cdiRedirXref[idx];
    cdiRedirXref[idx] = xref;
    nCdiRedirectedDevices++;
  }
  return 0;
}
CDI_EXPORT int cdi_getRedirectedDeviceNames(NAME16 *namebuffer,int namesize)
{
  int i,n=0;
  cdiRedirXRefStruct *xref;
  if (namebuffer == NULL) return nCdiRedirectedDevices;
  for (i=0; i<gCdiRedirectedDevicesHashSize; i++)
  {
    for (xref=cdiRedirXref[i]; xref; xref=xref->nxt)
    {
      if (n >= namesize) return n;
      strncpy(namebuffer[n++].name,xref->devName,16);
    }
  }
  return n;
}
int getNextFreeIndex(int cdiLine)
{
  int idx = 0;
    
  if (cdiLine < 0 || cdiLine >= NUMBER_LINES) return -illegal_line;
  idx = CdiNumIndiciesPerLine[cdiLine];
  if (idx >= busHandle[cdiLine]->numberDevices) return -resources_exhausted;
  CdiNumIndiciesPerLine[cdiLine] = idx + 1;
  return idx;
}
int checkDevNumber(char *devName, int cdiLine, int cdiIndex, int devNumber)
{ int i, j;

  if ( devNumber <= 0 )
  {
    cdilog("global device number %d invalid( >0 allowed)\n",devNumber);
    return invalid_index;
  }
  if ( cdiIndex < 0 || cdiIndex >= busHandle[cdiLine]->numberDevices)
  {
    cdilog("cdi Index: %d invalid (0 to %d allowed)\n",cdiIndex,busHandle[cdiLine]->numberDevices-1);
    return invalid_index;
  }
  for (i=0; i<numConnectedHandles; i++)
  {
    if ( busHandle[i] == NULL ) continue;
    for ( j=0; j<busHandle[i]->numberDevices; j++)
    {
      if ( strncmp(devName, busHandle[i]->pDevList[j].devName, 16) == 0 )
      {
        cdilog("device Name %s (nr: %d) repeated at cdiLine %d cdiIndex %d (nr %d)",
          devName, devNumber, i, j, busHandle[i]->pDevList[j].devNumber);
        return invalid_index;
      }
      if ( busHandle[i]->pDevList[j].devNumber != devNumber ) continue;
      
      cdilog("global device number %d invalid! In use by %s (cdiLine: %d cdiIndex: %d)", 
        busHandle[i]->pDevList[j].devName, devNumber, i, j);
      cdilog("try to use a device number in the range 0 to %d",cdiAddrDbSize+TEMPLATE_RECORD_GAP);
      return invalid_index;
    }
  }
  return 0;
}

void parseDeviceDescription(char *devDesc,char *desc,char *prec,char *units,float *mn,float *mx)
{
  char *c0, *c1, *c, *p, *v, strBuffer[128], *tok = " \t";
  int len;

  desc[0] = 0;
  prec[0] = units[0] = 0;
  *mn = *mx = 0;
  strncpy(strBuffer,devDesc,96); strBuffer[96] = 0;
  strtrm(strBuffer);
  if ((c0=strchr(strBuffer,'[')) == NULL)
  { /* this one's easy: no display info to parse */
    strncpy(desc,strBuffer,64);
    return;
  }
  *c0++ = 0;
  strncpy(desc,strBuffer,64); /* first section (or empty string) */
  len = strlen(desc);
  if ((c1=strchr(c0,']')) == NULL) 
  {
    cdilog("parseDeviceDescription : %s did not parse correctly!",devDesc);
    return;
  }
  *c1++ = 0;
  strncpy(strBuffer,c0,32);
  for (p=strtok(strBuffer,tok); p != NULL; p=strtok(NULL,tok))
  {
    if ((v=strchr(p,'=')) == NULL) continue;
    *v++ = 0;
    switch (tolower(*p))
    {
      case 'r': /* range */
        *mn = (float)strtod(v,NULL);
        if ((c=strchr(v,':')) != NULL)
        {
          *c++ = 0;
          *mx = (float)strtod(c,NULL);
        }
        break;
      case 'p': /* precision */
        strncpy(prec,v,16);
        break;
      case 'u': /* units */
        strncpy(units,v,16);
        break;
      default:
        break;
    }
  }
  /* second section */
  strncpy(strBuffer,c1,64); strBuffer[64] = 0;
  if ((c0=strchr(strBuffer,'[')) == NULL)
  { /* second set of display info : ignore ! */
    strncat(desc,strBuffer,64-len);
    return;
  }
  *c0++ = 0;
  strncat(desc,strBuffer,64-len);
  len = strlen(desc);
  if ((c1=strchr(c0,']')) == NULL) 
  {
    cdilog("parseDeviceDescription : %s did not parse correctly!",devDesc);
    return;
  }
  *c1++ = 0;
  strncat(desc,c1,64-len);
  return;
}

CDI_EXPORT int cdi_RegisterDevice(
  int cdiLine, char *devName, char *lngName, char *devDesc,
  int cdiIndex, int devNumber, int address, int sub_address
)
{ 
  CdiDeviceInfoBlk *pDev;
  cdiHashXRefStruct *xref;
  int idx;
  char devNrStr[32];
  char desc[64], prec[16], units[16];
  float mx, mn;

  if (devName == NULL) return argument_list_error;

  if (cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL)
  {
    if (cdiDebugLevel) dbglog("line number %d not open",cdiLine);
    return illegal_line;
  }
  if ( (idx = checkDevNumber(devName, cdiLine, cdiIndex, devNumber)) != 0 ) return idx;

  pDev = &(busHandle[cdiLine]->pDevList[cdiIndex]);
  strncpy(pDev->devName, devName, 16);
  if ( strlen(devName) > 16 )
  {
    cdilog("truncating device name %s to 16 characters !",devName);
  }

  if (lngName == NULL) pDev->lngName[0] = 0;
  else if ( strlen(lngName) > 64 )
  {
    cdilog("truncating long device name %s to 64 characters !",lngName);
    strncpy(pDev->lngName, lngName, 64);
  }
  else strncpy(pDev->lngName, lngName, 64);

  parseDeviceDescription(devDesc,desc,prec,units,&mn,&mx);
  strncpy(pDev->description, desc, 64);
  strncpy(pDev->precision,prec,16);
  strncpy(pDev->units,units,16);
  pDev->maximum = mx;
  pDev->minimum = mn;
  pDev->cdiAddr = address; /* can and sedac type */
  pDev->cdiSubAddr = sub_address;
  pDev->cdiIndex = cdiIndex;
  pDev->devNumber = devNumber;
  pDev->pcdiLine = busHandle[cdiLine];
  pDev->numberDeviceError = 0;
  pDev->dataMaskStartIndex = pDev->dataMaskStopIndex = -1; /* not applicable */
  pDev->dataPatternStartIndex = pDev->dataPatternStopIndex = -1; /* not applicable */

  if (busHandle[cdiLine]->devHashList != NULL) /* there's a hash table */
  {
    idx = (int)(ElfHash((unsigned char *)busHandle[cdiLine]->pDevList[cdiIndex].devName) % busHandle[cdiLine]->devHashSize);
    for (xref=busHandle[cdiLine]->devHashList[idx]; 
         (xref != NULL) && (xref->cdiLine != cdiLine || xref->cdiIndex != cdiIndex); 
         xref=xref->nxt);
    if (xref == NULL) /* it's a new one*/
    {
      if ((xref=(cdiHashXRefStruct *)SystemCalloc(1,sizeof(cdiHashXRefStruct))) == NULL) 
      {
        cdilog("CDI: cannot allocate hash table entry !");
        return out_of_server_memory;
      }
      xref->cdiLine = cdiLine;
      xref->cdiIndex = cdiIndex;
      xref->devNumber = devNumber;
      xref->nxt = busHandle[cdiLine]->devHashList[idx];
      busHandle[cdiLine]->devHashList[idx] = xref;
      /* now add the '#<devnr>' representation as well */
      if ((xref=(cdiHashXRefStruct *)SystemCalloc(1,sizeof(cdiHashXRefStruct))) == NULL) 
      {
        cdilog("CDI: cannot allocate hash table entry !");
        return out_of_server_memory;
      }
      xref->cdiLine = cdiLine;
      xref->cdiIndex = cdiIndex;
      xref->devNumber = devNumber;
      sprintf(devNrStr,"#%d",devNumber);
      idx = (int)(ElfHash((unsigned char *)devNrStr) % busHandle[cdiLine]->devHashSize);
      xref->nxt = busHandle[cdiLine]->devHashList[idx];
      busHandle[cdiLine]->devHashList[idx] = xref;
    }
  }
  if (cdiDebugLevel) dbglog("Register device name %s (devNum %d, addr %d, subaddr %d)",devName, devNumber, address, sub_address);
  return 0;
}
CDI_EXPORT int cdi_assignDefaultDeviceParameters(char *devName,int *devParameters)
{
  int cc, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;
  char paramName[36];

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0)
  {
    if (cdiDebugLevel) dbglog("invalid device name: %s",devName);
    return device_not_connected;
  }
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];

  strncpy(paramName,devName,16); paramName[16] = 0;
  strcat(paramName,"default");
  if ((cc=cdi_RegisterParameters(paramName,devParameters)) < 0) return -cc;
  pDev->devParameters = cdi_GetParameters(paramName);
  return 0;
}
CDI_EXPORT int cdi_assignAllowedAccess(char *devName,int access)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->allowedAccess = access;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignAllowedAccess for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDefaultDataMaskAndPattern(char *devName,BYTE *dataMask,int *maskRange,BYTE *pattern,int *patternRange,int patternTrigger)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  if (dataMask != NULL) 
  {
    if ((pDev->dataMask=(BYTE *)calloc(8,1)) == NULL) ccerr(out_of_server_memory);
    memcpy(pDev->dataMask,dataMask,8);
  }
  if (maskRange != NULL)
  {
    pDev->dataMaskStartIndex = maskRange[0];
    pDev->dataMaskStopIndex = maskRange[1];
  }
  if (pattern != NULL)
  {
    if ((pDev->dataPattern=(BYTE *)calloc(32,1)) == NULL) ccerr(out_of_server_memory);
    memcpy(pDev->dataPattern,pattern,32);
    pDev->dataPatternTrigger = patternTrigger;
  }
  if (patternRange != NULL)
  {
    pDev->dataPatternStartIndex = patternRange[0];
    pDev->dataPatternStopIndex = patternRange[1] != -1 ? patternRange[1] : patternRange[0];
  }
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDefaultDataMaskAndPattern for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDataLimits(char *devName,int dataSizeOut,int dataSizeIn)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->dataOuputSize = dataSizeOut;
  pDev->dataInputSize = dataSizeIn;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDefaultLimits for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDefaultInput(char *devName,BYTE *pDataInput,int dataSize,int dataFormatIn)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  if (dataFormatIn < 0) ccerr(illegal_format);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->dataInput = pDataInput;
  pDev->dataInputSize = dataSize;
  pDev->dataFormat = dataFormatIn;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDefaultInput for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDataFormat(char *devName,int dataFormat)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->dataFormat = dataFormat;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDataFormat for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDataCalibration(char *devName,CdiRule *clbrRules,CdiRule *rvrsRules,int dataBusFormat)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
    pDev->dataFormat = dataBusFormat;
  pDev->clbrRules = clbrRules;
  pDev->rvrsRules = rvrsRules;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDataFormat for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_eraseDefaultDataMaskAndPattern(char *devName,int erasedataMask, int erasepattern)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  if ( erasedataMask && pDev->dataMask != NULL) 
  {
    free(pDev->dataMask);
    pDev->dataMask = NULL;
  }
  if (erasepattern && pDev->dataPattern != NULL)
  {
    free(pDev->dataPattern);
    pDev->dataPattern = NULL;
  }
err:
  if (cc)
  {
    dbglog("CDI: cdi_eraseDefaultDataMaskAndPattern for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
int getNumberLines(char *busName, int flag)
{
  int i,j,cnt;
  if ( flag == 1 )  /* check busName */
  {
    for (i=0; i<numBusList; i++) if (!strncmp(busName, busList[i].busName,16)) break;
    if (i == numBusList) return -not_implemented;
  }
  cnt=0;
  for ( j = 0; j < numConnectedHandles; j++ )
  {
    if ( busHandle[j] == NULL ) continue;
    if ( flag == 1 && busList[ busHandle[j]->busIndex ].busType != busList[i].busType ) continue;
    cnt++;
  }
  return cnt;
}
CDI_EXPORT int cdi_GetNumberOpenedLines(void)
{
  return getNumberLines(NULL, 0);
}
CDI_EXPORT int cdi_GetNumberOpenedLinesByBus(char *busName)
{
  return getNumberLines(busName, 1);
}
CDI_EXPORT int cdi_GetCdiLine(char *busName, int busLine)
{
  int i,cdiLine;
  for (i=0; i<numBusList; i++) if (!strncmp(busName, busList[i].busName, 16)) break;
  if (i == numBusList) return -not_implemented;
  if ((cdiLine=checkLine(busList[i].busType,busLine)) >= 0) return cdiLine;
  return -no_such_line;
}
/************************************/
/* return CDI line index */
/************************************/
CDI_EXPORT int cdi_Open(char *busName, int busLine, char *parameter, int numberDevices, int numberRequestBuffers)
{ int i, cdiLine = -1, numBuffers;
  int shared, value;
  int fileHandle, retSem = 0, retSemCall = 0, cc = 0;
  static int insideCdiOpen = FALSE;

  if (insideCdiOpen) ccerr(operation_busy);
  insideCdiOpen = TRUE;
  if (cdiInitFlag == 0 && (i=cdiInitialize()) < 0) ccerr(-i);

  for (i=0; i<numBusList; i++) if (strncmp(busName, busList[i].busName, 16) == 0) break;
  if (i == numBusList)
  {
    cdilog("CDI: BusName not supported! %.16s",busName);
    ccerr(not_implemented);
  }

  if ((cdiLine = checkLine(busList[i].busType, busLine)) >= 0) ccerr(0);

  cdiLine = findFreeLine();
  if ( cdiLine < 0 )
  {
    cdilog("CDI: cannot find available slot!; maximum slots = %d", numConnectedHandles);
    ccerr(resources_exhausted);
  }
  cdilog("CDI: using free cdi line %d", cdiLine);

  if ( numberDevices < 1 ) numberDevices = 1; /* at least one device */
  fileHandle = connectToBus(busList[i].busType, busLine, parameter, cdiLine, numberDevices);
  if ( fileHandle < 0 )
  {
    if (cdiDebugLevel) dbglog("open line error; line Index = %d", busLine);
    ccerr(code_failure);
  } 
  if ( numberRequestBuffers <= 0 ) numBuffers = MULTI_N_DEVICE * numberDevices;
  else if ( numberRequestBuffers > MULTI_MAX_DEVICE * numberDevices ) numBuffers = MULTI_MAX_DEVICE * numberDevices;
  else numBuffers = numberRequestBuffers;
  /* allocate memory... */
  if ((busHandle[cdiLine]=(CdiLineInfoBlk *)calloc((size_t)LINE_INFOR_SIZE,1)) == NULL) 
    ccerr(out_of_server_memory);
  if ((busHandle[cdiLine]->pDevList=(CdiDeviceInfoBlk *)calloc((size_t)DEVICE_INFOR_SIZE,numberDevices)) == NULL) 
    ccerr(out_of_server_memory);
  if ((busHandle[cdiLine]->pReqList=(CdiRequestInfoBlk *)calloc((size_t)REQUEST_INFOR_SIZE,numBuffers)) == NULL) 
    ccerr(out_of_server_memory);

  busHandle[cdiLine]->busIndex = i;
  busHandle[cdiLine]->cdiLine = cdiLine;
  busHandle[cdiLine]->busLine = busLine;
  busHandle[cdiLine]->fileHandle = fileHandle;
  busHandle[cdiLine]->numberDevices = numberDevices;
  busHandle[cdiLine]->numberRequest = numBuffers;
  cdiMutexInit(&busHandle[cdiLine]->lineMutex, cdiLine);

  retSem = 0;
  for (i=0; i<busHandle[cdiLine]->numberRequest; i++)
  {
    shared = value = 0;
    if ((retSem=cdiGetSemaphore(&busHandle[cdiLine]->pReqList[i].sync_SemId,shared,value)) != 0) ccerr(semaphore_error);
  }

  shared = value = 0;
  if ((retSem=cdiGetSemaphore(&busHandle[cdiLine]->back_SemId,shared,value)) != 0) ccerr(semaphore_error);
  shared = value = 0;
  if ((retSemCall=cdiGetSemaphore(&busHandle[cdiLine]->call_SemId,shared,value)) != 0) ccerr(semaphore_error);
  if ((busHandle[cdiLine]->tIdCallback=(pthread_t)cdiCreateThread(busHandle[cdiLine], THREAD_INDEX_CALLBACK)) == 0) ccerr(software_error);
  if ((busHandle[cdiLine]->tIdMaster=(pthread_t)cdiCreateThread(busHandle[cdiLine], THREAD_INDEX_MASTER)) == 0) ccerr(software_error);
  busHandle[cdiLine]->threadActive = TRUE;

  if ( cdiLine == numConnectedHandles ) numConnectedHandles++;
  cdilog("CDI: cdi_Open -> cdi Line number %d success",cdiLine);

  makeCdiHashTable(cdiLine,numberDevices); /* will do nothing if already allocated */
err:
  if (cc && cdiLine >= 0 && busHandle[cdiLine] != NULL)
  {
    if (busHandle[cdiLine]->pDevList != NULL) free (busHandle[cdiLine]->pDevList);
    else cdilog("CDI: could not allocate memory for busHandle %d device list",cdiLine);
    if (busHandle[cdiLine]->pReqList != NULL) free (busHandle[cdiLine]->pReqList);
    else cdilog("CDI: could not allocate memory for busHandle %d request list",cdiLine);
    if (busHandle[cdiLine] != NULL) free(busHandle[cdiLine]);
    else cdilog("CDI: could not allocate memory for busHandle %d",cdiLine);
    if (busHandle[cdiLine]->tIdCallback != 0) cdiCancelThread((int)busHandle[cdiLine]->tIdCallback);
    else cdilog("CDI: could not create thread for busHandle %d",cdiLine);
    if (busHandle[cdiLine]->tIdMaster != 0) cdiCancelThread((int)busHandle[cdiLine]->tIdMaster);
    else cdilog("CDI: could not create thread for busHandle %d",cdiLine);
    if (retSemCall != 0) cdiDestroySemaphore(&busHandle[cdiLine]->call_SemId);
    else cdilog("CDI: could not create semaphore for busHandle %d",cdiLine);
    if (retSem != 0) cdiDestroySemaphore(&busHandle[cdiLine]->back_SemId);
    else cdilog("CDI: could not create semaphore for busHandle %d",cdiLine);
    cdilog("CDI: cdi_Open -> %s",erlst[cc]); 
    disconnectFromBus(busList[i].busType, busLine, fileHandle);
  }
  insideCdiOpen = FALSE;
  return cc ? -cc : cdiLine;
}

/**************************************/

CDI_EXPORT void cdi_Cleanup(void)
{ 
  int i;
  for (i=0; i<numConnectedHandles; i++) cdi_Close(i);
  return;
}

CDI_EXPORT int cdi_Close(int cdiLine)
{ 
  int status, i, cc;

  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL )
  {
    cdilog("CDI: cdi_Close -> cdi line %d invalid",cdiLine);
    return -argument_list_error;
  }
  status = disconnectFromBus(busList [ busHandle[cdiLine]->busIndex ].busType, busHandle[cdiLine]->busLine, busHandle[cdiLine]->fileHandle);
  if ( status >= 0 )
  {
    busHandle[cdiLine]->threadActive = FALSE;
    while (busHandle[cdiLine]->threadBusy) nanosleep( &busHandle[cdiLine]->backReq, NULL );
    cc = cdiCancelThread((int)busHandle[cdiLine]->tIdCallback); 
    cdilog("cdiCancelThread callback thread: %d",cc);
    cc = cdiCancelThread((int)busHandle[cdiLine]->tIdMaster);
    cdilog("cdiCancelThread background thread: %d",cc);
    cdiDestroySemaphore(&busHandle[cdiLine]->back_SemId);
    cdiDestroySemaphore(&busHandle[cdiLine]->call_SemId);
    for (i=0; i<busHandle[cdiLine]->numberDevices; i++)
    {
      if (busHandle[cdiLine]->pDevList[i].dataMask != NULL) free(busHandle[cdiLine]->pDevList[i].dataMask);
      if (busHandle[cdiLine]->pDevList[i].dataPattern != NULL) free(busHandle[cdiLine]->pDevList[i].dataPattern);
    }
    free ( busHandle[cdiLine]->pDevList );
    free ( busHandle[cdiLine]->pReqList );
    free ( busHandle[cdiLine] );
    busHandle[cdiLine] = NULL;
    if ( cdiLine == numConnectedHandles - 1 )
    {
      i = cdiLine;
      for ( --i; i >= 0 && busHandle[i] == NULL ; --i )
      numConnectedHandles = i + 1;
    }
  }

  if (cdiDebugLevel) dbglog("cdi_Close OK! line number: %d; status: %d",cdiLine, status);

  return status;
}
CDI_EXPORT int cdiRegisterBusFilter(char *busName,int (*fcn)(int *,int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName, 16); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register filter function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].filterBusFcn = fcn;
  cdilog("cdi filter function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusConfiguration(char *busName,int (*fcn)(void))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,16); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register configuration function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].cfgBusFcn = fcn;
  /* don't call this in cdi.c !! */
  if (busList[i].cfgBusFcn != NULL && gSimulationFlag == FALSE )
  {
    if (busList[i].cfgBusFcn() < 0)
    {
      dbglog("failure configuring %s",busList[i].busName);
    }
  }
  cdilog("cdi configuration function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusInitialization(char *busName,int (*fcn)(int,int,int,char *))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,16); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register initialization function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].initBusFcn = fcn;
  cdilog("cdi initialization function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusHandler(char *busName,void (*fcn)(CdiRequestInfoBlk *))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,16); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register bus handler for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].execBusFcn = fcn;
  cdilog("cdi bus handler for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusCleanup(char *busName,int (*fcn)(int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,16); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register cleanup function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].freeBusFcn = fcn;
  cdilog("cdi cleanup function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBus(char *busName,int busType)
{
  int id = busType; /* if positive, use it */
  if (id < 0) id = cdiGetBusType(busName); /* is this bus already registered ? */
  if (id >= 0) return id;  /* if registered already, just return the busType */
  id = cdiGetNextAvailableBusType(); /* it's a new one ! */
  return cdi_RegisterBus(busName,id,NULL,NULL,NULL,NULL,NULL);

}
CDI_EXPORT int cdiScanBus(char *busName,char *target,char *result,int resultBufferSize)
{
  int i;
  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,16); i++);
  if (i == numBusList) 
  {
#   ifdef WINNT
    _snprintf(result,32,"%s not registered",busName);
#   else
    snprintf(result,32,"%s not registered",busName);
#   endif
    return non_existent;
  }
  if (busList[i].scanBusFcn == NULL || gSimulationFlag != FALSE )
  {
#   ifdef WINNT
    _snprintf(result,32,"no bus scanner registered");
#   else
    snprintf(result,32,"no bus scanner registered");
#   endif
    return not_implemented;
  }
  return busList[i].scanBusFcn(target,result,resultBufferSize);
}
CDI_EXPORT int cdiRegisterBusScanner(char *busName,int (*fcn)(char *,char *,int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,16); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register initialization function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].scanBusFcn = fcn;
  cdilog("cdi bus scanner for bus %.16s registered",busName);
  return 0;
}
CdiTemplate *getCdiTemplate(char *templateName)
{
  int i;
 
  for (i=0; i<CdiTemplateTblSize; i++)
    if (!strncmp(templateName,CdiTemplateTbl[i]->name,16)) 
      return CdiTemplateTbl[i];

  if (i >= MAX_NUM_TEMPLATES)
  {
    cdilog("no resources to allocate new template for %.16s",templateName);
    return NULL;
  }
  cdilog("adding new template %.16s to the template table",templateName);
  if ((CdiTemplateTbl[i]=(CdiTemplate *)calloc(1,sizeof(CdiTemplate))) == NULL)
  {
    cdilog("cannot allocate new template %.16s",templateName);
    return NULL;
  }
  CdiTemplateTblSize++;
  CdiTemplateTbl[i]->length = 0;
  strncpy(CdiTemplateTbl[i]->name,templateName,16);
  return CdiTemplateTbl[i];
}
CdiTemplateList *findFieldInTemplate(CdiTemplate *tmpt,char *field)
{
  CdiTemplateList *t;
  if (tmpt == NULL || field == NULL) return NULL;
  for (t=tmpt->cdiTemplateList; t != NULL; t=t->next)
  {
    if (!strncmp(t->field,field,16)) return t;
  }
  return NULL;
}

int addFieldToTemplate(
  CdiTemplate *tmpt, char *field, char *lngfield, char *fieldDesc,
  int *addrParameters,int naddrParameters,
  BYTE *mask,int *maskRange,BYTE *pattern,int *patternRange,int patternTrigger,
  BYTE *dataIn,int maxBytesIn,int maxBytesOut,int dataFmt,
  int access,
  CdiRule *rules_rcv,CdiRule *rules_snd
)
{
  CdiTemplateList *t;
  if (tmpt == NULL || field == NULL) return argument_list_error;
  if (findFieldInTemplate(tmpt,field) != NULL) return link_exists;
  if ((t=(CdiTemplateList *)calloc(1,sizeof(CdiTemplateList))) == NULL) return out_of_local_memory;
  strncpy(t->field,field,16);
  strncpy(t->lngfield,lngfield,64);
  strncpy(t->description,fieldDesc,96);
  if (addrParameters != NULL) memcpy(t->addrParameters,addrParameters,NUMBER_EXEC_PARAM*sizeof(int));
  t->naddrParameters = naddrParameters;
  if (mask != NULL)
  {
    if ((t->dataMask=(BYTE *)calloc(8,1)) == NULL) return out_of_server_memory;
    memcpy(t->dataMask,mask,8);
  }
  if (maskRange != NULL) memcpy(t->dataMaskRange,maskRange,2*sizeof(int));
  if (pattern != NULL)
  {
    if ((t->dataPattern=(BYTE *)calloc(32,1)) == NULL) return out_of_server_memory;
    memcpy(t->dataPattern,pattern,32);
  }
  if (patternRange != NULL) memcpy(t->dataPatternRange,patternRange,2*sizeof(int));
  t->dataPatternTrigger = patternTrigger;
  t->dataInput = dataIn;
  t->dataInputSize = maxBytesIn;
  t->dataOutputSize = maxBytesOut;
  t->dataFormat = dataFmt;
  t->allowedAccess = access;
  t->clbrRules = rules_rcv;
  t->rvrsRules = rules_snd;
  /* add this one to the list */
  t->next = tmpt->cdiTemplateList;
  tmpt->cdiTemplateList = t;
  tmpt->length++;  
  return 0;
}

CDI_EXPORT int cdi_getTemplateNames(NAME16 *namebuffer,int namesize)
{ 
  int i;
  for (i=0; i<CdiTemplateTblSize && i<namesize; i++)
  {
    strncpy(namebuffer[i].name,CdiTemplateTbl[i]->name,16);
  }
  return CdiTemplateTblSize;
}
CDI_EXPORT int cdi_getTemplate(char *templateName,NAME16II *cdiTemplate,int *namesize)
{ 
  int i;
  CdiTemplate *tmpl;
  CdiTemplateList *t;

  if (templateName == NULL || cdiTemplate == NULL || 
      namesize == NULL || *namesize == 0) return argument_list_error;
  for (i=0; i<CdiTemplateTblSize; i++)
    if (!strncmp(templateName,CdiTemplateTbl[i]->name,16)) break;

  if (i == CdiTemplateTblSize) return non_existent;
  tmpl = CdiTemplateTbl[i];
  t = tmpl->cdiTemplateList;

  for (i=0; i<tmpl->length && i<*namesize && t != NULL; i++) 
  {
    strncpy(cdiTemplate[i].name,t->field,16);
    cdiTemplate[i].i1val = t->addrParameters[0];
    cdiTemplate[i].i2val = t->addrParameters[1];
    t = t->next;
  }
  *namesize = tmpl->length;
  return 0;
}
