/********************************************************/
/* Common Device Interface definition file */
/* ------------ cdi.h --------------- */
/********************************************************/

#ifndef CDI_H_
#define CDI_H_


#include "cdi.def"

#ifndef CDI_EXPORT
# define CDI_EXPORT
#endif

#include "cdi_prot.h"

extern time_t gCdiStartupTime;

#define CDI_PAYOFF_HASHSIZE 25
#define CDI_TINY_HASHSIZE 31
#define CDI_SMALL_HASHSIZE 251
#define CDI_MEDIUM_HASHSIZE 1021
#define CDI_LARGE_HASHSIZE 4093

typedef struct cdiHashXRef
{
  int cdiLine;
  int cdiIndex;
  int devNumber;
  struct cdiHashXRef *nxt;
} cdiHashXRefStruct;

#define DEF_INPUT_BYTES 32  // default input bytes for input
#define MAX_REDIRECTED_DEVICES 256
typedef struct cdiRedirXRef
{
  char devName[16];
  char devRedirectedServerName[64];
  int devNumber;
  struct cdiRedirXRef *nxt;
} cdiRedirXRefStruct;

#define BACK_THREAD_TIME 10000 /* 10 ms for usleep */
#define MULTI_N_DEVICE 3 /* allocate multiple RequestBlocks */
#define MULTI_MAX_DEVICE 20 /* allocate multiple RequestBlocks */
#define NUMBER_LINES 20 /* total number of bus lines */

#define NUMBER_EXEC_PARAM 16 /* number parameters for exec and poll Request call */

#define DATA_BACK 1
#define DATA_PASS_FILTER 2

#define THREAD_INDEX_MASTER		1
#define THREAD_INDEX_CALLBACK	2

#define RETURN_NAMES     0x2000
#define RETURN_STATUS    0x1000
#define RETURN_CALIBRATED 0x800
#define RETURN_REFORMATED 0x400
#define INPUT_CALIBRATED  0x200
#define INPUT_REFORMATED  0x100

#define CDICHAR_ADD '+'
#define CDICHAR_SUB '-'
#define CDICHAR_MUL '*'
#define CDICHAR_DIV '/'
#define CDICHAR_PWR '^'
#define CDICHAR_LOG 'L'
#define CDICHAR_MOD '%'
#define CDICHAR_RGH '>'
#define CDICHAR_LFT '<'
#define CDICHAR_UNS 'U'
#define CDICHAR_SGN 'S'
#define CDICHAR_MSG 'M'
#define CDICHAR_EXP 'E'
#define CDICHAR_FCN '|'
#define CDICHAR_FCNA 'F'
#define CDICHAR_XOR 'X'

enum cdiRuleOperation
{
  cdiRuleAdd,
  cdiRuleSub,
  cdiRuleMult,
  cdiRuleDiv,
  cdiRulePwr,
  cdiRuleLog,
  cdiRuleMod,
  cdiRuleShiftRight,
  cdiRuleShiftLeft,
  cdiRuleUnsigned,
  cdiRuleSigned,
  cdiRuleMessage,
  cdiRuleExp,
  cdiRuleFcn,
  cdiRuleXor
};

typedef struct cdiRuleStruct
{
  int operation;
  int isSigned;
  int msgOnValue;
  int reserved;
  double operand;
  double (*fcn)(double);
  char msgOn[32];
  char msgOff[32];
  struct cdiRuleStruct *nxt;
} CdiRule;

typedef struct cdiClbrFcnStruct
{
  char fcnName[64];
  double (*fcn)(double);
  struct cdiClbrFcnStruct *next;
} CdiClbrFcn;

extern CdiClbrFcn *cdiClbrFcnLst;
int registerClbrConversionFunction(char *fcnName,double (*fcn)(double));

typedef struct CdiTemplateListStruct
{
  char field[16];    /* field name to be applied */
  char lngfield[64]; /* long field name to be applied */
  char description[96]; /* field description */
  int  addrParameters[NUMBER_EXEC_PARAM]; /* relative offsets */
  int  naddrParameters;
  BYTE *dataMask;  /* default data mask */
  int  dataMaskRange[2];
  BYTE *dataPattern;/* default data tolerance/pattern */
  int  dataPatternRange[2];
  int  dataPatternTrigger;
  BYTE *dataInput;   /* default data input for WRITE commands */
  int dataInputSize; /* maximum allowed data input size in bytes (if > 0) */
  int dataOutputSize; /* maximum allowed data output size in bytes (if > 0) */
  int dataFormat;    /* default data type */
  int allowedAccess; /* allowed bus access flag */
  CdiRule *clbrRules;/* associated calibration rules for device readout */
  CdiRule *rvrsRules;/* associated reverse calibration rules for device sendto */
  struct CdiTemplateListStruct *next;
} CdiTemplateList;

typedef struct 
{
  char name[16];
  int length;
  CdiTemplateList *cdiTemplateList;
} CdiTemplate;

#define MAX_NUM_TEMPLATES 32
extern CdiTemplate *CdiTemplateTbl[MAX_NUM_TEMPLATES];
extern int CdiTemplateTblSize;
CdiTemplate *getCdiTemplate(char *templateName);
CdiTemplateList *findFieldInTemplate(CdiTemplate *tmpt,char *field);
int addFieldToTemplate(
  CdiTemplate *tmpt, char *field, char *lngfield, char *fieldDesc,
  int *addrParameters,int naddrParameters,
  BYTE *mask,int *maskRange,BYTE *pattern,int *patternRange,int patternTrigger,
  BYTE *dataIn,int maxBytesIn,int maxBytesOut,int dataFmt,
  int access,
  CdiRule *rules_rcv, CdiRule *rules_snd
);

/* bind address to device name */
typedef struct
{
  char devName[16];  /* registered (hashed) device name */
  int cdiIndex;      /* start from 0, for pDev[] */
  int devNumber;     /* start from 1, unique number */
  void *pcdiLine;    /*  pointer to CdiLineInfoBlk (before: int cdiLine; index for CdiLineInfoBlk) */
  int cdiAddr;       /* module address */
  int cdiSubAddr;    /* module sub-address */
  int numberDeviceError;
  int *devParameters;/* default bus parameters for this device */
  BYTE *dataMask;    /* default data mask */
  int dataMaskStartIndex;
  int dataMaskStopIndex;
  BYTE *dataPattern; /* default data tolerance/pattern */
  int dataPatternStartIndex;
  int dataPatternStopIndex;
  int dataPatternTrigger;
  BYTE *dataInput;   /* default data input for WRITE commands */
  int dataInputSize; /* maximum allowed data input size in bytes (if > 0) */
  int dataOuputSize; /* maximum allowed data output size in bytes (if > 0) */
  int dataFormat;    /* default data type */
  int reqFormat;     /* requested data type */
  int allowedAccess; /* allowed bus access flag */
  CdiRule *clbrRules;/* associated calibration rules for device readout */
  CdiRule *rvrsRules;/* associated reverse calibration rules for device sendto */
  char lngName[64];  /* associated long device name */
  char description[64];/* associated device description */
  float maximum;     /* display maximum */
  float minimum;     /* display minimum */
  char precision[16];/* display format a la width.decimal */ 
  char units[16];    /* display units */
} CdiDeviceInfoBlk;
#define DEVICE_INFOR_SIZE sizeof(CdiDeviceInfoBlk)

typedef struct
{
  char paramName[32];
  int devParameters[NUMBER_EXEC_PARAM];
} PARAM_STRUCT;

typedef struct
{
  int active;			/* request process steps */
  int callBackIndex;	/* can be user signed (< 0x1000) or requestNumber + 0x1000 */
  CdiDeviceInfoBlk *pDev;	/* referenced device */
  void *pInData;		/* from device */
  unsigned int inLength;
  void *pOutData;		/* to device */
  unsigned int outLength;
  short inType, outType;	/* char, short, int, float or double	*/
  short inFormatCode, outFormatCode;  /* original caller's code */
  int hasPatternFilter;
  BYTE patternFilter[32]; /* was ->  void *patternFilter; */ /* compare pattern or Filter, if string max 32 bytes, otherwise check 8 bytes(double) only */
  int hasDataMask;
  BYTE mask[8];  /*   was -> void *mask; */
  int maskValue;
  int devParameters[NUMBER_EXEC_PARAM];
  int timeout; /* in milliseconds: < 0 return only on time out or error; 0: wait forever */
  int registerTime; /* contract register time */
  int requestTime; /* send time */
  int requestCounter;
  short accessFlag, /* read 1, write 2, read then write 3, write then read 4 */
		conditionFlag; /* for reading data (readConditionNumeraters), bit: 0x8 for mask ON flag */
  int needsCalibration;
  int requestNumber;  /*  reqIndex + (cdiLine << 16) */
  int reqIndex;		/* own index number: 0, 1... */
  void *groupReq;	/* grouped request, group true if != NULL */
  short NgroupsReqs,		/* unmber of connected requests for mainRequest, but index number for each requested device start from 1  */
	  NgroupRets,		/* number of data returnd or timeout... */
    NgrpErrors,   /* number of errors of this group  */
    groupSyncStart;    /* synchronous flag for group start */
  short dataReturnCondition, /* databack: 1; conditionOK: 2 */
	  asynCallFlag;
  int deviceError;
  BYTE  cpOutData[DEF_INPUT_BYTES];	/* for default input */
  short bufLastData[4];	/* for compare	*/
  int errorCode, *perror;
  int responseTime; /* last response time in ms */
  short pollFlag, /* infinite loop for async call */
		lockRequestFlag,	/* lock this request until cancel calls	*/
		userCallDown,	/* callback or syncSem done flag */
		cancelFlag;		/* cancel callback */
  sem_t sync_SemId;
  int bufIndex;   /* for bus plug  */
  void (*callback)(int, int); /* callBackIndex, errorCode */
  int freeInDataMemoryOnCancel;
  int freeOutDataMemoryOnCancel;
} CdiRequestInfoBlk;
#define REQUEST_INFOR_SIZE sizeof(CdiRequestInfoBlk)

typedef struct
{
  int busIndex; /* index into CdiBusInfoBlk table */
  int cdiLine; /* line number for infor Struct, start from 0,1,2 */
  int busLine; /* bus line number, e.g sedac line 1,2,3,4... */
  int fileHandle; /* opened file handle */
  int numberRequest;
  int numberActiveRequest;
  int numberDevices;
  int backCounter, activeCounter;
  int readCounter, callCounter;
  struct timeval tv;
  int backTime; /* background task timer in msec */
  struct timespec backReq;
  CdiDeviceInfoBlk *pDevList;
  CdiRequestInfoBlk *pReqList;
  cdiHashXRefStruct **devHashList;
  int devHashSize;
  int threadActive;
  int threadBusy;
  sem_t back_SemId, call_SemId;
  cdi_mutex_t lineMutex; /* for user- and background-task */
  cdi_thread_t tIdMaster; /* background task */
  cdi_thread_t tIdCallback; /* callback task */
} CdiLineInfoBlk;
#define LINE_INFOR_SIZE sizeof(CdiLineInfoBlk)

typedef struct
{
  char busName[16];
  int busType;
  int (*filterBusFcn)(int *parameter, int accessFlag);/* associated bus parameter filter */
  int (*cfgBusFcn)(void);/* associated bus configurer */
  int (*initBusFcn)(int busline,int cdiline,int ndevices,char *paramList);/* associated bus initializer */
  void (*execBusFcn)(CdiRequestInfoBlk *pReq); /* associated bus executer */
  int (*freeBusFcn)(int busline);             /* associated bus disconnecter */
  int (*scanBusFcn)(char *target,char *result,int resultBufferSize); /* associated bus scanning function */
} CdiBusInfoBlk;

enum taskIndexNumeraters /* parameters for call cdiExecRequest() */
{
  readTaskIndex,
  backTaskIndex,
  execIndex,
};

enum reqStatus /* parameters for requestActive flag */
{
  idleState = 1,		/* after registration	 */
  transRequest,
  recvRequest,
  transRequest2nd,    // start from backgroud task immediatly
  recvRequest2nd,
  finishRequest,
  removeRequest,
};

enum cdiDataType	/* cdi support primitive data type */
{
  cdiChar,
  cdiByte,
  cdiShort,
  cdiInteger,
  cdiFloat,
  cdiDouble,
  cdiChar32,
};

extern PARAM_STRUCT *paramList;
extern int paramListSize;
extern int numBusList;
extern CdiLineInfoBlk *busHandle[NUMBER_LINES];
extern int CdiNumIndiciesPerLine[NUMBER_LINES];
extern int numConnectedHandles;
extern int cdiInitFlag;
extern int cdiDebugLevel;

CDI_EXPORT THREAD_TYPE ThreadMaster(THREAD_PARAM param);
CDI_EXPORT THREAD_TYPE ThreadCallback(THREAD_PARAM param);

extern cdiRedirXRefStruct **cdiRedirXref;
extern int nCdiRedirectedDevices;
extern int maxCdiRedirectedDevices;
extern int gCdiRedirectedDevicesHashSize;
int cdiGetNextAvailableBusType(void);
int getNextFreeIndex(int cdiLine);
int cdiGetBusType(char *busName);
int cdiRemapDataSets(DTYPE *dout,BYTE *dataout,int *errLst,NAME16 *nameout);
int cdiParseDeviceString(char *devString,NAME16 *devGroup,int *groupSize);
CDI_EXPORT int findDevice(char *devName, int *pCdiLine, int *pDevIndex);
CDI_EXPORT void checkReturnCondition(CdiRequestInfoBlk *pReq, char *lastData, char *str, int datalen, int stringType);
CDI_EXPORT void cdiExecRequest(CdiRequestInfoBlk *pReq, int taskIndex);
CDI_EXPORT int cdiInitialize(void);

/* supported bus type: */
/* BusType: "SEDAC", "RS232", "CAN", "GPIB" */
CDI_EXPORT int cdi_Open(char *busName, int busLine, char *parameter, int numberDevices, int numberRequestBuffers);
CDI_EXPORT int cdi_Close(int cdiLine);
CDI_EXPORT int cdi_RegisterDevice(
  int cdiLine, char *devName, char *lngName, char *devDesc,
  int cdiIndex, int devNumber, int address, int sub_address
  );
CDI_EXPORT int cdi_RegisterParameters(char *paramName, int *parameters);
CDI_EXPORT int *cdi_GetParameters(char *paramName);
/* get routines mainly for statistics */ 
CDI_EXPORT int cdi_GetCdiLine(char *busName, int busLine); /* get cdiLine index from harware bus name("SEDAC") and line number(sedac line number)  */ 
CDI_EXPORT int cdi_GetNumberRegisteredBus(void);
CDI_EXPORT int cdi_GetNumberRequests(int cdiLine);  /* cdicdiLine */
CDI_EXPORT int cdi_GetNumberOpenedLines(void);  /* totol opened lines */
CDI_EXPORT int cdi_GetNumberConnectedHandles(void);
CDI_EXPORT int cdi_GetNumberOpenedLinesByBus(char *busName);  /* for this type of bus */
CDI_EXPORT int cdi_GetNumberDevicesByLine(int cdiLine);  /* cdicdiLine */
CDI_EXPORT int cdi_GetRequestErrors(int requestIndex, int *errorArray);
CDI_EXPORT int cdi_GetRequestDeviceNames(int requestIndex, NAME16 *namebuffer);
CDI_EXPORT int cdi_GetRequestStat(int requestIndex, int *activeStatArray);
CDI_EXPORT int cdi_GetRequestNumberDevices(int requestIndex, int *numberOfDevices);
CDI_EXPORT int cdi_Request(
  char *devName, void *readData, short readType, unsigned int readLength, void *writeData, short writeType, 
  unsigned int writeLength, void *dataPatternWindow, void *dataMask, int *parameter, 
  int accessFlag, int returnCondition, int timeout, void (*callBack)(int, int), 
  int pollFlag, int *perror, int callbackID, int lockRequest
  );
CDI_EXPORT int cdi_loadGroupRequests(int requestIndexMain, int requestIndexFollower);
CDI_EXPORT int cdi_isValidRequestIndex(int requestIndex);
CDI_EXPORT int cdi_CancelRequest(int requestIndex);
CDI_EXPORT int cdi_StartRequest(int requestIndex);
CDI_EXPORT int cdi_StopRequest(int requestIndex);
CDI_EXPORT int cdi_assignDefaultDeviceParameters(char *devName,int *devParameters);
CDI_EXPORT int cdi_assignDefaultDataMaskAndPattern(char *devName,BYTE *dataMask,int *maskRange,BYTE *pattern,int *patternRange,int patternTrigger);
CDI_EXPORT int cdi_assignAllowedAccess(char *devName,int access);
CDI_EXPORT int cdi_eraseDefaultDataMaskAndPattern(char *devName,int erasedataMask, int erasepattern);

CDI_EXPORT int cdi_getDeviceNames(NAME16 *namebuffer, int namesize); /* for all cdi lines  */
CDI_EXPORT int cdi_getDeviceNamesByLine(NAME16 *namebuffer, int namesize, int cdiLine);  /* cdiLine is the cdiLineIndex  */
CDI_EXPORT char *cdi_getBusNameForCdiLine(int cdiLine);
CDI_EXPORT char *cdi_getBusNameForDevice(char *devName);
CDI_EXPORT int cdi_getBusNames(NAME16 *namebuffer,int namesize);
CDI_EXPORT int cdi_getDeviceNumberOfErrors(char *devName);
CDI_EXPORT DADDRESS *cdi_getBusAddressForDevice(char *devName);
CDI_EXPORT char *cdiGetRedirectedDeviceServer(char *devName);
CDI_EXPORT int cdi_getRedirectedDeviceNames(NAME16 *namebuffer,int namesize);
CDI_EXPORT int cdi_RegisterRedirectedDevice(
  char *devName, int devNumber, char *redirectedServerName
  );
CDI_EXPORT int cdi_RegisterBus(
  char *busName,
  int busType,
  int (*filterBusFcn)(int *parameter, int accessFlag),
  int (*cfgBusFcn)(void),
  int (*initBusFcn)(int busline,int cdiline,int ndevices,char *paramList),
  void (*execBusFcn)(CdiRequestInfoBlk *pReq),
  int (*freeBusFcn)(int busline)
  );
CDI_EXPORT int cdiRegisterBusFilter(char *busName,int (*fcn)(int *,int));
CDI_EXPORT int cdiRegisterBusConfiguration(char *busName,int (*fcn)(void));
CDI_EXPORT int cdiRegisterBusInitialization(char *busName,int (*fcn)(int,int,int,char *));
CDI_EXPORT int cdiRegisterBusHandler(char *busName,void (*fcn)(CdiRequestInfoBlk *));
CDI_EXPORT int cdiRegisterBusCleanup(char *busName,int (*fcn)(int));
CDI_EXPORT int cdiRegisterBus(char *busName,int busType);
CDI_EXPORT int cdiScanBus(char *busName,char *target,char *result,int resultBufferSize);
CDI_EXPORT int cdiRegisterBusScanner(char *busName,int (*fcn)(char *,char *,int));
CDI_EXPORT CdiLineInfoBlk *cdiGetBusHandle(int cdiLine);
CDI_EXPORT void cdi_Cleanup(void);
CDI_EXPORT void cdiCloseLinks(void);
CDI_EXPORT int cdi_assignDataLimits(char *devName,int dataSizeOut,int dataSizeIn);
CDI_EXPORT int cdi_assignDefaultInput(char *devName,BYTE *pDataInput,int dataSize,int dataFormatIn);
CDI_EXPORT int cdi_CalibrateRequest(int requestIndex,int applyRules);
CDI_EXPORT int cdi_assignDataFormat(char *devName,int dataFormat);
CDI_EXPORT int cdi_assignDataCalibration(char *devName,CdiRule *clbrRules,CdiRule *rvrsRules,int dataBusFormat);
CDI_EXPORT int getTypeLength(short dataType);
int clbrRequest(CdiRequestInfoBlk *pReq,int applyRules);
CDI_EXPORT int get_RequestBufferPointer(int requestIndex, CdiRequestInfoBlk **pReq);

#define CDI_INFO_ERRORS 1
#define CDI_INFO_NAMES  2
#define CDI_INFO_STATUS 3
#define CDI_INFO_NDEVS  4
#define CDI_INFO_ERRCNT 5

CDI_EXPORT int getTimeMilliseconds(void);
CDI_EXPORT int getGroupRequestInfo(int requestIndex, void *buffers, int index);
CDI_EXPORT int cdi_getTemplateNames(NAME16 *namebuffer,int namesize);
CDI_EXPORT int cdi_getTemplate(char *templateName,NAME16II *cdiTemplate,int *namesize);
CDI_EXPORT int cdi_getDescriptionForDevice(char *devName,char *devDesc);
CDI_EXPORT int signGroupInOutFormatCode(CdiRequestInfoBlk *pReq, int inFormat, int outFormat);
CDI_EXPORT int cdi_getUnitsForDevice(char *devName,char *devUnits);
CDI_EXPORT int cdi_getPrecisionForDevice(char *devName,char *devPrec);
CDI_EXPORT int cdi_getRangeForDevice(char *devName,float *devRange);
CDI_EXPORT int cdi_RegisterCalibrationFunction(char *fcnName,double (*fcn)(double));

#endif /* CDI_H_ */
