#include "prolog.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#include "rpctypes.h"
#include "errors.h"
#include "database.h"
#include "brdcast.h"
#include "servdef.h"
#include "cdiListener.h"
#include "tinecdi.h"

cdiListener **cdiListenerTable = NULL;
int nListeners = 0;
int defaultPollingMode = CM_POLL;

void listenerCb(int id,int cc)
{
  int siz, mlid = id - 1;
  cdiListener *ml = cdiListenerTable[mlid];
  if (cdiDebugLevel) dbglog("listenerCb: entered with id %d cc %d",id,cc);
  if (ml == NULL) 
  {
    if (cdiDebugLevel) dbglog("listenerCb: could not find listener for this id!");
    return;
  }
  ml->timestamp = makeDataTimeStamp();
  siz = ml->dout.dArrayLength*fmtsizeof(ml->dout.dFormat);
  memcpy(ml->rbbuf,ml->outbuf,siz);

  return;
}
int startListenerCompCode = 0;
cdiListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate,int flags)
{
  int i,lid,siz,mode=defaultPollingMode|CM_WAIT;
  cdiListener *ml=NULL;

  if (cdiDebugLevel) dbglog("startListener: starting new listener for %s %s at %d msec",devName,devProperty,pollrate);
  if (cdiListenerTable == NULL)
  {
    if ((cdiListenerTable=(cdiListener **)SystemCalloc(MAX_NUM_LISTENERS,sizeof(cdiListener *))) == NULL)
    {
      cdilog("Cannot allocate listener table !");
      startListenerCompCode = out_of_local_memory;
      return NULL;
    }
  }
  for (i=0; i<MAX_NUM_LISTENERS && cdiListenerTable[i] != NULL; i++);
  if (i == MAX_NUM_LISTENERS) 
  {
    cdilog("listener table full !");
    startListenerCompCode = resources_exhausted;
    return NULL;
  }
  siz = dout->dArrayLength * fmtsizeof(dout->dFormat);
  if ((ml=(cdiListener *)calloc(1,sizeof(cdiListener))) == NULL) 
  {
    cdilog("Cannot allocate new listener !");
    startListenerCompCode = out_of_local_memory;
    return NULL;
  }
  if ((ml->outbuf=(BYTE *)calloc(siz,1)) == NULL) 
  {
    free(ml);
    cdilog("Cannot allocate data buffer for listener !");
    startListenerCompCode = out_of_local_memory;
    return NULL;
  }
  if ((ml->rbbuf=(BYTE *)calloc(siz,1)) == NULL) 
  {
    free(ml->outbuf);
    free(ml);
    cdilog("Cannot allocate data buffer for listener !");
    startListenerCompCode = out_of_local_memory;
    return NULL;
  }
  ml->dout.dArrayLength = dout->dArrayLength;
  ml->dout.dFormat = dout->dFormat;
  memcpy(ml->dout.dTag,dout->dTag,8);
  ml->rbbuf;  /* pointer to original data reference */
  ml->dout.data.vptr = ml->outbuf; /* pointer to working area buffer */
  if (din != NULL && (siz=din->dArrayLength * fmtsizeof(din->dFormat)) > 0)
  {
    if ((ml->inbuf=(BYTE *)calloc(siz,1)) == NULL) 
    {
      if (ml->outbuf != NULL) free(ml->outbuf);
      if (ml->rbbuf != NULL) free(ml->rbbuf);
      free(ml);
      cdilog("Cannot allocate data buffer for listener !");
      startListenerCompCode = out_of_local_memory;
      return NULL;
    }
    memcpy(ml->din.data.vptr,din->data.vptr,siz);
    ml->din.dArrayLength = din->dArrayLength;
    ml->din.dFormat = din->dFormat;
    memcpy(ml->din.dTag,din->dTag,8);
  }
  else
  {
    ml->din.dFormat = CF_NULL;
    ml->din.dArrayLength = 0;
    ml->din.data.vptr = NULL;
    memset(ml->din.dTag,0,8);
  }
  strncpy(ml->devName,devName,64);
  strncpy(ml->devProperty,devProperty,32);

  ml->mlid = i;
  cdiListenerTable[i] = ml;

  if (flags > 0) mode |= flags;
  lid = AttachLinkEx(devName,devProperty,&ml->dout,&ml->din,CA_READ,pollrate,listenerCb,mode,ml->mlid+1);

  if (lid < 0)
  {
    if (ml->rbbuf != NULL) free(ml->rbbuf);
    if (ml->outbuf != NULL) free(ml->outbuf);
    if (ml->inbuf != NULL) free(ml->inbuf);
    free(ml);
    cdiListenerTable[i] = NULL;
    cdilog("startListener for %s %s failed : %s",devName,devProperty,erlst[-lid]);
    startListenerCompCode = -lid;
    return NULL;
  }
  if (ml->dout.dArrayLength != 0)
  {
    memcpy(dout->data.vptr,ml->outbuf,dout->dArrayLength*fmtsizeof(dout->dFormat));
  }
  ml->pollrate = pollrate;
  ml->lid = lid;
  nCdiListeners++;
  startListenerCompCode = 0;
  if (cdiDebugLevel) dbglog("startListener: assign listener %d (link %d) for %s %s",i,lid,devName,devProperty);
  return ml;
}
int stopListener(cdiListener *ml)
{
  int mlid;
  cdiListener *m = NULL;
  if (ml == NULL) return non_existent_elem;
  if (cdiDebugLevel) dbglog("stopListener: stop listener %d (link %d)",ml->mlid,ml->lid);
  Cancel(ml->lid);
  mlid = ml->mlid;
  if (cdiListenerTable[mlid] == NULL) return not_running;
  if (ml->rbbuf != NULL) free(ml->rbbuf);
  if (ml->outbuf != NULL) free(ml->outbuf);
  if (ml->inbuf != NULL) free(ml->inbuf);
  free(ml);
  cdiListenerTable[mlid] = NULL;
  nCdiListeners--;
  return 0;
}
int nCdiListeners = 0;
cdiListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din)
{
  cdiListener *ml=NULL;
  UINT32 dsiz;
  int i;

  if (cdiListenerTable == NULL) return NULL;
  if (cdiDebugLevel) dbglog("getListener: looking for %s %s among %d registered listeners",devName,devProperty,nCdiListeners);
  for (i=0; i<nCdiListeners; i++)
  {
    if ((ml=cdiListenerTable[i]) == NULL) continue;
    if (strncmp(devName,ml->devName,64)) continue;
    if (strlen(devProperty) > 0 && strncmp(devProperty,ml->devProperty,32)) continue;
    if (dout != NULL)
    {
      if (dout->dFormat != ml->dout.dFormat) continue;
      if (dout->dArrayLength != ml->dout.dArrayLength) continue;
    }
    if (din != NULL && ml->din.dArrayLength != 0)
    {
      if (din->dFormat != ml->din.dFormat) continue;
      if (din->dArrayLength != ml->din.dArrayLength) continue;
      dsiz = ml->din.dArrayLength*fmtsizeof(ml->din.dFormat);
      if (memcmp(din->data.vptr,ml->inbuf,dsiz)) continue;
    }
    if (cdiDebugLevel) dbglog("getListener: found listener (%d of %d)",i,nCdiListeners);
    break;
  }
  return i < nCdiListeners ? ml : NULL;  
}
int cdiDumpListeners(int i1,int i2,int i3,int i4)
{
  cdiListener *ml=NULL;
  int i;
  i1 = i1; i2 = i2; i3 = i3; i4 = i4;
  dbglog("Currently running CDI Listeners :\n");
  if (cdiListenerTable == NULL) return 0;
  for (i=0; i<nCdiListeners; i++)
  {
    if ((ml=cdiListenerTable[i]) == NULL) continue;
    dbglog("%s %s @ %d ms : last access %10.3f <%d>",ml->devName,ml->devProperty,ml->pollrate,ml->timestamp,ml->status);
  }
  return 0;
}
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,double *dTime,int rate,int flags)
{
  cdiListener *ml;
  UINT32 siz;
  struct timespec ts;

  if (devName == NULL || devProperty == NULL || dout == NULL) 
  {
    if (cdiDebugLevel) dbglog("getDataFromListener: device %s property %s data %s",
      devName != NULL ? devName : "null", devProperty != NULL ? devProperty : "null", dout != NULL ? "okay" : "null");
    return argument_list_error;
  }
  if (cdiDebugLevel) dbglog("getDataFromListener: %s %s",devName,devProperty);
  if ( (ml=getListener(devName,devProperty,dout,din)) == NULL && 
       (ml=startListener(devName,devProperty,dout,din,rate,flags)) == NULL )
  {
    if (cdiDebugLevel) dbglog("getDataFromListener: startListener returns %s",erlst[startListenerCompCode]);
    return startListenerCompCode;
  }
  if (dout->dFormat != ml->dout.dFormat) 
  {
    if (cdiDebugLevel) dbglog("getDataFromListener: calling format %d not equal to listener format %d",
      dout->dFormat,ml->dout.dFormat);
    return illegal_format;
  }
  if (dout->data.vptr == NULL) 
  {
    if (cdiDebugLevel) dbglog("getDataFromListener: data pointer is null !");
    return argument_list_error;
  }
  if (flags & CM_WAIT)
  {
    if (cdiDebugLevel) dbglog("wait for timestamp!");
    while (ml->timestamp == 0)
    {
      ts.tv_sec = 0;
      ts.tv_nsec = 100000;
      nanosleep(&ts,NULL);
    }
  }
  dout->dTimeStamp = ml->timestamp;
  siz = MIN(dout->dArrayLength,ml->dout.dArrayLength)*fmtsizeof(dout->dFormat);
  memcpy(dout->data.vptr,ml->rbbuf,siz);
  if (dTime != NULL) *dTime = dout->dTimeStamp;
  if (cdiDebugLevel) dbglog("getDataFromListener: returning status %d (timestamp %g)",ml->status,ml->timestamp);
  return ml->status;
}
