#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#ifdef UNIX
#include <dlfcn.h>
#include "unixcmd.h"
#endif
#ifdef WINNT
#include "ntcmd.h"
#endif
#include "errors.h"
#include "rpctypes.h"
#include "database.h"
#include "swaplib.h"
#include "servdef.h"
#include "almsrv.h"
#include "toolkit.h"
#include "cdi.h"
#include "tinecdi.h"
#include "cdiListener.h"

int cdiDebugLevel = 0;
int cdiStandAlone = FALSE;
static BYTE cdiCurrentByteMask[8];
static int cdiHasByteMask = 0;
char CdiDBpath[132];
int errListMemPoolSize = MAX_ERROR_LIST_SIZE;
int *errListMemPool = NULL;
NAME16 *nameListMemPool = NULL;
int gMaxTemplateLength = 1;
char gCdiExportName[32];
char gCdiExportContext[32];
char gCdiExportPrefix[96];
int gCdiExportPrefixLength = 0;
int isCdiStandAlone(void)
{
  char *ptr = getenv("CDI_STANDALONE");
  if (ptr == NULL || strnicmp(ptr,"TRUE",4)) return FALSE;
  cdilog("using cdi in standalone mode");
  return TRUE;
}
int cdiSystemInitialization(void)
{
  static int cc = database_not_loaded;
  static int initDone = FALSE;
  static int inside = FALSE;
  if (initDone) return cc;
  if (inside) return operation_busy;
  inside = TRUE;
  if ((cc=readCdiManifest()) == 0) 
  {
    cdiStandAlone = isCdiStandAlone();
    if (!cdiStandAlone) initTineCDI();
    readCdiDataBase();
  }
  initDone = TRUE;
  inside = FALSE;
  return cc;
}  
CDI_EXPORT void rotateLogFile(void)
{
  static int done = FALSE;
  static char fn[148] = "", fb[148] = "";

  if (done) return;
  done = TRUE;
  if (strlen(fn) == 0) sprintf(fn,"%scdi.log",CdiDBpath);
  if (strlen(fb) == 0) sprintf(fb,"%scdi.bak",CdiDBpath);
  remove(fb);
  rename(fn,fb);
  return;
}
CDI_EXPORT void cdilog(char *text,...)
{
  double ts = makeDataTimeStamp();
  FILE *fp = NULL, *fpb = NULL;
  long flen;
  char s[512];
  static char fn[148] = "";
  va_list args;

  if (strlen(fn) == 0) sprintf(fn,"%scdi.log",CdiDBpath);
  if ((fp=fopen(fn,"a")) == NULL)
  {
    perror("cdi: fopen");
    goto error;
  }
  va_start(args,text);
  sprintf(s,"%s : ",getDataTimeString(ts,0));
  vsprintf(&s[strlen(s)],text,args);
  va_end(args);

  if (s[strlen(s)-1] != '\n') strcat(s,"\n");
  fprintf(fp,s);
 
  fseek(fp, 0L, SEEK_END);
  flen = ftell(fp);

  if (flen > 64000)
  {
    fclose(fp); fp = NULL;
    rename("cdi.log","cdi.bak");
  }
  if (cdiDebugLevel) dbglog(s);
error:
  if (fp) fclose(fp);
  return;
}

CDI_EXPORT void cdiSetDataByteMask(void *dataMask,int sizeInBytes)
{
  memcpy(cdiCurrentByteMask,dataMask,MIN(8,sizeInBytes));
  cdiHasByteMask = TRUE;
}
CDI_EXPORT void *cdiGetDataByteMask(void)
{
  void *bmptr = NULL;
  if (cdiHasByteMask) bmptr = cdiCurrentByteMask; 
  cdiHasByteMask = 0; /* always clear */
  return bmptr;
}
CDI_EXPORT int cdiCloseLink(int id)
{
  if (cdi_isValidRequestIndex(id)) 
  {
    cdi_CancelRequest(id);
    return 0;
  }
  return invalid_index;
}
CDI_EXPORT void cdiCloseLinks(void)
{
  int i, j;

  cdilog("stopping all active links");
  for (i=0; i<numConnectedHandles; i++)
  {
    if ( busHandle[i] == NULL ) continue;
    for (j=0; j<busHandle[i]->numberActiveRequest; j++)
      cdi_CancelRequest(busHandle[i]->pReqList[j].requestNumber + 0x10000);
  }
  return;

}

CDI_EXPORT int cdiRegisterDevice(
  char *devName, int devNumber, char *lngName, char *devDesc,
  int cdiLine, int cdiIndex, int address, int subAddress,int *devParameters
)
{
  int cc;
  if ((cc=cdi_RegisterDevice(cdiLine,devName,lngName,devDesc,cdiIndex,devNumber,address,subAddress)) != 0) return cc;
  return devParameters == NULL ? 0 : cdi_assignDefaultDeviceParameters(devName,devParameters);
}

CDI_EXPORT int cdiRegisterDeviceNames(void)
{
  int i, ndevs, nredir;
  NAME16 *devs = NULL;
  char dname[32];
  ndevs = cdi_getDeviceNames(NULL,0); /* first get the number of devices */
  if (ndevs > 0)
  {
    if ((devs=(NAME16 *)alloca(ndevs*16)) == NULL) return out_of_local_memory;
    cdi_getDeviceNames(devs,ndevs);
    for (i=0; i<ndevs; i++) 
    {
      strncpy(dname,devs[i].name,16); dname[16] = 0;
      RegisterDeviceName(CDIEQM_TAG,dname,i);
    }
  }
  nredir = cdi_getRedirectedDeviceNames(NULL,0);
  if (nredir == 0) return 0;
  if (nredir > ndevs)
  {
    if ((devs=(NAME16 *)alloca(nredir*16)) == NULL) return out_of_local_memory;
  }
  cdi_getRedirectedDeviceNames(devs,nredir);
  for (i=0; i<nredir; i++)
  {
    strncpy(dname,devs[i].name,16); dname[16] = 0;
    RegisterDeviceName(CDIEQM_TAG,dname,i+ndevs);
  }
  return 0;
}

ExtCdiProperty *extCdiPrps;
int gNumExtCdiProperties = 0;

int splitCdiDeviceName(char *cdiname,char *dev,char *prp)
{
  char cname[32], *c, *tok = ".";
  if (cdiname == NULL || dev == NULL || prp == NULL) return argument_list_error;
  strncpy(cname,cdiname,16); cname[16] = 0;
  if ((c=strstr(cname,tok)) == NULL)
  { /* not splittable */
    strncpy(prp,cname,16); prp[16] = 0;
    strncpy(dev,cname,16);
  }
  else
  {
    *c++ = 0;
    strncpy(prp,c,16); prp[16] = 0;
    strncpy(dev,cname,16);
  }
  return 0;
}
CDI_EXPORT int cdiRegisterExtendedProperties(void)
{
  int i, k, n, ndevs, nxprps = 0, prpid, cc = 0;
  int cdinr_last, cdinr_this, ln, ix, iscont = TRUE;
  NAME16 *devs = NULL, *pdevs = NULL;
  char dname[32],pname[32],tname[32];
  DTYPE dout, din;
  ndevs = cdi_getDeviceNames(NULL,0); /* first get the number of devices */
  if (ndevs > 0)
  {
    if ((extCdiPrps=(ExtCdiProperty *)calloc(ndevs,sizeof(ExtCdiProperty))) == NULL)
      ccerr(out_of_local_memory);
    if ((devs=(NAME16 *)alloca(ndevs*16)) == NULL) ccerr(out_of_local_memory);
    if ((pdevs=(NAME16 *)alloca(ndevs*16)) == NULL) ccerr(out_of_local_memory);
    cdi_getDeviceNames(devs,ndevs);
    for (i=0; i<ndevs; i++) 
    {
      splitCdiDeviceName(devs[i].name,dname,pname);
      /* is property registered ? */
      if ((prpid=GetPropertyId(CDIEQM_TAG,pname)) < 0)
      { /* find them all : */
        strncpy(pdevs[0].name,dname,16);
        cdinr_last = findDevice(devs[i].name,&ln,&ix);
        for (n=1,k=i+1; k<ndevs; k++)
        {
          splitCdiDeviceName(devs[k].name,dname,tname);
          if (strncmp(pname,tname,32)) continue;
          cdinr_this = findDevice(devs[k].name,&ln,&ix);
          if (cdinr_this != cdinr_last + 1) iscont = FALSE;
          cdinr_last = cdinr_this;
          strncpy(pdevs[n].name,dname,16);
          n++;
        }
        strncpy(extCdiPrps[nxprps].prp,pname,32);
        extCdiPrps[nxprps].ndevs = n;
        extCdiPrps[nxprps].isContiguous = iscont;
        extCdiPrps[nxprps].fmt = CF_FLOAT;
        dout.dFormat = CF_FLOAT;
        dout.dArrayLength = n;
        din.dFormat = CF_FLOAT;
        din.dArrayLength = 1;
        extCdiPrps[nxprps].id = RegisterPropertyInformation(CDIEQM_TAG,pname,&dout,&din,
                                       CA_READ|CA_WRITE,AT_CHANNEL,(UINT16)n,"extended property",
                                       PRP_EXTENDED_OFFSET+nxprps,NULL);

        AssignDeviceListToProperty(CDIEQM_TAG,pname,pdevs,n);
        if ((extCdiPrps[nxprps].devs=(NAME16 *)malloc(n*sizeof(NAME16))) == NULL) ccerr(out_of_local_memory);
        memcpy(extCdiPrps[nxprps].devs,pdevs,n*sizeof(NAME16));
        nxprps++;
      }
    }
  }
  gNumExtCdiProperties = nxprps;
err:
  return cc ;
}

CDI_EXPORT char *cdiGetFullDeviceName(char *devName)
{
  int cdiLine,cdiIndex;
  char *c = cdiGetRedirectedDeviceServer(devName);
  static char localDeviceServerName[96];
  if (c != NULL) return c;
  if (findDevice(devName,&cdiLine,&cdiIndex))
  {
    sprintf(localDeviceServerName,"/%s/%s.CDI/%s",gDeviceContext,gFecName,devName);
    return localDeviceServerName;
  }
  return NULL;
}
NAME16 *cdiBusNames = NULL;
int nCdiBusNames = 0;
char **cdiHardwareTypes;
int nCdiHardwareTypes = 0;
CDI_EXPORT NAME16 *cdiGetBusNames(void)
{
  int i;
  if (cdiBusNames == NULL)
  {
    nCdiBusNames = cdi_getBusNames(NULL,0);
    if ((cdiBusNames=(NAME16 *)calloc(nCdiBusNames,16)) == NULL)
    {
      cdilog("cdiGetBusNames : out of memory!");
      return NULL;
    }
    if ((cdiHardwareTypes=(char **)malloc((nCdiBusNames+1)*sizeof(char *))) == NULL)
    {
      cdilog("cdiGetBusNames : out of memory!");
      return NULL;
    }
    nCdiHardwareTypes = nCdiBusNames + 1;
    cdi_getBusNames(cdiBusNames,nCdiBusNames);
    cdiHardwareTypes[0] = "CDI"; /* zeroth position is general ! */
    for (i=0; i<nCdiBusNames; i++)
    {
      cdiHardwareTypes[i+1] = cdiBusNames[i].name;
    }
  }
  return cdiBusNames;
}
CDI_EXPORT char *getCdiLinkName(char *devName,char *cdiDevName)
{
  int i;
  char *lclDevName;
  if (devName == NULL || strlen(devName) < 8) return NULL;
  cdiGetBusNames(); /* make sure it's filled in ... */
  if (!strnicmp(&devName[1],"LOCAL",5) && (lclDevName=strchr(&devName[1],'/')) != NULL)
  {
    lclDevName++;
    for (i=0; i<nCdiHardwareTypes; i++)
    {
      if (!strnicmp(lclDevName,cdiHardwareTypes[i],strlen(cdiHardwareTypes[i])))
      {
        strncpy(cdiDevName,&lclDevName[strlen(cdiHardwareTypes[i])+1],256);
        return cdiDevName; /* return the pointer given */
      }
    }
  }
  if (!strnicmp(gCdiExportPrefix,devName,gCdiExportPrefixLength))
  {
    strncpy(cdiDevName,&devName[gCdiExportPrefixLength],256);
    return cdiDevName; /* return the pointer given */
  }
  return NULL;
}
#define cdiParsePropertyForAccess(prp) (\
  !strcmp(prp,"RECV") ? cdiReadFlag :\
  !strcmp(prp,"SEND") ? cdiWriteFlag :\
  !strcmp(prp,"RECV.SEND.ATOM") ? cdiReadWriteFlag :\
  !strcmp(prp,"SEND.RECV.ATOM") ? cdiWriteReadFlag :\
  !strcmp(prp,"SEND.RECV") ? cdiWriteReadFlag :\
  !strcmp(prp,"SEND.SEND.ATOM") ? cdiWriteWriteFlag :\
  !strcmp(prp,"SEND.RECV.SEND.ATOM") ? cdiWriteReadWriteFlag : -1\
  )

short getBaseFormat(short tineFormat,int *needsErrList,int *needsNameList)
{
  *needsNameList = FALSE;
  *needsErrList = FALSE;
  switch (tineFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL: *needsNameList = TRUE;
    case CF_DBLDBL: *needsErrList = TRUE;
    case CF_DOUBLE: return CF_DOUBLE;
    case CF_NAME16FI: *needsNameList = TRUE;
    case CF_FLTINT: *needsErrList = TRUE;
    case CF_FLOAT:  return CF_FLOAT;
    case CF_UINT16: return CF_UINT16;
    case CF_NAME32I: *needsErrList = TRUE;
    case CF_NAME32: return CF_NAME32;
    case CF_NAME16II: *needsNameList = TRUE;
    case CF_LNGINT: *needsErrList = TRUE;
    case CF_UINT32: return CF_UINT32;
    case CF_UINT8:  return CF_UINT8;
    case CF_TEXT:   return CF_TEXT;
    default: break;
  }
  return -1;
}

short getCdiBaseFormat(short tineFormat)
{
  switch (tineFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
    case CF_DBLDBL:
    case CF_DOUBLE: return cdiDouble;
    case CF_NAME16FI:
    case CF_FLTINT:
    case CF_FLOAT:  return cdiFloat;
    case CF_UINT16: return cdiShort;
    case CF_NAME16II:
    case CF_LNGINT:
    case CF_UINT32: return cdiInteger;
    case CF_UINT8:  return cdiByte;
    case CF_TEXT:   return cdiChar;
    case CF_NAME32I:
    case CF_NAME32: return cdiChar32;
    default: break;
  }
  return -1;
}

CDI_EXPORT short getCdiFormat(short tineFormat)
{
  switch (tineFormat)
  {
    case CF_DOUBLE: return cdiDouble;
    case CF_FLOAT:  return cdiFloat;
    case CF_UINT16: return cdiShort;
    case CF_UINT32: return cdiInteger;
    case CF_UINT8:  return cdiByte;
    case CF_TEXT:   return cdiChar;
    case CF_NAME32: return cdiChar32;
    default: break;
  }
  return -1;
}
CDI_EXPORT int getCdiAllowedAccess(char *accessString)
{
  char access[256], *s=NULL, *c=NULL, *tok = " +|";
  int allowedAccess = 0;
  if (accessString == NULL || strlen(accessString) == 0) return 0xffff;
  strncpy(access,accessString,255); access[255] = 0;
  strupr(access);
  for (s=strtok_r(access,tok,&c); s != NULL; s=strtok_r(NULL,tok,&c))
  {
    if (!strcmp(s,"READ") || !strcmp(s,"RD")) allowedAccess |= 0x01;
    if (!strcmp(s,"WRITE") || !strcmp(s,"WR")) allowedAccess |= 0x02;
    if (!strcmp(s,"READWRITE") || !strcmp(s,"RDWR")) allowedAccess |= 0x04;
    if (!strcmp(s,"WRITEREAD") || !strcmp(s,"WRRD")) allowedAccess |= 0x08;
    if (!strcmp(s,"WRITEWRITE") || !strcmp(s,"WRWR")) allowedAccess |= 0x10;
    if (!strcmp(s,"WRITEREADWRITE") || !strcmp(s,"WRRDWR")) allowedAccess |= 0x20;
  }
  if ((allowedAccess & 0x03) == 0x03) allowedAccess = 0xffff; /* READ & WRITE separately => everything's okay */
  return allowedAccess;
}
int cdiRemapDataSets(DTYPE *dout,BYTE *dataout,int *errLst,NAME16 *nameLst)
{
  int i, namlen = 0, fmtsiz = CF_NULL;
  int *errout = NULL;
  NAME16 *nameout = NULL;
  double dStatus;
  BYTE *dataptr = NULL;
char buf[64];

  if (dout == NULL) return argument_list_error;
  fmtsiz = fmtsizeof(dout->dFormat);
  dataptr = dout->data.bptr;

  if (errListMemPoolSize < (int)dout->dArrayLength) return resources_exhausted;
  if (errListMemPool == NULL && (errListMemPool=(int *)calloc(errListMemPoolSize,sizeof(int))) == NULL)
    return out_of_local_memory;
  if (nameListMemPool == NULL && (nameListMemPool=(NAME16 *)calloc(errListMemPoolSize,sizeof(NAME16))) == NULL)
    return out_of_local_memory;

  /* copy error array out to the errList memory pool (must remap e.g. INTINT structs into 2 arrays) */
  if (errLst != NULL)
  {
    errout = errListMemPool;
    memcpy(errout,errLst,dout->dArrayLength*sizeof(int));
  }
  if (nameLst != NULL)
  {
    nameout = nameListMemPool;
    memcpy(nameout,nameLst,dout->dArrayLength*sizeof(NAME16));
  }
  switch (dout->dFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      if (nameout == NULL) return argument_list_error;
      namlen = fmtsiz - 16;
    case CF_DBLDBL:
      if (errout == NULL) return argument_list_error;
      for (i=dout->dArrayLength-1; i >= 0;i--)
      { /* dataptr = dataout => traverse the list in reverse */
        dStatus = (double)errout[i];        
        memcpy(dataptr + i*fmtsiz + namlen + 8,&dStatus,8);  /* status */
        memcpy(dataptr + i*fmtsiz + namlen,dataout + i*8,8); /* data */
        if (namlen > 0) memcpy(dataptr + i*fmtsiz,&nameout[i],16); /* name */
      }
    case CF_NAME16FI:
    case CF_NAME16II:
      if (nameout == NULL) return argument_list_error;
      namlen = 16;
    case CF_FLTINT:
    case CF_LNGINT:
      if (errout == NULL) return argument_list_error;
      for (i=dout->dArrayLength-1; i >= 0;i--)
      { /* dataptr = dataout => traverse the list in reverse */
        memcpy(dataptr + i*fmtsiz + namlen + 4,&errout[i],4);  /* status */
        memcpy(dataptr + i*fmtsiz + namlen,dataout + i*4,4); /* data */
        if (namlen > 0) memcpy(dataptr + i*fmtsiz,&nameout[i],16); /* name */
      }
      break;
    case CF_NAME32I:
      if (errout == NULL) return argument_list_error;
      for (i=dout->dArrayLength-1; i >= 0;i--)
      { /* dataptr = dataout => traverse the list in reverse */
        memcpy(dataptr + i*fmtsiz + 32,&errout[i],4);  /* status */
        memcpy(buf,dataout + i*32,32);
        /* memcpy(dataptr + i*fmtsiz,dataout + i*32,32); < this doesn't compile correctly on VC 6.0 @ 8byte boundaries */
        memcpy(dataptr + i*fmtsiz,buf,32); /* data */
      }
      break;
    default:
      break;
  }
  return 0;
}
/*
 * synchronous call :
 * e.g. ExecLinkEx("/LOCALHOST/CDI/CanTestDevNumber_4/CanParamSet_0","WRITE",NULL,&din,CA_WRITE,3000);
 * context LOCAL => either cdiExecLinkEx() or execLocalLink()
 * if group == CDI, or explicitly CAN, SEDAC, etc. => jump to cdiExecLinkEx()
 */
CDI_EXPORT int cdiExecLinkEx(
  char *devName,     /* device name */
  char *devProperty, 
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  UINT16 timeout     /* timeout in msec. */
)
{
  char *c=NULL,fulname[4096],devPropertyStr[64];
  char *dn = NULL,*pn = NULL;
  int *cdiParameters = NULL;
  void *cdiDataMask = NULL;
  int cc = 0, id, rc, cdiAccessFlag, cdiReturnCondition = 0, cdiPollingFlag = 0, cdiNeedsCalibration = 0;
  UINT32 doutlen=0, dinlen=0, doutlenInBytes=0;
  void *doutptr = NULL, *dinptr = NULL;
  short dinFormat=0, doutFormat=0, baseFormat = CF_NULL;
  int hasErrLst = FALSE, hasNameLst = FALSE;
  int *errLst = NULL;
  NAME16 *nameLst = NULL;
  CdiRequestInfoBlk *pReq;

  if (cdiDebugLevel) dbglog("cdiExecLinkEx: %s %s with timeout %d",devName,devProperty,timeout);
  if (devName == NULL || strlen(devName) == 0) ccerr(argument_list_error);
  if (devProperty == NULL || strlen(devProperty) == 0) ccerr(argument_list_error);
  strncpy(devPropertyStr,devProperty,32); devPropertyStr[32] = 0;
  /* parse devName for hdw device name and hdw parameter set */
  strncpy(fulname,devName,4095); fulname[4095] = 0;
  remapCdiDeviceAndProperty(fulname,devPropertyStr,dout,din,access);
  if ((dn=strtok_r(fulname,DELIMITER,&c)) == NULL) ccerr(argument_list_error);
  if ((pn=strtok_r(NULL,"<",&c)) != NULL)
  { 
    if ((c=strchr(pn,'>')) != NULL) *c = 0;
    if ((cdiParameters=cdi_GetParameters(pn)) == NULL) cdilog("input parameter list not found");
  }
  if (dout != NULL && dout->dArrayLength > 0) 
  {
    baseFormat = getBaseFormat(dout->dFormat,&hasErrLst,&hasNameLst);
    doutlen = dout->dArrayLength;
    doutlenInBytes = dout->dArrayLength*fmtsizeof(baseFormat);
    doutptr = dout->data.vptr;
    doutFormat = getCdiFormat(baseFormat);
    if ( doutFormat < 0 ) ccerr(illegal_format)
    if (hasErrLst) 
    {
      errLst = (int *)((int *)doutptr + doutlen);
      if (dout->dFormat == CF_NAME32I) errLst = (int *)((BYTE *)doutptr + doutlen * 32);
    }
  }
  if (din != NULL && din->dArrayLength > 0) 
  {
    /* does SEND call need reverse calibration ? */
    if ((c=strstr(devPropertyStr,"SEND.CLBR")) != NULL)
    {
      cdiReturnCondition |= INPUT_CALIBRATED;
      *c = 0;
      strcat(devPropertyStr,"SEND");
    }
    dinlen = din->dArrayLength*fmtsizeof(din->dFormat);
    dinptr = din->data.vptr;
    dinFormat = getCdiFormat(din->dFormat);
    if ( dinFormat < 0 ) ccerr(illegal_format)
  }
  /* is it a calibration property ? */
  if ((c=strstr(devPropertyStr,".CLBR")) != NULL)
  {
    cdiNeedsCalibration = TRUE;
    *c = 0;
  }
  /* has dataMask been set ? */
  /* cdi_applyDataByteMask()  */
  cdiDataMask = cdiGetDataByteMask();
  cdiAccessFlag = cdiParsePropertyForAccess(devPropertyStr);
  if (cdiAccessFlag < 0) ccerr(illegal_property);
  if (cdiAccessFlag > cdiReadFlag && (access & CA_WRITE) == 0) ccerr(illegal_read_write);
  /* synchronous link will not look for pattern or tolerance ! */
  /* returnCondition = maskChange since there's no callback */
  id = cdi_Request(dn,doutptr,doutFormat,doutlenInBytes,dinptr,dinFormat,dinlen,
                   NULL,cdiDataMask,cdiParameters,cdiAccessFlag,cdiReturnCondition,timeout,NULL,cdiPollingFlag,errLst,-1, 0);
  if (id < 0) return -id;
  if ((rc=get_RequestBufferPointer(id, &pReq)) < 0 ) return -rc;
  rc = signGroupInOutFormatCode(pReq, din != NULL ? din->dFormat : CF_NULL, dout != NULL ? dout->dFormat : CF_NULL);

  rc = cdi_StartRequest(id);
  if (rc < 0 && doutlen == 1) cc = hardware_error;
  if (cc == 0 && doutlen > 0)
  {
    rc = cdi_CalibrateRequest(id,cdiNeedsCalibration);
    if (rc < 0) cc = -rc;
  }
  if (hasNameLst && doutptr != NULL)
  {
    nameLst = (NAME16 *)((int *)doutptr + doutlen*2);
    cdi_GetRequestDeviceNames(id,nameLst);
  }
  cdi_CancelRequest(id);
  cdiRemapDataSets(dout,(BYTE *)doutptr,errLst,nameLst);
err:
  if (cdiDebugLevel) dbglog("cdiExecLinkEx: finished with completion code %d",cc);
  return cc;
}

CDI_EXPORT int cdiExecLinkHookFcn(
  char *devName,     /* device name */
  char *devProperty, 
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  UINT16 timeout     /* timeout in msec. */
)
{
  char *dn;
  int cc = cdiExecLinkEx(devName,devProperty,dout,din,access,timeout);
  if (cc == server_redirection)
  {
    if ((dn=cdiGetFullDeviceName(devName)) == NULL) return non_existent_fec;
    strncpy(devName,dn,512);
  }
  return cc;
}
/* asynchronous call:
 * e.g. AttachLinkEx("/LOCALHOST/CDI/CanTestDevNumber_4/CanParamSet_0","WRITE",&dout,&din,CA_READ,200,cbCan,CM_REFRESH,id);
 * context LOCAL => cdiAttachLinkEx()
 * if group == CDI, or explicitly CAN, SEDAC, etc. => jump to cdiAttachLinkEx()
 */
CDI_EXPORT int cdiAttachLinkEx(
  char *devName,     /* device name */
  char *devProperty, 
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  int pollingRate,   /* polling rate in msec. */
  void (*callback)(int,int), /* callback routine */
  int mode,          /* polling mode (SINGLE, POLL, REFRESH, etc.) */
  UINT32 callbackID  /* callback ID */
)
{
  char *c=NULL;
  char fulname[4096],devPropertyStr[64];
  char *dn = NULL,*pn = NULL;
  int *cdiParameters = NULL;
  void *cdiDataMask = NULL;
  void *cdiDataPatternWindow = NULL;
  int cc = 0, cdiAccessFlag, cdiReturnCondition = -1, cdiPollingFlag = 0, cdiNeedsCalibration = 0, rc;
  UINT32 doutlen=0, dinlen=0, doutlenInBytes=0;
  void *doutptr = NULL, *dinptr = NULL;
  short dinFormat=0, doutFormat=0, baseFormat = CF_NULL;
  int hasErrLst = FALSE, hasNameLst = FALSE, isGrouped = FALSE, isWait = FALSE;
  int *errLst = NULL;
  CdiRequestInfoBlk *pReq;

  if (devName == NULL || strlen(devName) == 0) ccerr(argument_list_error);
  if (devProperty == NULL || strlen(devProperty) == 0) ccerr(argument_list_error);
  strncpy(devPropertyStr,devProperty,32); devPropertyStr[32] = 0;
  /* parse devName for hdw device name and hdw parameter set */
  strncpy(fulname,devName,4095); fulname[4095] = 0;
  remapCdiDeviceAndProperty(fulname,devPropertyStr,dout,din,access);
  if ((dn=strtok_r(fulname,DELIMITER,&c)) == NULL) ccerr(argument_list_error);
  if ((pn=strtok_r(NULL,"<",&c)) != NULL)
  { 
    if ((c=strchr(pn,'>')) != NULL) *c = 0;
    if ((cdiParameters=cdi_GetParameters(pn)) == NULL) cdilog("input parameter list not found");
  }
  if (dout != NULL && dout->dArrayLength > 0) 
  {
    baseFormat = getBaseFormat(dout->dFormat,&hasErrLst,&hasNameLst);
    doutlen = dout->dArrayLength;
    doutlenInBytes = dout->dArrayLength*fmtsizeof(baseFormat);
    doutptr = dout->data.vptr;
    doutFormat = getCdiFormat(baseFormat);
    if ( doutFormat < 0 ) ccerr(illegal_format)
    if (hasErrLst) 
    {
      errLst = (int *)((int *)doutptr + doutlen);
    }
  }
  if (din != NULL && din->dArrayLength > 0) 
  {
    /* does SEND call need reverse calibration ? */
    dinlen = din->dArrayLength*fmtsizeof(din->dFormat);
    dinptr = din->data.vptr;
    dinFormat = getCdiFormat(din->dFormat);
    if ( dinFormat < 0 ) ccerr(illegal_format)
  }
  /* is it a calibration property ? */
  if ((c=strstr(devPropertyStr,".CLBR")) != NULL)
  {
    cdiNeedsCalibration = TRUE;
    *c = 0;
  }
  if (mode & CM_GROUPED) isGrouped = TRUE;
  if (mode & CM_WAIT) isWait = TRUE;
  /* has dataMask been set ? */
  /* cdi_applyDataByteMask()  */
  cdiDataMask = cdiGetDataByteMask();
  /* is there a data pattern or tolerance ? */
  if (mode & CM_SEEKPATTERN) cdiReturnCondition = maskCompEqual;
  else if (mode & CM_MISSPATTERN) cdiReturnCondition = maskCompNotEqual;
  else if (mode & CM_NODUPLICATES) cdiReturnCondition = maskChangeWindow;
  if (cdiReturnCondition < 0)
  {
    cdiReturnCondition = 0;
  }
  else
  {
    if ((cdiDataPatternWindow=calloc(doutlenInBytes,1)) == NULL) ccerr(out_of_local_memory);
    memcpy(cdiDataPatternWindow,dout->data.vptr,doutlenInBytes);
  }
  /* accessFlag */
  cdiAccessFlag = cdiParsePropertyForAccess(devPropertyStr);
  if (cdiAccessFlag < 0) ccerr(illegal_property);
  if (cdiAccessFlag > cdiReadFlag && (access & CA_WRITE) == 0) ccerr(illegal_read_write);
  if (BASEMODE(mode) > CM_SINGLE) cdiPollingFlag = 1;
  if (cdiNeedsCalibration) cdiReturnCondition |= RETURN_CALIBRATED;
  if (hasErrLst) cdiReturnCondition |= RETURN_STATUS;
  if (hasNameLst) cdiReturnCondition |= RETURN_NAMES;
  /* returned handle includes high word! */
  /* how to use the user's preferred callbackID ? */
  cc = cdi_Request(dn,doutptr,doutFormat,doutlenInBytes,dinptr,dinFormat,dinlen,
                   cdiDataPatternWindow,cdiDataMask,cdiParameters,cdiAccessFlag,
                   cdiReturnCondition,pollingRate,callback,cdiPollingFlag,errLst,callbackID, 0);
  if (cc < 0) ccerr(cc);
  if ((rc=get_RequestBufferPointer(cc, &pReq)) < 0 ) ccerr(rc);
  rc = signGroupInOutFormatCode(pReq, din->dFormat, dout->dFormat);
  cdi_StartRequest(cc);
err:
  return cc;  /* cc > 0 => id */
}
CDI_EXPORT int cdiAttachLinkHookFcn(
  char *devName,     /* device name */
  char *devProperty, 
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  int pollingRate,   /* polling rate in msec. */
  void (*callback)(int,int), /* callback routine */
  int mode,          /* polling mode (SINGLE, POLL, REFRESH, etc.) */
  UINT32 callbackID  /* callback ID */
)
{
  char *dn;
  int cc = cdiAttachLinkEx(devName,devProperty,dout,din,access,pollingRate,callback,mode,callbackID);
  if (cc == -server_redirection)
  {
    if ((dn=cdiGetFullDeviceName(devName)) == NULL) return -non_existent_fec;
    strncpy(devName,dn,512);
  }
  return cc;
}
CDI_EXPORT void AssignCdiSystemInitHook(void (*hook)(void))
{
  cdiSystemInitHook = hook;
}
CDI_EXPORT void AssignGetCdiLinkNameHook(char *(*hook)(char *,char *))
{
  getCdiLinkNameHook = hook;
}
CDI_EXPORT void AssignCdiCloseLinkHook(int (*hook)(int id))
{
  cdiCloseLinkHook = hook;
}
CDI_EXPORT void AssignCdiExecLinkHook(int (*hook)(char *,char *,DTYPE *,DTYPE *,short,UINT16))
{
  cdiExecLinkHook = hook;
}
CDI_EXPORT void AssignCdiAttachLinkHook(int (*hook)(char *,char *,DTYPE *,DTYPE *,short,int,void (*)(int,int),int,UINT32))
{
  cdiAttachLinkHook = hook;
}
CDI_EXPORT void initTineCDI(void)
{
  AssignCdiSystemInitHook(cdiPostSystemInit);
  AssignGetCdiLinkNameHook(getCdiLinkName);
  AssignCdiCloseLinkHook(cdiCloseLink);
  AssignCdiExecLinkHook(cdiExecLinkHookFcn);
  AssignCdiAttachLinkHook(cdiAttachLinkHookFcn);
}

void fixCdiRepository(void)
{
  char *ptr;
  int len;
  static int done=0;
  if (done) return;
  done = TRUE;
  memset(CdiDBpath,0,80);
  if ((ptr=getenv("CDI_HOME")) != NULL || (ptr=getenv("CDIDB")) != NULL ||
      (ptr=getenv("FEC_HOME")) != NULL || (ptr=getenv("FECDB")) != NULL)
  {
    strncpy(CdiDBpath,ptr,80);
#   ifdef FS_DELIMITER
    len = strlen(CdiDBpath);
    if (CdiDBpath[len-1] != FS_DELIMITER) CdiDBpath[len] = FS_DELIMITER;
#   endif
  }
  rotateLogFile();
  cdilog("cdi initializing");
  cdilog("CDI_HOME : %s",CdiDBpath);
}
#ifdef UNIX
typedef void* HANDLE;
#endif
HANDLE cdiLoadLib(const char *libpath)
{
# if defined(UNIX)
  return dlopen(libpath, RTLD_LAZY);
# elif defined(WINNT)
  return LoadLibrary(libpath);
# else
  return NULL;
# endif
}
char *cdiLoadLibError(void)
{
# if defined(UNIX)
  return dlerror();
# elif defined(WINNT)
  LPVOID lpMsgBuf;
  static char lclStrBuf[256] = "";
  FormatMessage( 
    FORMAT_MESSAGE_ALLOCATE_BUFFER | 
    FORMAT_MESSAGE_FROM_SYSTEM | 
    FORMAT_MESSAGE_IGNORE_INSERTS,
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
    (LPTSTR) &lpMsgBuf,
    0,
    NULL 
  );
  strncpy(lclStrBuf,lpMsgBuf,256);
  LocalFree(lpMsgBuf);
  return lclStrBuf;
# else
  return "not implemented";
# endif
}

#define CDIMF_FILE "cdimf.csv"
typedef struct cdiCsvMfStruct
{
  char bus[32];
  char lib[64];
  char env[128];
  int id;
} cdiCsvMfStructType;
cdiCsvMfStructType gCdiMf;
CSVCOLUMN csvCdiMfColumns[] =
{
  { "BUS","",0,CF_TEXT,32,(BYTE *)gCdiMf.bus,NULL,0 },
  { "LIBRARY","",0,CF_TEXT,64,(BYTE *)gCdiMf.lib,NULL,0 },
  { "BUS_ENV","",0,CF_TEXT,128,(BYTE *)gCdiMf.env,NULL,0 },
  { "BUS_ID","",0,CF_LONG,1,(BYTE *)&gCdiMf.id,NULL,0 },
};
int csvCdiMfIterator(struct CSVDBStruct *db)
{
  int cc = 0, id;
  char libPath[256],*ccstr = "success";
# if defined(WINNT)
  char *libExt = ".dll";
  char *libPre = NULL;
  HANDLE hLib=NULL;
# elif defined(UNIX)
  void *hLib=NULL;
  char *libExt = ".so";
  char *libPre = "lib";
# endif

  if (strlen(gCdiMf.bus) > 0)
  { /* explicitly named in the manifest file */
    if ((id=cdiRegisterBus(gCdiMf.bus,gCdiMf.id)) < 0) 
    {
      cc = hardware_error;
      ccstr = erlst[cc];
      goto err;
    }
  }
  /* the loaded library will register its function handlers ... */
  memset(libPath,0,256);
  if (libPre != NULL && strncmp(libPre,gCdiMf.lib,strlen(libPre))) strcpy(libPath,libPre);
  strncat(libPath,gCdiMf.lib,200);
  if (strcmp(&libPath[strlen(libPath)-strlen(libExt)],libExt)) strcat(libPath,libExt);

  /* use bus environment to set port parameters */
  /* e.g. CANJANZ_CDI=1:500;2:300 */
  /* e.g. RS232_CDI=1:9600:0:1:1:0:0:0;2:3200:0:1:1:0:0:0 */
  if (strlen(gCdiMf.env) > 0) 
  {
    cdilog("set environment  %s : %s", gCdiMf.env, putenv(gCdiMf.env) ? "failed" : "success");
  }
  if (!strnicmp(libPath,"NULL",4))
  {
    cc = 0;
    ccstr = "process environment only";
    goto err;
  }
  if ((hLib=cdiLoadLib(libPath)) == NULL)
  {
    cc = software_error;
    ccstr = cdiLoadLibError();
    goto err;
  }
err:
  cdilog("load bus library %s : %s",libPath,ccstr);
  db->cnt++;
  return cc;
}
CSVDB csvCdiMfDb =
{
  csvCdiMfColumns,
  sizeof(csvCdiMfColumns)/sizeof(csvCdiMfColumns[0]),
  0,
  NULL,
  csvCdiMfIterator,
  NULL,
  0,
  NULL,
  0
};
CDI_EXPORT int readCdiManifest(void)
{
  int cc=0;
  fixCdiRepository();
  cc = csvReadFile(CdiDBpath,CDIMF_FILE,&csvCdiMfDb,NULL);
  cdilog("read database file %s%s : %s",CdiDBpath,CDIMF_FILE,erlst[cc]);
  return cc;
}

#define CDIDB_FILE "cdiaddr.csv"
#define CDI_PARAMETER_DELIMETER ":"
typedef struct cdiCsvAddrStruct
{
  char bus[132];
  char addr[132];
  char access[64];
  char name[16];
  char lngname[64];
  char description[96];
  int line;
  int index;
  int number;
  char mask[64];
  char pattern[64];
  char tolerance[16];
  char dataformat[16];
  char datainput[132];
  char datalimit[64];
  char clbrrule[132];
  char clbrrule_rcv[132];
  char clbrrule_snd[132];
} cdiCsvAddrStructType;
cdiCsvAddrStructType gCdiCsvAddr;
CSVCOLUMN csvCdiAddrColumns[] =
{
  { "BUS","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.bus,NULL,0 },
  { "LINE","",0,CF_LONG,1,(BYTE *)&gCdiCsvAddr.line,NULL,0 },
  { "ADDRESS","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.addr,NULL,0 },
  { "MASK","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.mask,NULL,0 },
  { "PATTERN","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.pattern,NULL,0 },
  { "ACCESS","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.access,NULL,0 },
  { "TOLERANCE","",0,CF_TEXT,16,(BYTE *)gCdiCsvAddr.tolerance,NULL,0 },
  { "NUMBER","",0,CF_LONG,1,(BYTE *)&gCdiCsvAddr.number,NULL,0 },
  { "NAME","",0,CF_TEXT,16,(BYTE *)gCdiCsvAddr.name,NULL,0 },
  { "FORMAT","short",0,CF_TEXT,16,(BYTE *)gCdiCsvAddr.dataformat,NULL,0 },
  { "INPUT","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.datainput,NULL,0 },
  { "LIMIT","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.datalimit,NULL,0 },
  { "RULE","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.clbrrule,NULL,0 },
  { "RULE_RECV","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.clbrrule_rcv,NULL,0 },
  { "RULE_SEND","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.clbrrule_snd,NULL,0 },
  { "LONG_NAME","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.lngname,NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.description,NULL,0 },
};
BYTE *parseInputData(char *str,int dataFormat,int *dataSize)
{
  char *s, *c;
  int cc=0, i, n, sizeInBytes=0;
  BYTE *inputData=NULL;

  if (str == NULL || dataSize == NULL) ccerr(argument_list_error);
  if ((sizeInBytes=fmtsizeof(dataFormat)) == 0) ccerr(illegal_format);
  /* find out how many ...*/
  for (c=str, n=1; (c=strstr(c,CDI_PARAMETER_DELIMETER)) != NULL; c++, n++);
  if (dataFormat == CF_TEXT)
  {
    if (n != 1) ccerr(illegal_string_set);
    sizeInBytes = strlen(str) + 1; /* add space for trailing '0' */
  }
  if ((inputData=(BYTE *)calloc(n,sizeInBytes)) == NULL) ccerr(out_of_server_memory);
  c = NULL;
  if ((s=strtok_r(str,CDI_PARAMETER_DELIMETER,&c)) == NULL) ccerr(string_expected);
  *dataSize = n * sizeInBytes;
  switch (dataFormat)
  {
    case CF_BYTE:
      for (i=0; i<n && s != NULL; i++) 
      {
        inputData[i] = (BYTE)strtol(s,NULL,0);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c);
      }
      break;
    case CF_SHORT:
      for (i=0; i<n; i++) 
      {
        ((short *)inputData)[i] = (short)strtol(s,NULL,0);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c);
      }
      break;
    case CF_LONG:
      for (i=0; i<n; i++) 
      {
        ((UINT32 *)inputData)[i] = (UINT32)strtol(s,NULL,0);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c);
      }
      break;
    case CF_FLOAT:
      for (i=0; i<n; i++) 
      {
        ((float *)inputData)[i] = (float)strtod(s,NULL);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c);
      }
      break;
    case CF_DOUBLE:
      for (i=0; i<n; i++) 
      {
        ((double *)inputData)[i] = (double)strtod(s,NULL);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c);
      }
      break;
    case CF_TEXT:
      *dataSize = sizeInBytes - 1;
      strncpy(inputData,s,n * sizeInBytes);
      break;
  }
err:
  if (cc != 0)
  {
    cdilog("CDI: parseInputData : %s",erlst[cc]);
    if (inputData != NULL) free(inputData);
    inputData = NULL;
  }
  return inputData;
}
int cdiParseBusString(char *str,char *bus,int *parameters,int *nparameters)
{
  char *s,*c=NULL;
  int i;
  if (str == NULL || bus == NULL) return argument_list_error;
  if ((s=strtok_r(str,CDI_PARAMETER_DELIMETER,&c)) == NULL) return string_expected;
  strncpy(bus,s,32);
  if (parameters == NULL) return 0;
  memset(parameters,0,NUMBER_EXEC_PARAM*sizeof(int));
  for (i=0; i<NUMBER_EXEC_PARAM && 
           (s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c)) != NULL; i++)
  {
    parameters[i] = strtol(s,NULL,0);
  }
  if (nparameters != NULL) *nparameters = i;
  return 0;
}
int cdiParseAddressString(char *str,int *addr,int *subaddr, int *parameters,int *nparameters)
{
  char *s,*as,*c=NULL;
  int i;
  if (str == NULL || addr == NULL) return argument_list_error;
  as = strtok_r(str,CDI_PARAMETER_DELIMETER,&c);
  if ((s=strchr(as,'.')) != NULL)
  {
    *s++ = 0;
    *addr = strtol(as,NULL,0);
    if (subaddr != NULL) *subaddr = strtol(s,NULL,0);
  }
  else
  {
    *addr = strtol(as,NULL,0);
    *subaddr = 0;
  }
  if (parameters == NULL) return 0;
  memset(parameters,0,NUMBER_EXEC_PARAM*sizeof(int));
  for (i=0; i<NUMBER_EXEC_PARAM && 
           (s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c)) != NULL; i++)
  {
    parameters[i] = strtol(s,NULL,0);
  }
  if (nparameters != NULL) *nparameters = i;
  return 0;
}
CdiClbrFcn *cdiClbrFcnLst = NULL;

int registerClbrConversionFunction(char *fcnName,double (*fcn)(double))
{
  CdiClbrFcn *f = NULL;
  if (fcnName == NULL) return argument_list_error;
  for (f=cdiClbrFcnLst; f != NULL; f=f->next)
  {
    if (strncmp(f->fcnName,fcnName,64)) continue;
    break;
  }
  if (f != NULL)
  {
    cdilog("replacing conversion function for %s",fcnName);
    f->fcn = fcn;
    return 0;
  }
  if ((f=(CdiClbrFcn *)calloc(1,sizeof(CdiClbrFcn))) == NULL) return out_of_server_memory;
  strncpy(f->fcnName,fcnName,64);
  f->fcn = fcn;
  f->next = cdiClbrFcnLst;
  cdiClbrFcnLst = f;
  cdilog("register conversion function %s",fcnName);
  return 0;
}
CDI_EXPORT int cdi_RegisterCalibrationFunction(char *fcnName,double (*fcn)(double))
{
  return registerClbrConversionFunction(fcnName,fcn);
}
int assignClbrConversionFunction(char *fcnName,CdiRule *rule)
{
  CdiClbrFcn *f = NULL;
  if (fcnName == NULL || rule == NULL) return argument_list_error;
  for (f=cdiClbrFcnLst; f != NULL; f=f->next)
  {
    if (strncmp(f->fcnName,fcnName,64)) continue;
    break;
  }
  rule->fcn = f == NULL ? NULL : f->fcn;
  return 0;
}
int chrtorule(char c)
{
  switch (c)
  {
    case CDICHAR_ADD: return cdiRuleAdd;
    case CDICHAR_SUB: return cdiRuleSub;
    case CDICHAR_MUL: return cdiRuleMult;
    case CDICHAR_DIV: return cdiRuleDiv;
    case CDICHAR_PWR: return cdiRulePwr;
    case CDICHAR_LOG: return cdiRuleLog;
    case CDICHAR_MOD: return cdiRuleMod;
    case CDICHAR_RGH: return cdiRuleShiftRight;
    case CDICHAR_LFT: return cdiRuleShiftLeft;
    case CDICHAR_UNS: return cdiRuleUnsigned;
    case CDICHAR_SGN: return cdiRuleSigned;
    case CDICHAR_MSG: return cdiRuleMessage;
    case CDICHAR_EXP: return cdiRuleExp;
    case CDICHAR_FCNA:
    case CDICHAR_FCN: return cdiRuleFcn;
    case CDICHAR_XOR: return cdiRuleXor;
  }
  return -1;
}
int getMsgStr(const char *source,char *target,int pos,int len)
{
  char str[128], *c=str;
  int i;

  strncpy(str,source,128);
  for (i=0; i<pos; i++)
  {
    if ((c=strchr(c,'<')) == NULL) break;
    c++;
  }
  if (c == NULL) 
  {
    *target = 0;
  }
  else
  {
    strncpy(target,c,len);
    if ((c=strchr(target,'>')) != NULL) *c = 0;
  }
  return 0;
}
CdiRule *parseCdiRules(char *rulesString)
{
  int cc = 0, r;
  CdiRule *rule = NULL, *next = NULL, *clbrRules = NULL;
  char *s,*c=NULL,fnam[64];
  int val;

  if (rulesString == NULL || strlen(rulesString) == 0) return NULL;
  for (s=strtok_r(rulesString,CDI_PARAMETER_DELIMETER,&c); s != NULL; s=strtok_r(NULL,CDI_PARAMETER_DELIMETER,&c))
  {
    if (strlen(s) == 0) continue;
    while (*s == ' ' || *s == '\t') s++; /* remove white space */
    if ((r=chrtorule(*s++)) < 0) ccerr(illegal_operator);
    while (isalpha(*s)) s++; /* skip to end of string */
    if ((next=(CdiRule *)calloc(1,sizeof(CdiRule))) == NULL) ccerr(out_of_local_memory);
    next->operation = r;
    switch (r)
    {
      case cdiRuleSigned:
        next->isSigned = TRUE;
        break;
      case cdiRuleUnsigned:
        next->isSigned = FALSE;
        break;
      case cdiRuleMessage:
        next->msgOnValue = strtol(s,&s,0);
        /* parse for <msgOn><msgOff> */
        getMsgStr(s,next->msgOn,1,32);
        getMsgStr(s,next->msgOff,2,32);
        break;
      case cdiRuleFcn:
        getMsgStr(s,fnam,1,64);
        assignClbrConversionFunction(fnam,next);
        break;
      case cdiRuleExp:
        val = 0;
        if (strlen(s) > 0) val = (int)strtod(s,NULL);
        if (val == 0) val = 10;
        next->operand = (double) val;
        break;
      default:
        next->operand = strtod(s,NULL);
        break;
    }
    if (clbrRules == NULL) 
    { /* first in the list */
      clbrRules = next;
      rule = next;
    }
    else
    { /* advance it */
      rule->nxt = next;
      rule = next;
    }
  }

err:
  if (cc)
  {
    cdilog("cannot parse rules string : %s",erlst[cc]);
    if (clbrRules != NULL)
    {
      for (rule=clbrRules; rule != NULL; rule=next)
      {
         next = rule->nxt;
         free(rule);
      }
      clbrRules = NULL;
    }
  }
  return clbrRules;
}
void csvCdiAddrInitializer(struct CSVDBStruct *db)
{
  if (db->siz == 0)
  {
    cdilog("cdi database empty");
    return;
  }
  if ((paramList=(PARAM_STRUCT *)SystemCalloc(db->siz*gMaxTemplateLength,sizeof(PARAM_STRUCT))) == NULL)
  {
    cdilog("unable to allocate device parameter list space");
    return;
  }
  paramListSize = db->siz * gMaxTemplateLength;
}

int cdiRegisterTemplateDevices(int ln,cdiCsvAddrStructType *gCdiCsvAddr,CdiTemplate *tmpl,int nrecords)
{
  int cc, i, devnr, devidx, lnidx, addr, saddr, *prms;
  static int nparams, params[NUMBER_EXEC_PARAM];
  char s[512], dname[32], lname[128], desc[96];
  CdiTemplateList *t;

  if (tmpl == NULL || gCdiCsvAddr == NULL) return argument_list_error;
  if ((cc=cdiParseAddressString(gCdiCsvAddr->addr,&addr,&saddr,params,&nparams)) != 0)
  {
    cdilog("cannot parse address string -> %s",erlst[cc]);
    return cc;
  }
  lnidx = getNextFreeIndex(ln);
  if (lnidx < 0)
  {
    cc = -lnidx;
    cdilog("cannot acquire line index for line %d -> %s",ln,erlst[cc]);
    return cc;
  }
  for (i=0,t=tmpl->cdiTemplateList; i<tmpl->length && t != NULL; i++, t=t->next)
  {
#   ifdef WINNT
    _snprintf(dname,32,"%s.%s",gCdiCsvAddr->name,t->field);
    _snprintf(lname,64,"%s.%s",gCdiCsvAddr->lngname,t->lngfield);
    _snprintf(desc,96,"%s (%s)",gCdiCsvAddr->description,t->description);
#   else
    snprintf(dname,32,"%s.%s",gCdiCsvAddr->name,t->field);
    snprintf(lname,64,"%s.%s",gCdiCsvAddr->lngname,t->lngfield);
    snprintf(desc,96,"%s (%s)",gCdiCsvAddr->description,t->description);
#   endif
    sprintf(s,"(%d.%d:%d:%d)",addr,saddr,t->addrParameters[0],t->addrParameters[1]);
    devidx = lnidx + nrecords * i;
    devnr = gCdiCsvAddr->number + (nrecords + TEMPLATE_RECORD_GAP)* i;
    if ((cc=cdi_RegisterDevice(ln,dname,lname,desc,devidx,devnr,addr,saddr)) != 0)
    {
      cdilog("cannot register device %s at devIndex %d and devNr %d on line %d -> %s",dname,devidx,devnr,ln,erlst[cc]);
      return cc;
    }
    cdilog("%s %s line %d line index %d %s : cdi dev nr %d on cdi bus #%d",
      gCdiCsvAddr->bus,dname, gCdiCsvAddr->line, devidx, s, devnr, ln);
     
    prms = t->naddrParameters > 0 ? t->addrParameters : params;
    if ((cc=cdi_assignDefaultDeviceParameters(dname,prms)) != 0)
    {
      cdilog("assign device parameters to %s -> %s",dname,erlst[cc]);
      return cc;
    }
    if ((cc=cdi_assignDefaultDataMaskAndPattern(dname,t->dataMask,t->dataMaskRange,t->dataPattern,t->dataPatternRange,t->dataPatternTrigger)) != 0)
    {
      cdilog("assign mask/pattern to %s -> %s",dname,erlst[cc]);
      return cc;
    }
    if ((cc=cdi_assignDataFormat(dname,t->dataFormat)) != 0)
    {
      cdilog("data format for %s could not be assigned : %s",dname,erlst[cc]);
      return cc;
    }
    if (t->dataInput != NULL)
    {
      if (t->dataFormat == CF_NULL)
      {
        cdilog("data input provided without a valid format!");
        return illegal_format;
      }
      if ((cc=cdi_assignDefaultInput(dname,t->dataInput,t->dataInputSize,t->dataFormat)) != 0)
      {
        cdilog("data input for %s could not be assigned : %s",dname,erlst[cc]);
        return cc;
      }
    }
    if ((cc=cdi_assignDataLimits(dname,t->dataOutputSize,t->dataInputSize)) != 0)
    {
      cdilog("data limits for %s could not be assigned : %s",dname,erlst[cc]);
      return cc;
    }
    if (t->clbrRules != NULL || t->rvrsRules != NULL)
    {
      /*if ((cc=cdi_assignDataCalibration(dname,t->clbrRules,getCdiFormat((short)t->dataFormat))) != 0) */
      if ((cc=cdi_assignDataCalibration(dname,t->clbrRules,t->rvrsRules,t->dataFormat)) != 0)
      {
        cdilog("data calibration rule for %s could not be assigned : %s",dname,erlst[cc]);
        return cc;
      }
    }
    cdi_assignAllowedAccess(dname,t->allowedAccess);
  }
  return 0;
}

int csvCdiAddrIterator(struct CSVDBStruct *db)
{
  int cc, i, fmt, fsiz, siz=0, sizin=0, sizout=0, ln, addr, saddr, dw, access=0xffff;
  static int nparams, params[NUMBER_EXEC_PARAM];
  char bus[32], busParams[132];
  BYTE *pPattern=NULL, *pDataMask=NULL, *pDataInput=NULL;
  char *c, *c1, s[512];
  CdiTemplate *tmpl;
  int patRange[2],mskRange[2],patTrigger=CM_SEEKPATTERN;

  if (strlen(gCdiCsvAddr.bus) == 0) return 0;
  /* gCdiCsvAddr has been filled in */
  /* if a TEMPLATE entry then do nothing here ... */
  if (!stricmp(gCdiCsvAddr.bus,"TEMPLATE")) return 0;
  /* parameters from address and bus strings ... */
  strncpy(s,gCdiCsvAddr.bus,132);
  if ((c=strstr(s,CDI_PARAMETER_DELIMETER)) != NULL)
  {
    *c++ = 0;
    strncpy(busParams,c,132);
  }
  else
  {
    busParams[0] = 0;
  }
  strncpy(bus,s,32);
  /* number of devices -> twice the database size should allow for enough dynamic entries */
  /* repeated calls will just return the line index ! */
  if ((ln=cdi_Open(bus, gCdiCsvAddr.line, busParams, db->siz*gMaxTemplateLength, 0)) < 0)
  {
    cdilog("cannot open bus %s -> %s",bus,erlst[-ln]);
    return -ln;
  }
  /* check address for template */
  if ((c=strstr(gCdiCsvAddr.addr,":<")) != NULL)
  {
    *c = 0; /* peel it off */
    strncpy(s,c+2,32); s[32] = 0;
    if ((c=strstr(s,">")) != NULL) *c = 0;
    if ((tmpl=getCdiTemplate(s)) == NULL)
    {
      cdilog("could not get template %s for %s!",s,gCdiCsvAddr.name);
      return invalid_structure_tag;
    }
    return cdiRegisterTemplateDevices(ln,&gCdiCsvAddr,tmpl,db->siz);
  }
  /* no template assigned:  just this record */
  nparams = 0; s[0] = 0;
  if ((cc=cdiParseAddressString(gCdiCsvAddr.addr,&addr,&saddr,params,&nparams)) != 0)
  {
    cdilog("cannot parse address string -> %s",erlst[cc]);
    return cc;
  }
  for (i=0; i<nparams; i++) 
    sprintf(&s[strlen(s)],i == 0 ? "(%d" : " %d",params[i]);
  if (nparams > 0) strcat(s,")");
  if ((gCdiCsvAddr.index = getNextFreeIndex(ln)) < 0)
  {
    cc = -gCdiCsvAddr.index;
    cdilog("cannot acquire index for device %s, number: %d on line %d -> %s",
      gCdiCsvAddr.name,gCdiCsvAddr.number,ln,erlst[cc]);
    return cc;
  }
  if ((cc=cdi_RegisterDevice(ln,gCdiCsvAddr.name,gCdiCsvAddr.lngname,gCdiCsvAddr.description,gCdiCsvAddr.index,gCdiCsvAddr.number,addr,saddr)) != 0)
  {
    cdilog("cannot register device %s, number: %d at index %d on line %d -> %s",
      gCdiCsvAddr.name,gCdiCsvAddr.number,gCdiCsvAddr.index,ln,erlst[cc]);
    return cc;
  }
  cdilog("%s %s %s line %d line index %d %s : cdi dev nr %d on cdi bus #%d",
    bus,busParams,gCdiCsvAddr.name, gCdiCsvAddr.line, gCdiCsvAddr.index, s, gCdiCsvAddr.number, ln);
  if ((cc=cdi_assignDefaultDeviceParameters(gCdiCsvAddr.name,params)) != 0)
  {
    cdilog("assign device parameters to %s -> %s",gCdiCsvAddr.name,erlst[cc]);
    return cc;
  }
  /*
    returnCondition established when making the call to AttachLink() or ExecLink()
    at which time the pattern is extracted from the dout data array ....
    supplying a pattern below is in the present state useless ...
  */
  patRange[0] = patRange[1] = -1;
  if (strlen(gCdiCsvAddr.pattern) > 0)
  { /* 32 bytes */
    if (!strncmp(gCdiCsvAddr.pattern,"!=",2))
    { /* looking for pattern not equal to that given */
      patTrigger = CM_MISSPATTERN;
      memset(s,0,64);
      strncpy(s,&gCdiCsvAddr.pattern[2],62);
      strtrm(s);
      memcpy(gCdiCsvAddr.pattern,s,62);
    }
    if ( gCdiCsvAddr.pattern[0] >= '0' && gCdiCsvAddr.pattern[0] <= '9' )
    { /* interpret as a byte pattern (up to 4 bytes, i.e. 8 chars) */
      strncpy(s,gCdiCsvAddr.pattern,64); s[63] = 0;
      if ((c=strchr(s,':')) != NULL)
      {
        *c++ = 0;
        if ((c1=strchr(c,':')) != NULL)
        {
          *c1++ = 0;
          patRange[1] = strtol(c1,NULL,0);
        }
        patRange[0] = strtol(c,NULL,0);
      }
      else
      {
        s[10] = 0;
      }
      dw = strtol(s,NULL,0);
      memset(gCdiCsvAddr.pattern,0,32);
      memcpy(gCdiCsvAddr.pattern,&dw,4);
      /* *(int *)gCdiCsvAddr.pattern = dw; */
    }
    pPattern = gCdiCsvAddr.pattern;
  }
  mskRange[0] = mskRange[1] = -1;
  if (strlen(gCdiCsvAddr.mask) > 0)
  { /* 8 bytes */
    if (!strncmp(gCdiCsvAddr.mask,"0x",2) || !strncmp(gCdiCsvAddr.mask,"0X",2)) 
    { /* interpret as a byte pattern (up to 4 bytes, i.e. 8 chars) */
      strncpy(s,gCdiCsvAddr.mask,64); s[63] = 0;
      if ((c=strchr(s,':')) != NULL)
      {
        *c++ = 0;
        if ((c1=strchr(c,':')) != NULL)
        {
          *c1++ = 0;
          mskRange[1] = atoi(c1);
        }
        mskRange[0] = atoi(c);
      }
      else
      {
        s[10] = 0;
      }
      dw = strtol(s,NULL,16);
      memset(gCdiCsvAddr.mask,0,32);
      memcpy(gCdiCsvAddr.mask,&dw,4);
    }
    pDataMask = gCdiCsvAddr.mask;
  }
  if ((cc=cdi_assignDefaultDataMaskAndPattern(gCdiCsvAddr.name,pDataMask,mskRange,pPattern,patRange,patTrigger)) != 0)
  {
    cdilog("assign mask/pattern to %s -> %s",gCdiCsvAddr.name,erlst[cc]);
    return cc;
  }
  fmt = getSimpleFormatType(strupr(gCdiCsvAddr.dataformat));
  if ((cc=cdi_assignDataFormat(gCdiCsvAddr.name,getCdiFormat((short)fmt))) != 0)
  {
    cdilog("data format could not be assigned : %s",erlst[cc]);
    return cc;
  }
  if (strlen(gCdiCsvAddr.datainput) > 0)
  {
    if (fmt == CF_NULL)
    {
      cdilog("data input provided without a valid format!");
      return illegal_format;
    }
    if ((pDataInput=parseInputData(gCdiCsvAddr.datainput,fmt,&siz)) == NULL)
    {
      cdilog("data input could not be parsed!");
      return invalid_data;
    }
    if ((cc=cdi_assignDefaultInput(gCdiCsvAddr.name,pDataInput,siz,getCdiFormat((short)fmt))) != 0)
    {
      cdilog("data input could not be assigned : %s",erlst[cc]);
      return cc;
    }
  }
  if (strlen(gCdiCsvAddr.datalimit) > 0)
  {
    if ((fmt=getSimpleFormatType(strupr(gCdiCsvAddr.dataformat))) == CF_NULL)
    {
      cdilog("data input provided without a valid format!");
      return illegal_format;
    }
    fsiz = fmtsizeof(fmt);
    strncpy(s,gCdiCsvAddr.datalimit,16);
    if ((c=strstr(s,CDI_PARAMETER_DELIMETER)) != NULL) 
    {
      *c++ = 0;
      sizin = strtol(c,NULL,10) * fsiz;
      if (siz > 0 && sizin > siz) sizin = siz;
    }
    sizout = strtol(s,NULL,10) * fsiz;
    if ((cc=cdi_assignDataLimits(gCdiCsvAddr.name,sizout,sizin)) != 0)
    {
      cdilog("data limits could not be assigned : %s",erlst[cc]);
      return cc;
    }
  }
  if (strlen(gCdiCsvAddr.clbrrule_rcv) == 0 && strlen(gCdiCsvAddr.clbrrule) > 0)
  { /* support legacy csv column */
    strncpy(gCdiCsvAddr.clbrrule_rcv,gCdiCsvAddr.clbrrule,132);
  }
  if (strlen(gCdiCsvAddr.clbrrule_rcv) > 0 || strlen(gCdiCsvAddr.clbrrule_snd) > 0 || strlen(gCdiCsvAddr.clbrrule) > 0)
  {
    if ((cc=cdi_assignDataCalibration(gCdiCsvAddr.name,parseCdiRules(gCdiCsvAddr.clbrrule_rcv),
      parseCdiRules(gCdiCsvAddr.clbrrule_snd),getCdiFormat((short)fmt))) != 0)
    {
      cdilog("data calibration rule could not be assigned : %s",erlst[cc]);
      return cc;
    }
  }
  access = getCdiAllowedAccess(gCdiCsvAddr.access);
  cdi_assignAllowedAccess(gCdiCsvAddr.name,access);
  db->cnt++;
  return 0;
}
CSVDB csvCdiAddrDb =
{
  csvCdiAddrColumns,
  sizeof(csvCdiAddrColumns)/sizeof(csvCdiAddrColumns[0]),
  0,
  csvCdiAddrInitializer,
  csvCdiAddrIterator,
  NULL,
  0,
  NULL,
  0
};

int csvCdiTemplateIterator(struct CSVDBStruct *db)
{
  int cc, i, fmt, fsiz, siz=0, sizin=0, sizout=0, addr, saddr, dw, access=0xffff;
  static int nparams, params[NUMBER_EXEC_PARAM];
  char templateName[32], templateField[32], templateLongField[64], fieldDescription[96];
  BYTE *pPattern=NULL, *pDataMask=NULL, *pDataInput=NULL;
  char *c, *c1, s[512];
  CdiTemplate *tmpl;
  CdiRule *rule_rcv=NULL,*rule_snd=NULL;
  int mskRange[2], patRange[2], patTrigger=CM_SEEKPATTERN;

  /* gCdiCsvAddr has been filled in */
  /* if not a TEMPLATE entry then do nothing ... */
  if (stricmp(gCdiCsvAddr.bus,"TEMPLATE")) return 0;
  /* template name is in the device name string */
  strncpy(s,gCdiCsvAddr.name,132);
  if ((c=strstr(s,CDI_PARAMETER_DELIMETER)) == NULL)
  { /* badly formed ! */
    cdilog("template entry without field (%s)!",gCdiCsvAddr.name);
    return argument_list_error;
  }
  *c++ = 0;
  strncpy(templateName,s,16);
  if ((tmpl=getCdiTemplate(templateName)) == NULL)
  {
    cdilog("could not get template for %s!",templateName);
    return resources_exhausted;
  }
  strncpy(templateField,c,16);
  strncpy(templateLongField,gCdiCsvAddr.lngname,64);
  strncpy(fieldDescription,gCdiCsvAddr.description,96);

  nparams = 0; s[0] = 0;
  if ((cc=cdiParseAddressString(gCdiCsvAddr.addr,&addr,&saddr,params,&nparams)) != 0)
  {
    cdilog("cannot parse address string -> %s",erlst[cc]);
    return cc;
  }
  for (i=0; i<nparams; i++) 
    sprintf(&s[strlen(s)],i == 0 ? "(%d" : " %d",params[i]);
  if (nparams > 0) strcat(s,")");

  cdilog("Template %s field %s : parameters %s",templateName,templateField,s);

  patRange[0] = patRange[1] = -1;
  if (strlen(gCdiCsvAddr.pattern) > 0)
  { /* 32 bytes */
    if (!strncmp(gCdiCsvAddr.pattern,"!=",2))
    { /* looking for pattern not equal to that given */
      patTrigger = CM_MISSPATTERN;
      memset(s,0,64);
      strncpy(s,&gCdiCsvAddr.pattern[2],62);
      strtrm(s);
      memcpy(gCdiCsvAddr.pattern,s,62);
    }
    if ( gCdiCsvAddr.pattern[0] >= '0' && gCdiCsvAddr.pattern[0] <= '9' )
    { /* interpret as a byte pattern (up to 4 bytes, i.e. 8 chars) */
      strncpy(s,gCdiCsvAddr.pattern,64); s[63] = 0;
      if ((c=strchr(s,':')) != NULL)
      {
        *c++ = 0;
        if ((c1=strchr(c,':')) != NULL)
        {
          *c1++ = 0;
          patRange[1] = strtol(c1,NULL,0);
        }
        patRange[0] = strtol(c,NULL,0);
      }
      else
      {
        s[10] = 0;
      }
      dw = strtol(s,NULL,0);
      memset(gCdiCsvAddr.pattern,0,32);
      *(int *)gCdiCsvAddr.pattern = dw;
    }
    pPattern = gCdiCsvAddr.pattern;
  }
  mskRange[0] = mskRange[1] = -1;
  if (strlen(gCdiCsvAddr.mask) > 0)
  { /* 8 bytes */
    if (!strncmp(gCdiCsvAddr.mask,"0x",2) || !strncmp(gCdiCsvAddr.mask,"0X",2)) 
    { /* interpret as a byte pattern (up to 4 bytes, i.e. 8 chars) */
      strncpy(s,gCdiCsvAddr.mask,64); s[63] = 0;
      if ((c=strchr(s,':')) != NULL)
      {
        *c++ = 0;
        if ((c1=strchr(c,':')) != NULL)
        {
          *c1++ = 0;
          mskRange[1] = atoi(c1);
        }
        mskRange[0] = atoi(c);
      }
      else
      {
        s[10] = 0;
      }
      dw = strtol(s,NULL,16);
      memset(gCdiCsvAddr.mask,0,32);
      memcpy(gCdiCsvAddr.mask,&dw,4);
    }
    pDataMask = gCdiCsvAddr.mask;
  }
  fmt = getSimpleFormatType(strupr(gCdiCsvAddr.dataformat));
  if (strlen(gCdiCsvAddr.datainput) > 0)
  {
    if (fmt == CF_NULL)
    {
      cdilog("data input provided without a valid format!");
      return illegal_format;
    }
    if ((pDataInput=parseInputData(gCdiCsvAddr.datainput,fmt,&siz)) == NULL)
    {
      cdilog("data input could not be parsed!");
      return invalid_data;
    }
  }
  if (strlen(gCdiCsvAddr.datalimit) > 0)
  {
    if ((fmt=getSimpleFormatType(strupr(gCdiCsvAddr.dataformat))) == CF_NULL)
    {
      cdilog("data input provided without a valid format!");
      return illegal_format;
    }
    fsiz = fmtsizeof(fmt);
    strncpy(s,gCdiCsvAddr.datalimit,16);
    if ((c=strstr(s,CDI_PARAMETER_DELIMETER)) != NULL) 
    {
      *c++ = 0;
      sizin = strtol(c,NULL,10) * fsiz;
      if (siz > 0 && sizin > siz) sizin = siz;
    }
    sizout = strtol(s,NULL,10) * fsiz;
  }
  if (strlen(gCdiCsvAddr.clbrrule_rcv) == 0 && strlen(gCdiCsvAddr.clbrrule) > 0)
  { /* support legacy csv column */
    strncpy(gCdiCsvAddr.clbrrule_rcv,gCdiCsvAddr.clbrrule,132);
  }
  if (strlen(gCdiCsvAddr.clbrrule_rcv) > 0)
  {
    rule_rcv = parseCdiRules(gCdiCsvAddr.clbrrule_rcv);
  }
  if (strlen(gCdiCsvAddr.clbrrule_snd) > 0)
  {
    rule_snd = parseCdiRules(gCdiCsvAddr.clbrrule_snd);
  }
  access = getCdiAllowedAccess(gCdiCsvAddr.access);

  if ((cc=addFieldToTemplate(tmpl,templateField,templateLongField,fieldDescription,
            params,nparams,pDataMask,mskRange,pPattern,patRange,patTrigger,
            pDataInput,sizin,sizout,getCdiFormat((short)fmt),
            access,rule_rcv,rule_snd)) != 0)
  {
    cdilog("cannot add template field %s to template %s : %s",templateName,templateField,erlst[cc]);
    return cc;
  }

  db->cnt++;
  return 0;
}

CSVDB csvCdiTmplDb =
{
  csvCdiAddrColumns,
  sizeof(csvCdiAddrColumns)/sizeof(csvCdiAddrColumns[0]),
  0,
  NULL,
  csvCdiTemplateIterator,
  NULL,
  0,
  NULL,
  0
};
int cdiAddrDbSize = 0;
int getMaxTemplateLength(void)
{
  int i, maxsiz=1;

  for (i=0; i<CdiTemplateTblSize; i++)
  {
    if (CdiTemplateTbl[i] == NULL) continue;
    if (CdiTemplateTbl[i]->length > maxsiz) maxsiz = CdiTemplateTbl[i]->length;
  }
  return maxsiz;
}
CDI_EXPORT int readCdiDataBase(void)
{
  int cc=0;
  fixCdiRepository();
  /* scan for templates */
  cc = csvReadFile(CdiDBpath,CDIDB_FILE,&csvCdiTmplDb,NULL);
  gMaxTemplateLength = getMaxTemplateLength();
  cdilog("largest template has %d fields",gMaxTemplateLength);
  /* scan for records */
  cc = csvReadFile(CdiDBpath,CDIDB_FILE,&csvCdiAddrDb,NULL);
  cdiAddrDbSize = csvCdiAddrDb.siz;
  cdilog("read database file %s%s : %s",CdiDBpath,CDIDB_FILE,erlst[cc]);
  return cc;
}
#ifdef UNIX
/* prolog code for UNIX shared library (WINDOWS uses DllMain in cdi32.c) */
__attribute__((constructor)) void cdiLibMain(void)
{
  cdiInitialize();
}
#endif
/* ===============================================================
  Below are routines which provide the server functionality
 ================================================================= */

/* call cdiPostSystemInit() following the call to SystemInit() ... */
CDI_EXPORT void cdiPostSystemInit(void)
{
  char cdiExportName[32];
  int ndevices;
  int devParameters[NUMBER_EXEC_PARAM];
  static int initialized = FALSE;
  if (initialized) return;
  cdiGetBusNames();
  if (strlen(gFecName) == 0)
  {
    cdilog("fecname unknown at time of CDI initialization!");
    return;
  }
  memset(devParameters,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdi_RegisterParameters("null", devParameters);
  strncpy(cdiExportName,gFecName,12); cdiExportName[12] = 0;
  strcat(cdiExportName,".CDI");
  cdilog("CDI export name : %s",cdiExportName);
  ndevices = cdi_getDeviceNames(NULL,0) + nCdiRedirectedDevices;
  RegisterEquipmentModule(cdiExportName,CDIEQM_TAG,ndevices,cdieqm,cdieqm_ini,cdieqm_bkg,CDIEQM_RATE,cdieqm_exi);
  strncpy(gCdiExportName,cdiExportName,32);
  strncpy(gCdiExportContext,GetExportedContext(CDIEQM_TAG),32);
  sprintf(gCdiExportPrefix,"/%s/%s/",gCdiExportName,gCdiExportContext);
  gCdiExportPrefixLength = strlen(gCdiExportPrefix);
  cdilog("CDI remote server initialized");
  initialized = TRUE;
}
CDI_EXPORT void cdieqm_bkg(void)
{
}
CDI_EXPORT int cdiRegisterProperties(void)
{
  int cc = 0;

  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSADDR",PRP_ADDR_SIZE,CF_ADDRESS,CA_READ,"[0:65535 ]HW address in use by selected device",PRP_ADDR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSNAME",PRP_BUS_NAME_SIZE,CF_NAME16,CA_READ,"[0:65535 ]Bus in use by selected device",PRP_BUS_NAME)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSERRORS",PRP_BUS_ERRORS_SIZE,CF_LONG,CA_READ,"[0:65535 ]Number of bus errors since server start",PRP_BUS_ERRORS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSNAMES",PRP_BUS_NAMES_SIZE,CF_NAME16,CA_READ,"[0:65535 ]Supported Bus names",PRP_BUS_NAMES)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSSCAN",PRP_BUS_SCAN_SIZE,CF_TEXT,CA_READ,"[0:65535 ]Scan bus according to input given",PRP_BUS_SCAN)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSSCAN.NAM",PRP_BUS_SCAN_NAM_SIZE,CF_NAME16,CA_READ,"[0:65535 ]registered bus names",PRP_BUS_SCAN_NAM)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"TEMPLATE",PRP_TEMPLATE_SIZE,CF_NAME16II,CA_READ,"[0:65535 ]template information",PRP_TEMPLATE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"TEMPLATE.NAM",PRP_TEMPLATE_NAM_SIZE,CF_NAME16,CA_READ,"[0:65535 ]registered template names",PRP_TEMPLATE_NAM)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.ATOM",PRP_WRITEREAD_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Write+Read Telegram Pair",PRP_WRITEREAD)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV.SEND.ATOM",PRP_READWRITE_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Read+Write Telegram Pair",PRP_READWRITE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND",PRP_WRITE_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Write Telegram",PRP_WRITE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV",PRP_READ_SIZE,CF_SHORT,CA_READ,"[0:65535 ]Issue CDI Read Telegram",PRP_READ)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV.CLBR",PRP_READ_SIZE,CF_SHORT,CA_READ,"[0:65535 ]Issue CDI Read Telegram with calibration",PRP_READ_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.CLBR",PRP_WRITEREAD_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Write+Read Telegram Pair with calibration",PRP_WRITEREAD_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"DESCRIPTION",PRP_DESCRIPTION_SIZE,CF_NAME64,CA_READ,"[0:0 ]description selected device",PRP_DESCRIPTION)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.SEND.ATOM",PRP_WRITE_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Write+Write Telegram",PRP_WRITEWRITE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.SEND.ATOM",PRP_READ_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Write+READ+Write Telegram",PRP_WRITEREADWRITE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"NUMBER",PRP_CDINUMBER_SIZE,CF_ADDRESS,CA_READ,"[0:1000 ]CDI device number (plus line and index) assigned to selected device",PRP_CDINUMBER)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RANGE",PRP_RANGE_SIZE,CF_FLOAT,CA_READ,"[0:1000 ]value range (calibrated) for selected device",PRP_RANGE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"UNITS",PRP_UNITS_SIZE,CF_NAME16,CA_READ,"[0:0 ]assigned units (calibrated) for selected device",PRP_UNITS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"PRECISION",PRP_PRECISION_SIZE,CF_NAME16,CA_READ,"[0:0 ]assigned format precision (calibrated) for selected device",PRP_PRECISION)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.CLBR",PRP_WRITE_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Write Telegram",PRP_WRITE_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV.SEND.CLBR",PRP_READWRITE_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Read+Write Telegram Pair",PRP_READWRITE_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.SEND.CLBR",PRP_WRITE_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Write+Write Telegram",PRP_WRITEWRITE_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.SEND.CLBR",PRP_READ_SIZE,CF_SHORT,CA_READ|CA_WRITE,"[0:65535 ]Issue CDI Write+READ+Write Telegram",PRP_WRITEREADWRITE_CLBR)) < 0) goto err;
err:
  return cc < 0 ? -cc : 0;
}
CDI_EXPORT void cdieqm_ini(void)
{
  int cc;
  if ((cc=cdiRegisterProperties()) != 0) cdilog("cdiRegisterProperties : %s",erlst[cc]);
  if ((cc=cdiRegisterDeviceNames()) != 0) cdilog("cdiRegisterDeviceNames : %s",erlst[cc]);
  if ((cc=cdiRegisterExtendedProperties()) != 0) cdilog("cdiRegisterExtendedProperties : %s",erlst[cc]);
# if defined(UNIX) || defined(WINNT)
  RegisterUserCommand("cdidebug",NULL,(int *)&cdiDebugLevel,NULL,CA_READ|CA_WRITE);
  RegisterUserCommand("cdilisteners",cdiDumpListeners,NULL,NULL,CA_READ);
# endif
}
CDI_EXPORT void cdieqm_exi(void)
{
}
#define MAX_QUERYABLE_LIST_SIZE 1024
CDI_EXPORT int cdieqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc,devnr=0,prpid,extid=-1,errcnt=0,i,len,namlen,issync,rate,flags;
  int isfirst = (access & CA_FIRST);
  int islast = (access & CA_LAST);
  int cdiDevNr, cdiLine, cdiIndex, numDevices;
  float range[2];
  char *c = NULL, *scanTarget = NULL, dn[512], strbuffer[256], devDesc[64];
  static NAME16 devList[MAX_QUERYABLE_LIST_SIZE];
  DADDRESS da = { 0,0,0,0 }, *dap;
  CdiTemplate *tmpl;
  CdiTemplateList *tmplst;
  cdiListener *ml;
  double dt;

  /* prepare local device name */
  sprintf(dn,"/LOCALHOST/CDI/%s",devName);
  
  /* get device number from device name */
  devnr = GetDeviceNumberEx(CDIEQM_TAG,devName,devProperty);
  prpid = GetPropertyId(CDIEQM_TAG,devProperty);
  if (prpid >= PRP_EXTENDED_OFFSET &&
      prpid < PRP_EXTENDED_OFFSET + gNumExtCdiProperties)
  {
    extid = prpid - PRP_EXTENDED_OFFSET;
  }

  issync = (isfirst && islast) ? TRUE : FALSE; /* is this a synchronous request ? */

  switch (prpid)
  {
    case PRP_RANGE:
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,devList,&numDevices)) != 0) return cc;
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        switch (dout->dFormat)
        {
          case CF_FLOAT:
          case CF_LONG:
          case CF_SHORT:
          case CF_DOUBLE:
            if (numDevices > 1) return illegal_format;
            cdi_getRangeForDevice(devName,range);
            return putValuesFromFloat(dout,range,2);
          case CF_FLTFLT:
           if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
            for (i=0; i<numDevices; i++)
            {
              cdi_getRangeForDevice(devList[i].name,range);
              ((DPOINT *)dout->data.vptr)[i].x = range[0];
              ((DPOINT *)dout->data.vptr)[i].y = range[1];
            }
            break;
          case CF_DBLDBL:
            if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
            for (i=0; i<numDevices; i++)
            {
              cdi_getRangeForDevice(devList[i].name,range);
              ((DBLDBL *)dout->data.vptr)[i].d1val = (double)range[0];
              ((DBLDBL *)dout->data.vptr)[i].d2val = (double)range[1];
            }
            break;
          default:
            return illegal_format;
        }
        return 0;
    case PRP_UNITS:
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,devList,&numDevices)) != 0) return cc;
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
        len = fmtsizeof(dout->dFormat);
        switch (dout->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            namlen = 16;
            break;
          case CF_TEXT:
            if (dout->dArrayLength > 1) return illegal_format;
            namlen = MIN(dout->dArrayLength,16);
            break;
          default:
            return illegal_format;
        }
        for (i=0; i<numDevices; i++)
        {
          cdi_getUnitsForDevice(devList[i].name,devDesc);
          strncpy(dout->data.cptr+i*namlen,devDesc,namlen);
        }
        return 0;
    case PRP_PRECISION:
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,devList,&numDevices)) != 0) return cc;
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
        len = fmtsizeof(dout->dFormat);
        switch (dout->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            namlen = 16;
            break;
          case CF_TEXT:
            if (dout->dArrayLength > 1) return illegal_format;
            namlen = MIN(dout->dArrayLength,16);
            break;
          default:
            return illegal_format;
        }
        for (i=0; i<numDevices; i++)
        {
          cdi_getPrecisionForDevice(devList[i].name,devDesc);
          strncpy(dout->data.cptr+i*namlen,devDesc,namlen);
        }
        return 0;
    case PRP_CDINUMBER:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength < 1) return dimension_error;
        cdiDevNr = findDevice(devName,&cdiLine,&cdiIndex);
        if (cdiDevNr < 0) return -cdiDevNr;
        switch (dout->dFormat)
        {
          case CF_ADDRESS:
            ((DADDRESS *)dout->data.vptr)[0].i1val = cdiDevNr;
            ((DADDRESS *)dout->data.vptr)[0].i2val = cdiLine;
            ((DADDRESS *)dout->data.vptr)[0].i3val = cdiIndex;
            break;
          default:
            return putValuesFromLong(dout,&cdiDevNr,1);
        }
        return 0;
    case PRP_DESCRIPTION:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,devList,&numDevices)) != 0) return cc;
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
        len = fmtsizeof(dout->dFormat);
        switch (dout->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            namlen = len;
            break;
          case CF_TEXT:
            if (dout->dArrayLength > 1) return illegal_format;
            namlen = MIN(dout->dArrayLength,64);
            break;
          default:
            return illegal_format;
        }
        for (i=0; i<numDevices; i++)
        {
          cdi_getDescriptionForDevice(devList[i].name,devDesc);
          strncpy(dout->data.cptr+i*namlen,devDesc,namlen);
        }
        return 0;
    case PRP_ADDR:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_ADDR_SIZE) return dimension_error;
          if (dout->dFormat != CF_ADDRESS) return illegal_format;
          if ((dap=cdi_getBusAddressForDevice(devName)) == NULL) dap = &da;
          memcpy(dout->data.vptr,dap,dout->dArrayLength*fmtsizeof(dout->dFormat));
        }
        return 0;
    case PRP_BUS_NAME:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_BUS_NAME_SIZE) return dimension_error;
          if (dout->dFormat != CF_NAME16) return illegal_format;
          if ((c=cdi_getBusNameForDevice(devName)) == NULL) c = "[none]";
          memcpy(dout->data.vptr,c,dout->dArrayLength*fmtsizeof(dout->dFormat));
        }
        return 0;
    case PRP_BUS_ERRORS:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_BUS_ERRORS_SIZE) return dimension_error;
          if ((errcnt=cdi_getDeviceNumberOfErrors(devName)) < 0) return -errcnt;
          return putValuesFromLong(dout,&errcnt,1);
        }
        return 0;
    case PRP_BUS_SCAN_NAM:
    case PRP_BUS_NAMES:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_NAME16) return illegal_format;
          if (nCdiBusNames < (int)dout->dArrayLength) dout->dArrayLength = nCdiBusNames;
          if ((c=(char *)cdiGetBusNames()) == NULL) c = "[none]";
          memcpy(dout->data.vptr,c,dout->dArrayLength*fmtsizeof(dout->dFormat));
        }
        return 0;
    case PRP_BUS_SCAN:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (din->dArrayLength > 0)
        {
          if (din->dFormat != CF_TEXT) return illegal_format;
          strncpy(strbuffer,din->data.cptr,din->dArrayLength);
          strbuffer[din->dArrayLength] = 0;
          scanTarget = strbuffer;
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_TEXT) return illegal_format;
          if ((cc=cdiScanBus(devName,scanTarget,dout->data.cptr,dout->dArrayLength)) != 0) return cc;
          if (strlen(dout->data.cptr) < dout->dArrayLength) dout->dArrayLength = strlen(dout->data.cptr);
        }
        return 0;
    case PRP_TEMPLATE_NAM:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_NAME16) return illegal_format;
          for (i=0; i<CdiTemplateTblSize && i<(int)dout->dArrayLength; i++)
            strncpy(((NAME16 *)dout->data.vptr)[i].name,CdiTemplateTbl[i]->name,16);
        }
        return 0;    
    case PRP_TEMPLATE:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (devnr < 0 || devnr >= CdiTemplateTblSize) return illegal_equipment_number;
        tmpl = CdiTemplateTbl[devnr];
        tmplst = tmpl->cdiTemplateList;
        if (dout->dArrayLength > 0)
        {
          len = fmtsizeof(dout->dFormat);
          switch (dout->dFormat)
          {
            case CF_NAME16II:
              namlen = len - 8;
              for (i=0; i<tmpl->length && i<(int)dout->dArrayLength && tmplst != NULL; i++) 
              {
                strncpy(dout->data.cptr + i*len,tmplst->field,16);
                *(int *)(dout->data.bptr + i*len + namlen) = tmplst->addrParameters[0];
                *(int *)(dout->data.bptr + i*len + namlen + 4) = tmplst->addrParameters[1];
                tmplst = tmplst->next;
              }
              dout->dArrayLength = i;
              break;
            case CF_NAME16I:
            case CF_NAME32I:
            case CF_NAME64I:
              namlen = len - 4;
              for (i=0; i<tmpl->length && i<(int)dout->dArrayLength && tmplst != NULL; i++) 
              {
                strncpy(dout->data.cptr + i*len,tmplst->field,16);
                *(int *)(dout->data.bptr + i*len + namlen) = tmplst->addrParameters[0];
                tmplst = tmplst->next;
              }
              dout->dArrayLength = i;
              break;
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME64:
              namlen = len;
              for (i=0; i<tmpl->length && i<(int)dout->dArrayLength && tmplst != NULL; i++) 
              {
                strncpy(dout->data.cptr + i*len,tmplst->field,16);
                tmplst = tmplst->next;
              }
              dout->dArrayLength = i;
              break;
            default:
              return illegal_format;
          }
        }
        return 0;
    default:
        if (extid < 0) 
        {
          return illegal_property;
        }
    case PRP_WRITEREADWRITE_CLBR:
    case PRP_WRITEWRITE_CLBR:
    case PRP_READWRITE_CLBR:
    case PRP_WRITE_CLBR:
    case PRP_WRITEREAD_CLBR:
    case PRP_READ_CLBR:
    case PRP_WRITEREADWRITE:
    case PRP_WRITEWRITE:
    case PRP_WRITEREAD:
    case PRP_READWRITE:
    case PRP_WRITE:
    case PRP_READ:
        if (issync)
        { /* is synchronous : just do it */
          return ExecLinkEx(dn,devProperty,dout,din,access,2000);
        }
        else
        { /* return buffered data */
          if ((rate=getContractAccessRate(-1)) == 0) return code_failure;
          flags = isfirst ? CM_WAIT : 0; /* coming into scope ?*/
          cc = getDataFromListener(dn,devProperty,dout,din,&dt,rate,flags);
          setDataTimeStamp(dt);
          if (islast) 
          { /* going out of scope: stop listener */
            if ((ml=getListener(dn,devProperty,dout,din)) != NULL)
              stopListener(ml);
          }
          return cc;
        }
  }
}

void remapCdiDeviceAndProperty(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr=0,prpid,extid=-1,i,k;
  char *c = NULL, strbuffer[256];
  char *dn = devName;

  devnr = GetDeviceNumberEx(CDIEQM_TAG,devName,devProperty);
  prpid = GetPropertyId(CDIEQM_TAG,devProperty);
  if (prpid >= PRP_EXTENDED_OFFSET &&
      prpid < PRP_EXTENDED_OFFSET + gNumExtCdiProperties)
  {
    extid = prpid - PRP_EXTENDED_OFFSET;
  }
  if (extid < 0) return;
  if (cdiDebugLevel) dbglog("remapCdiDeviceAndProperty: input %s %s",devName,devProperty);
  if (strncmp(devName,devProperty,16))
  { /* a split extended property ! */
    if (devnr < 0) return;
    sprintf(dn,"%s.%s",devName,devProperty);
    if (access & CA_WRITE) 
    { /* a "SEND" to bus command */
      if (din->dArrayLength > 1)
      {
        k = devnr + (int)din->dArrayLength - 1;
        if (k >= extCdiPrps[extid].ndevs) k = extCdiPrps[extid].ndevs - 1;
        if (extCdiPrps[extid].isContiguous)
        { /* contiguous list ? */
          if (k > devnr)
          {
            sprintf(strbuffer," - %s.%s",extCdiPrps[extid].devs[k].name,devProperty);
            strcat(dn,strbuffer);
          }
        }
        else
        { /* group them one-by-one then */
          for (i=devnr+1; i<=k; i++)
          {
            sprintf(strbuffer,",%s.%s",extCdiPrps[extid].devs[i].name,devProperty);
            strcat(dn,strbuffer);
            if (strlen(dn) >= 4080)
            {
              k = i; break;
            }
          }
        }
        din->dArrayLength = k - devnr + 1;
      }
    }
    else
    { /* a "RECV" from bus command */
      if (dout->dArrayLength > 1)
      {
        k = devnr + (int)dout->dArrayLength - 1;
        if (k >= extCdiPrps[extid].ndevs) k = extCdiPrps[extid].ndevs - 1;
        if (extCdiPrps[extid].isContiguous)
        { /* contiguous list ? */
          if (k > devnr)
          {
            sprintf(strbuffer," - %s.%s",extCdiPrps[extid].devs[k].name,devProperty);
            strcat(dn,strbuffer);
          }
        }
        else
        { /* group them one-by-one then */
          for (i=devnr+1; i<=k; i++)
          {
            sprintf(strbuffer,",%s.%s",extCdiPrps[extid].devs[i].name,devProperty);
            strcat(dn,strbuffer);
            if (strlen(dn) >= 4080)
            {
              k = i; break;
            }
          }
        }
        dout->dArrayLength = k - devnr + 1;
      }
    }
  }        
  if (access & CA_WRITE) 
  {
    strcpy(devProperty,"SEND.CLBR");
  }
  else 
  {
    strcpy(devProperty,"RECV.CLBR");
  }
  if (cdiDebugLevel) dbglog("remapCdiDeviceAndProperty: output %s %s",devName,devProperty);
}